/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.control;

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;
import zleurtor.scanmanager.view.gui.ConfiguratorDialog;
import zleurtor.scanmanager.view.gui.GuiUtils;
import zleurtor.scanmanager.view.gui.SlavesPanel;

/**
 * The action triggered when the user click the "Configurator" button.
 *
 * @author Zleurtor
 */
public class ConfiguratorAction extends AbstractAction {

    private static final long serialVersionUID = -5457148423714960749L;

    /**
     * The model object which has to execute the options process.
     */
    private final Master master;

    /**
     * The dialog to display when the user trigger the action.
     */
    private ConfiguratorDialog dialog;

    /**
     * The slaves panel involved in the action.
     */
    private final SlavesPanel panel;

    /**
     * The options field involved in the action.
     */
    private final JTextField optionsField;

    /**
     * Constructs a new configurator action.
     *
     * @param thePanel The slaves panel involved in the action.
     * @param theOptionsField The options field involved in the action.
     * @param theMaster The model object which has to execute the options process.
     */
    public ConfiguratorAction(final Master theMaster, final SlavesPanel thePanel, final JTextField theOptionsField) {
        super(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_CONFIGURATOR));
        this.master = theMaster;
        this.panel = thePanel;
        this.optionsField = theOptionsField;
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        if (dialog == null) {
            try {
                dialog = new ConfiguratorDialog(SwingUtilities.getWindowAncestor(panel));
            } catch (final IOException e) {
                GuiUtils.showError(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_IMAGEREADINGERROR),
                        e, panel);
            }
            dialog.pack();
            dialog.setLocationRelativeTo(SwingUtilities.getWindowAncestor(panel));
        }

        dialog.parse(optionsField.getText());
        dialog.setColorBalance(panel.getCommonColorBalance());
        dialog.setColorsSamples(computeColorsSamples());
        dialog.setVisible(true);

        switch (dialog.getCloseOption()) {
            case APPLY_ALL:
                for (final SlaveInformation slave : panel.getSlaves()) {
                    master.setOptions(dialog.getOptions(), slave.getName());
                    slave.setColorBalance(dialog.getColorBalance());
                }
                panel.displayOptions();
                panel.repaint();
                break;
            case APPLY_SELECTED:
                for (final SlaveInformation slave : panel.getSelected()) {
                    master.setOptions(dialog.getOptions(), slave.getName());
                    slave.setColorBalance(dialog.getColorBalance());
                }
                panel.displayOptions();
                panel.repaint();
                break;
            case CANCEL:
            default:
                // Do nothing
        }
    }

    /**
     * Compute the list of files containing snapshots used as sample for color shift setting.
     *
     * @return The list of files containing snapshots used as sample for color shift setting.
     */
    private List<File> computeColorsSamples() {
        final List<File> colorSamples = new ArrayList<>();

        if (panel.getSelected() != null && !panel.getSelected().isEmpty()) {
            for (final SlaveInformation slaveInformation : panel.getSelected()) {
                if (slaveInformation.getSnapshotFilename() != null) {
                    colorSamples.add(new File(slaveInformation.getSnapshotFilename()));
                }
            }
        } else {
            for (final SlaveInformation slaveInformation : panel.getSlaves()) {
                if (slaveInformation.getSnapshotFilename() != null) {
                    colorSamples.add(new File(slaveInformation.getSnapshotFilename()));
                }
            }
        }

        return colorSamples;
    }

}
