/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.control;

import java.awt.Component;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.MessageFormat;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;
import zleurtor.scanmanager.view.gui.ConfiguratorDialog;
import zleurtor.scanmanager.view.gui.GuiUtils;
import zleurtor.scanmanager.view.gui.SaveDialog;
import zleurtor.scanmanager.view.gui.SlavesPanel;

/**
 * The action triggered when the user click the "Save" button.
 *
 * @author Zleurtor
 */
public class SaveAction extends AbstractAction {

    private static final long serialVersionUID = 2298782119990589383L;

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(SaveAction.class);

    /**
     * The model object which has to execute the save process.
     */
    private final Master master;

    /**
     * The slave panel containing all the information on the slaves and the snapshots to save.
     */
    private SlavesPanel slavesPanel;

    /**
     * The dialog displayed during the save process.
     */
    private final SaveDialog dialog = new SaveDialog();

    /**
     * Constructs a new save action.
     *
     * @param theMaster The model object which has to execute the save process.
     */
    public SaveAction(final Master theMaster) {
        super();
        this.master = theMaster;

        master.addMasterListener(dialog);
    }

    /**
     * Sets a new slave panel containing all the information on the slaves and the snapshots to save.
     *
     * @param newSlavesPanel The new slave panel containing all the information on the slaves and the snapshots to save.
     */
    public final void setSlavesPanel(final SlavesPanel newSlavesPanel) {
        this.slavesPanel = newSlavesPanel;
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        final JFileChooser fileChooser = new JFileChooser();
        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setFileFilter(new FileFilter() {
            @Override
            public boolean accept(final File f) {
                return f.isDirectory() || f.getName().endsWith(".zip");
            }

            @Override
            public String getDescription() {
                return ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_ZIPDESCRIPTION);
            }
        });

        File outFile = null;
        if (fileChooser.showSaveDialog(slavesPanel) != JFileChooser.APPROVE_OPTION) {
            return;
        }

        outFile = fileChooser.getSelectedFile();
        if (!outFile.getName().endsWith(".zip")) {
            outFile = new File(outFile.getParentFile(), outFile.getName() + ".zip");
        }

        final Window owner = SwingUtilities.getWindowAncestor((Component) evt.getSource());
        if (outFile.exists() && JOptionPane.showConfirmDialog(
                owner,
                MessageFormat.format(ResourcesManager.getInstance().getString(ResourcesManager.KEY_CONFIRM_OVERWRITE), outFile),
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_CONFIRM), JOptionPane.YES_NO_OPTION)
                != JOptionPane.YES_OPTION) {
            return;
        }

        try {
            final File optionsFile = File.createTempFile("options-", ".txt");
            final Map<File, int[]> inFiles = new LinkedHashMap<>();

            try (PrintWriter writer = new PrintWriter(new FileOutputStream(optionsFile), true)) {
                for (final SlaveInformation si : slavesPanel.getSlaves()) {
                    if (si.getSnapshotFilename() != null) {
                        inFiles.put(new File(si.getSnapshotFilename()), si.getColorBalance());
                    }
                    writer.println(si.getIp() + ":");
                    writer.println(ConfiguratorDialog.parseToReadableFormat(si.getOptions()));
                }
            }

            dialog.setModal(false);
            dialog.setVisible(true);

            final File finalOutFile = outFile;
            final Thread t = new Thread(() -> {
                try {
                    while (!dialog.isVisible()) {
                        // Do nothing ... Wait
                    }

                    master.save(inFiles, optionsFile, finalOutFile);
                } catch (final IOException e) {
                    GuiUtils.showWarning(e, slavesPanel);
                    LOGGER.warn(e.getMessage(), e);
                } finally {
                    dialog.dispose();
                }
            });
            t.setName("Saver Thread");
            t.start();

            dialog.setModal(true);
            dialog.setVisible(true);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, slavesPanel);
            LOGGER.warn(e.getMessage(), e);
        }
    }

}
