/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.control;

import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;
import zleurtor.scanmanager.view.gui.GuiUtils;
import zleurtor.scanmanager.view.gui.SlavesPanel;

/**
 * The action triggered when the user click the "Shutdown" button.
 *
 * @author Zleurtor
 */
public class ShutdownAction extends AbstractAction {

    private static final long serialVersionUID = -639880709605429157L;

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(ShutdownAction.class);

    /**
     * The model object which has to execute the Shutdown process.
     */
    private final Master master;

    /**
     * The panel from which the action shall be done.
     */
    private final SlavesPanel slavesPanel;

    /**
     * Constructs a new shutdown action.
     *
     * @param theMaster The model object which has to execute the Shutdown process.
     * @param theSlavesPanel The panel from which the action shall be done.
     */
    public ShutdownAction(final Master theMaster, final SlavesPanel theSlavesPanel) {
        super();
        this.master = theMaster;
        this.slavesPanel = theSlavesPanel;
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        final String rebootOption = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_REBOOT);
        final String shutdownOption = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_SHUTDOWN);
        final String rebootSelectedOption = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_REBOOTSELECTED);
        final String shutdownSelectedOption = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_SHUTDOWNSELECTED);

        try {
            final Object selected = JOptionPane.showInputDialog(slavesPanel,
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_INPUTREBOOTSHUTDOWN),
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_INPUT),
                    JOptionPane.QUESTION_MESSAGE,
                    new ImageIcon(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_OPTIONPANE_QUESTIONICON)),
                    evt.getSource() instanceof SlaveInformation
                            ? new String[] { rebootOption, shutdownOption }
                            : new String[] { rebootOption, shutdownOption, rebootSelectedOption, shutdownSelectedOption },
                    shutdownOption);

            if (selected == null) {
                return;
            } else if (evt.getSource() instanceof SlaveInformation) {
                // Reboot shutdown the slave
                master.shutdown(((SlaveInformation) evt.getSource()).getName(), selected.equals(rebootOption));
            } else if (selected.equals(rebootOption) || selected.equals(shutdownOption)) {
                // Reboot shutdown all slaves
                master.shutdown(selected.equals(rebootOption));
            } else if (selected.equals(rebootSelectedOption) || selected.equals(shutdownSelectedOption)) {
                // Reboot shutdown selected slaves
                slavesPanel.getSelected().forEach(
                    (SlaveInformation si) -> master.shutdown(si.getName(), selected.equals(rebootSelectedOption)));
            }
        } catch (final HeadlessException | IOException e) {
            GuiUtils.showWarning(e, slavesPanel);
            LOGGER.warn(e.getMessage(), e);
        }
    }

}
