/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.control;

import java.awt.event.ActionEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.JOptionPane;

import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;
import zleurtor.scanmanager.view.gui.SlavesPanel;
import zleurtor.scanmanager.view.gui.SnapshotsDialog;

/**
 * The action triggered when the user click the "Take snapshot" button.
 *
 * @author Zleurtor
 */
public class SnapshotAction extends AbstractAction {

    private static final long serialVersionUID = -2429136427817322416L;

    /**
     * The model object which has to execute the save process.
     */
    private final Master master;

    /**
     * The slave panel containing all the information on the slaves.
     */
    private SlavesPanel slavesPanel;

    /**
     * The dialog displayed during the snapshot process.
     */
    private SnapshotsDialog snapshotsDialog;

    /**
     * Constructs a new snapshot action.
     *
     * @param theMaster The model object which has to execute the save process.
     * @param theSlavesPanel The slave panel containing all the information on the slaves.
     */
    public SnapshotAction(final Master theMaster, final SlavesPanel theSlavesPanel) {
        super();
        this.master = theMaster;
        this.slavesPanel = theSlavesPanel;
        this.snapshotsDialog = new SnapshotsDialog();
        snapshotsDialog.setModal(true);
        master.addMasterListener(snapshotsDialog);
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        if (slavesPanel.getSlaves().isEmpty()) {
            JOptionPane.showMessageDialog(slavesPanel,
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_NOSLAVE),
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_INFO),
                    JOptionPane.INFORMATION_MESSAGE);
        } else {
            new Thread(() -> {
                snapshotsDialog.reset();
                snapshotsDialog.setTotalImages(slavesPanel.getSlaves().size());
                snapshotsDialog.setVisible(true);
            }).start();

            // Create the options map
            final Map<String, String> options = new HashMap<>();
            for (final SlaveInformation si : slavesPanel.getSlaves()) {
                options.put(si.getName(), si.getOptions());
            }

            slavesPanel.setWaitForQuack(true);

            master.takeSnapshot();
        }
    }

}
