/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.control;

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;
import java.util.zip.GZIPInputStream;

import javax.swing.AbstractAction;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import org.kamranzafar.jtar.TarEntry;
import org.kamranzafar.jtar.TarInputStream;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.model.Utils;
import zleurtor.scanmanager.model.events.MasterUpdaterEvent;
import zleurtor.scanmanager.model.update.ArchivaConnector;
import zleurtor.scanmanager.model.update.ArtifactRepositoryManagerConnector;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;
import zleurtor.scanmanager.view.gui.GuiUtils;
import zleurtor.scanmanager.view.gui.SlavesPanel;
import zleurtor.scanmanager.view.gui.UpdateDialog;

/**
 * The action triggered when the user click the "Update" button.
 *
 * @author Zleurtor
 */
public class UpdateAction extends AbstractAction {

    private static final long serialVersionUID = 1179803724809879136L;

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(ShutdownAction.class);

    /**
     * The application's configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    /**
     * The "download latest update file" option.
     */
    private static final String DOWNLOAD_OPTION = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_DOWNLOAD_UPDATE);

    /**
     * The "use a local update file" option.
     */
    private static final String LOCAL_FILE_OPTION = ResourcesManager.getInstance().getString(ResourcesManager.KEY_INPUT_LOCAL_UPDATE_FILE);

    /**
     * The file dialog allowing the usert to select a local update file.
     */
    private static final JFileChooser FILE_DIALOG = new JFileChooser();

    /**
     * The dialog to display while a new version is searched.
     */
    private static final UpdateDialog UPDATE_DIALOG = new UpdateDialog();

    /**
     * The archiva artifact repository manager URL.
     */
    private static final String ARCHIVA_URL = CONFIGURATION.getString("artifactRepositoryManager.url");

    /**
     * The slave's group id.
     */
    private static final String GROUP_ID = CONFIGURATION.getString("slave.groupId");

    /**
     * The slave's artifact id.
     */
    private static final String ARTIFACT_ID = CONFIGURATION.getString("slave.artifactId");

    /**
     * The slave's classifier.
     */
    private static final String CLASSIFIER = CONFIGURATION.getString("slave.classifier");

    /**
     * The slave's packaging.
     */
    private static final String PACKAGING = CONFIGURATION.getString("slave.packaging");

    /**
     * The model object which has to execute the update process.
     */
    private final Master master;

    /**
     * The slave panel containing all the information on the slaves.
     */
    private SlavesPanel slavesPanel;

    /**
     * Constructs a new update action.
     *
     * @param theMaster The model object which has to execute the update process.
     * @param theSlavesPanel The slave panel containing all the information on the slaves.
     */
    public UpdateAction(final Master theMaster, final SlavesPanel theSlavesPanel) {
        super();
        this.master = theMaster;
        this.slavesPanel = theSlavesPanel;
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        new Thread(() -> {
            try {
                // Download or select a local file
                final File updateFile;
                final String filePolicy = selectFilePolicy();
                if (filePolicy != null && filePolicy.equals(DOWNLOAD_OPTION)) {
                    updateFile = downloadLatestUpdateFile();
                } else if (filePolicy != null && filePolicy.equals(LOCAL_FILE_OPTION)) {
                    updateFile = selectUpdateFile();
                    if (updateFile == null) {
                        // The user closed the selection dialog
                        return;
                    }
                } else {
                    // The user closed the selection dialog
                    return;
                }

                UPDATE_DIALOG.setVisible(false);

                final Collection<SlaveInformation> toUpdate = checkVersion(updateFile);
                if (toUpdate == null) {
                    // No update !
                    return;
                }

                UPDATE_DIALOG.installing(null);
                UPDATE_DIALOG.setVisible(true);

                toUpdate.stream().forEach(slaveInformation -> master.updateSlave(updateFile,
                        ARTIFACT_ID + "-" + CLASSIFIER + "-",
                        "." + PACKAGING,
                        slaveInformation.getName()));
            } catch (final KeyManagementException | NoSuchAlgorithmException | KeyStoreException | IOException | URISyntaxException e) {
                GuiUtils.showError(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_UPDATEERROR), e, slavesPanel);
                return;
            } finally {
                if (UPDATE_DIALOG.isVisible()) {
                    UPDATE_DIALOG.setVisible(false);
                }
            }
        }).start();
    }

    /**
     * Check the update file version against the slaves version then return the list of slaves needing an update.
     *
     * @param updateFile The update file from which the version shall be checked.
     * @return The list of slaves needing an update.
     */
    private Collection<SlaveInformation> checkVersion(final File updateFile) {
        // Search the version in the update file
        final String version = extractUpdateVersion(updateFile);

        // Search the slaves to update
        final Collection<SlaveInformation> toUpdate = new ArrayList<>();
        if (version != null) {
            for (final SlaveInformation si : slavesPanel.getSlaves()) {
                if (si.getVersion() == null || Utils.compareVersion(version, si.getVersion()) > 0) {
                    toUpdate.add(si);
                }
            }
        }

        if (version == null || toUpdate.isEmpty()) {
            JOptionPane.showMessageDialog(slavesPanel,
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_NOUPDATE),
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_INFO),
                    JOptionPane.INFORMATION_MESSAGE);
            return null;
        }

        if (JOptionPane.showConfirmDialog(slavesPanel,
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_UPDATESLAVES, version, toUpdate.size()),
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_CONFIRM),
                JOptionPane.YES_NO_OPTION,
                JOptionPane.INFORMATION_MESSAGE) == JOptionPane.NO_OPTION) {
            return null;
        }

        return toUpdate;
    }

    /**
     * Ask the user if the update file shall be downloaded or if a local file shall be used instead.
     *
     * @return The update file obtention policy (download or local).
     */
    private String selectFilePolicy() {
        try {
            final Object selected = JOptionPane.showInputDialog(slavesPanel,
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_INPUTDOWNLOADLOCAL),
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_INPUT),
                    JOptionPane.QUESTION_MESSAGE,
                    new ImageIcon(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_OPTIONPANE_QUESTIONICON)),
                    new String[] { DOWNLOAD_OPTION, LOCAL_FILE_OPTION },
                    DOWNLOAD_OPTION);

            return selected != null ? selected.toString() : null;
        } catch (final IOException e) {
            GuiUtils.showWarning(e, slavesPanel);
            LOGGER.warn(e.getMessage(), e);

            return null;
        }
    }

    /**
     * Search the latest available version.
     *
     * @return The latest available version.
     */
    private String searchLatestVersion() {
        String latestVersion = null;

        try (ArtifactRepositoryManagerConnector connector = new ArchivaConnector(ARCHIVA_URL)) {
            latestVersion = connector.searchLatestVersion(GROUP_ID, ARTIFACT_ID);
        } catch (final KeyManagementException | NoSuchAlgorithmException | KeyStoreException | IOException e) {
            JOptionPane.showMessageDialog(slavesPanel,
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_UPDATEERROR),
                    ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_ERROR),
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        return latestVersion;
    }

    /**
     * Download the latest update file (if needed).
     *
     * @return The lastest update filr found or null if not any file was found.
     * @throws KeyManagementException If the SSL context builder fails to initialize.
     * @throws NoSuchAlgorithmException If no Provider supports aTrustManagerFactorySpi implementation for the "trust all" strategy.
     * @throws KeyStoreException If the TrustStrategy fails to load.
     * @throws URISyntaxException If a URI is malformed.
     */
    private File downloadLatestUpdateFile()
            throws KeyManagementException, NoSuchAlgorithmException, KeyStoreException, IOException, URISyntaxException {
        UPDATE_DIALOG.updateFound(null);
        UPDATE_DIALOG.setVisible(true);

        final String latestVersion = searchLatestVersion();

        try (ArtifactRepositoryManagerConnector connector = new ArchivaConnector(ARCHIVA_URL)) {
            UPDATE_DIALOG.setVisible(true);

            final URL artifactUrl = connector.searchArtifactUrl(GROUP_ID, ARTIFACT_ID, latestVersion, CLASSIFIER, PACKAGING);
            final int artifactSize = connector.searchArtifactSize(GROUP_ID, ARTIFACT_ID, latestVersion, CLASSIFIER, PACKAGING);
            final String prefix = ARTIFACT_ID + "-" + CLASSIFIER + "-";
            final String suffix = "." + PACKAGING;
            final File dest = File.createTempFile(prefix, suffix);
            connector.addArtifactRepositoryManagerConnectorListener(evt -> {
                UPDATE_DIALOG.downloading(new MasterUpdaterEvent(this, evt.getArtifactSize(), evt.getDownloadedSize()));
            });
            connector.download(artifactUrl, artifactSize, dest);

            return dest;
        }
    }

    /**
     * Lets the user select a local tar gz update file.
     *
     * @return The selected update file.
     */
    private File selectUpdateFile() {
        FILE_DIALOG.setFileSelectionMode(JFileChooser.FILES_ONLY);
        FILE_DIALOG.setMultiSelectionEnabled(false);
        FILE_DIALOG.setFileFilter(new FileFilter() {
            @Override
            public boolean accept(final File f) {
                return f.isDirectory() || f.getName().endsWith("." + PACKAGING);
            }

            @Override
            public String getDescription() {
                return "Gzipped tar file.";
            }
        });
        FILE_DIALOG.showOpenDialog(slavesPanel);

        return FILE_DIALOG.getSelectedFile();
    }

    /**
     * Extract the version number from the given update file.
     *
     * @param updateFile The file from which the version shall be extracted.
     * @return The version number of the given update file.
     */
    private String extractUpdateVersion(final File updateFile) {
        final Properties configuration = new Properties();

        try (TarInputStream in = new TarInputStream(new GZIPInputStream(new FileInputStream(updateFile)))) {
            for (TarEntry entry = in.getNextEntry(); entry != null; entry = in.getNextEntry()) {
                if (entry.getName().equals("conf/configuration.properties")) {
                    configuration.load(in);
                    return configuration.get("version").toString();
                }
            }
        } catch (final IOException e) {
            GuiUtils.showError(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_UPDATEERROR), slavesPanel);
            return null;
        }

        return null;
    }

}
