/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.xpath.XPathExpressionException;

import org.xml.sax.SAXException;

import zleurtor.scanmanager.model.messages.AbstractMessage;

/**
 * The mother class of entity objects that can communicate (ie read/write messages).
 *
 * @author Zleurtor
 */
public abstract class AbstractCommunicator {

    /**
     * Read a message from the specified input stream.
     *
     * @param in The input stream from which the next message shall be read.
     * @return The next message in the specified stream.
     * @throws IOException If any exception occurs.
     */
    protected AbstractMessage readMessage(final InputStream in) throws IOException {
        try {
            final byte[] sizeBuffer = new byte[Integer.SIZE / Byte.SIZE];
            int read = in.read(sizeBuffer);

            final int size = Utils.toInt(sizeBuffer);
            if (size == 0) {
                throw new IOException("Empty message");
            }

            final byte[] buffer = new byte[size];
            int totalRead = 0;

            while (totalRead < size) {
                read = in.read(buffer, totalRead, size - totalRead);
                totalRead += read;
            }

            final AbstractMessage message = AbstractMessage.unserialize(buffer);
            return message;
        } catch (final IOException | XPathExpressionException | InstantiationException | IllegalAccessException | ClassNotFoundException
                     | SAXException | ParserConfigurationException | IllegalArgumentException | InvocationTargetException
                     | NoSuchMethodException | SecurityException e) {
            throw new IOException(e);
        }
    }

    /**
     * Write the specified message into the specified output stream.
     *
     * @param out The output stream into which the next message shall be written.
     * @param message The message which shall be written.
     * @throws IOException If any exception occurs.
     */
    protected synchronized void writeMessage(final OutputStream out, final AbstractMessage message) throws IOException {
        try {
            final byte[] serialized = message.serialize();

            out.write(Utils.toArray(serialized.length));
            out.write(serialized);

            out.flush();
        } catch (final ParserConfigurationException | TransformerFactoryConfigurationError | TransformerException
                | IOException e) {
            throw new IOException(e);
        }
    }

}
