/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Enumeration;
import java.util.Properties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The class having in charge to read and store the configuration.
 *
 * @author Zleurtor
 */
public final class Configuration {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(Configuration.class);

    /**
     * The singleton instance of the configuration.
     */
    private static final Configuration INSTANCE = new Configuration();

    /**
     * The properties containing all the configuration.
     */
    private final Properties properties = new OrderedProperties();

    /**
     * The configuration file.
     */
    private final File file;

    /**
     * Constructs a new configuration by reading the configuration file.
     */
    private Configuration() {
        File tmpFile = null;

        try {
            final Enumeration<URL> resources = getClass().getClassLoader().getResources("configuration.properties");
            while (resources.hasMoreElements()) {
                final URL resource = resources.nextElement();
                LOGGER.debug("Loading from " + resource + "...");
                properties.load(resource.openStream());
            }

            final URL resource = getClass().getResource("/configuration.properties");
            tmpFile = new File(URLDecoder.decode(resource.getFile(), "UTF-8"));
            LOGGER.debug("Loading from " + resource + "...");
            properties.load(resource.openStream());
        } catch (final IOException e) {
            LOGGER.error(e.getMessage(), e);
            LOGGER.error("Exiting.");
            System.exit(-1);
        }

        file = tmpFile;
    }

    /**
     * Returns the singleton instance of the configuration.
     *
     * @return The singleton instance of the configuration.
     */
    public static Configuration getInstance() {
        return INSTANCE;
    }

    /**
     * Searches for the property with the specified key in this property list.
     *
     * @param key The property key.
     * @return The value in this property list with the specified key value.
     */
    public String getString(final String key) {
        return properties.getProperty(key);
    }

    /**
     * Searches for the property with the specified key in this property list.
     *
     * @param key The property key.
     * @return The value (as an integer) in this property list with the specified key value.
     */
    public int getInt(final String key) {
        return Integer.parseInt(properties.getProperty(key));
    }

    /**
     * Set the specified value to the specified key.
     *
     * @param key The key for which the value shall be set.
     * @param value The value to set to the key.
     */
    public void setString(final String key, final String value) {
        if (properties.getProperty(key) == null || !properties.getProperty(key).equals(value)) {
            properties.setProperty(key, value);

            try (OutputStream out = new FileOutputStream(file)) {
                LOGGER.debug("Saving to file " + file + "...");
                properties.store(out, null);
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
            }
        }
    }

    /**
     * Update the version in the main configuration file.
     *
     * @param newVersion The new version to set in the configuration file.
     * @throws IOException if an error occurred when reading/writing the configuration file.
     */
    public void updateVersion(final String newVersion)
            throws IOException {
        final Properties prop = new OrderedProperties();
        final URL resource = getClass().getResource("/configuration.properties");
        final File dest = new File(resource.getFile());

        try (InputStream in = resource.openStream()) {
            prop.load(in);
        }
        try (OutputStream out = new FileOutputStream(dest)) {
            prop.setProperty("version", newVersion);
            prop.store(out, null);
        }
    }

}
