/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

/**
 * A class to read Exif files (jpeg with metadata). See { @link https://www.media.mit.edu/pia/Research/deepview/exif.html }.
 * @author Zleurtor
 */
public class ExifReader {

    /**
     * The enumeration of all tags used in TIFF.
     *
     * @author Zleurtor
     */
    private static enum Tag {
        /**
         * The ImageDescription tag.
         */
        ImageDescription((short) 0x010e, "ImageDescription"),
        /**
         * The Make tag.
         */
        Make((short) 0x010f, "Make"),
        /**
         * The Model tag.
         */
        Model((short) 0x0110, "Model"),
        /**
         * The Orientation tag.
         */
        Orientation((short) 0x0112, "Orientation"),
        /**
         * The XResolution tag.
         */
        XResolution((short) 0x011a, "XResolution"),
        /**
         * The YResolution tag.
         */
        YResolution((short) 0x011b, "YResolution"),
        /**
         * The ResolutionUnit tag.
         */
        ResolutionUnit((short) 0x0128, "ResolutionUnit"),
        /**
         * The Software tag.
         */
        Software((short) 0x0131, "Software"),
        /**
         * The DateTime tag.
         */
        DateTime((short) 0x0132, "DateTime"),
        /**
         * The WhitePoint tag.
         */
        WhitePoint((short) 0x013e, "WhitePoint"),
        /**
         * The PrimaryChromaticities tag.
         */
        PrimaryChromaticities((short) 0x013f, "PrimaryChromaticities"),
        /**
         * The YCbCrCoefficients tag.
         */
        YCbCrCoefficients((short) 0x0211, "YCbCrCoefficients"),
        /**
         * The YCbCrPositioning tag.
         */
        YCbCrPositioning((short) 0x0213, "YCbCrPositioning"),
        /**
         * The ReferenceBlackWhite tag.
         */
        ReferenceBlackWhite((short) 0x0214, "ReferenceBlackWhite"),
        /**
         * The Copyright tag.
         */
        Copyright((short) 0x8298, "Copyright"),
        /**
         * The ExifOffset tag.
         */
        ExifOffset((short) 0x8769, "ExifOffset"),
        /**
         * The ExposureTime tag.
         */
        ExposureTime((short) 0x829a, "ExposureTime"),
        /**
         * The FNumber tag.
         */
        FNumber((short) 0x829d, "FNumber"),
        /**
         * The ExposureProgram tag.
         */
        ExposureProgram((short) 0x8822, "ExposureProgram"),
        /**
         * The ISOSpeedRatings tag.
         */
        ISOSpeedRatings((short) 0x8827, "ISOSpeedRatings"),
        /**
         * The ExifVersion tag.
         */
        ExifVersion((short) 0x9000, "ExifVersion"),
        /**
         * The DateTimeOriginal tag.
         */
        DateTimeOriginal((short) 0x9003, "DateTimeOriginal"),
        /**
         * The DateTimeDigitized tag.
         */
        DateTimeDigitized((short) 0x9004, "DateTimeDigitized"),
        /**
         * The ComponentConfiguration tag.
         */
        ComponentConfiguration((short) 0x9101, "ComponentConfiguration"),
        /**
         * The CompressedBitsPerPixel tag.
         */
        CompressedBitsPerPixel((short) 0x9102, "CompressedBitsPerPixel"),
        /**
         * The ShutterSpeedValue tag.
         */
        ShutterSpeedValue((short) 0x9201, "ShutterSpeedValue"),
        /**
         * The ApertureValue tag.
         */
        ApertureValue((short) 0x9202, "ApertureValue"),
        /**
         * The BrightnessValue tag.
         */
        BrightnessValue((short) 0x9203, "BrightnessValue"),
        /**
         * The ExposureBiasValue tag.
         */
        ExposureBiasValue((short) 0x9204, "ExposureBiasValue"),
        /**
         * The MaxApertureValue tag.
         */
        MaxApertureValue((short) 0x9205, "MaxApertureValue"),
        /**
         * The SubjectDistance tag.
         */
        SubjectDistance((short) 0x9206, "SubjectDistance"),
        /**
         * The MeteringMode tag.
         */
        MeteringMode((short) 0x9207, "MeteringMode"),
        /**
         * The LightSource tag.
         */
        LightSource((short) 0x9208, "LightSource"),
        /**
         * The Flash tag.
         */
        Flash((short) 0x9209, "Flash"),
        /**
         * The FocalLength tag.
         */
        FocalLength((short) 0x920a, "FocalLength"),
        /**
         * The MakerNote tag.
         */
        MakerNote((short) 0x927c, "MakerNote"),
        /**
         * The UserComment tag.
         */
        UserComment((short) 0x9286, "UserComment"),
        /**
         * The FlashPixVersion tag.
         */
        FlashPixVersion((short) 0xa000, "FlashPixVersion"),
        /**
         * The ColorSpace tag.
         */
        ColorSpace((short) 0xa001, "ColorSpace"),
        /**
         * The ExifImageWidth tag.
         */
        ExifImageWidth((short) 0xa002, "ExifImageWidth"),
        /**
         * The ExifImageHeight tag.
         */
        ExifImageHeight((short) 0xa003, "ExifImageHeight"),
        /**
         * The RelatedSoundFile tag.
         */
        RelatedSoundFile((short) 0xa004, "RelatedSoundFile"),
        /**
         * The ExifInteroperabilityOffset tag.
         */
        ExifInteroperabilityOffset((short) 0xa005, "ExifInteroperabilityOffset"),
        /**
         * The FocalPlaneXResolution tag.
         */
        FocalPlaneXResolution((short) 0xa20e, "FocalPlaneXResolution"),
        /**
         * The FocalPlaneYResolution tag.
         */
        FocalPlaneYResolution((short) 0xa20f, "FocalPlaneYResolution"),
        /**
         * The FocalPlaneResolutionUnit tag.
         */
        FocalPlaneResolutionUnit((short) 0xa210, "FocalPlaneResolutionUnit"),
        /**
         * The SensingMethod tag.
         */
        SensingMethod((short) 0xa217, "SensingMethod"),
        /**
         * The FileSource tag.
         */
        FileSource((short) 0xa300, "FileSource"),
        /**
         * The SceneType tag.
         */
        SceneType((short) 0xa301, "SceneType"),
        /**
         * The ImageWidth tag.
         */
        ImageWidth((short) 0x0100, "ImageWidth"),
        /**
         * The ImageLength tag.
         */
        ImageLength((short) 0x0101, "ImageLength"),
        /**
         * The BitsPerSample tag.
         */
        BitsPerSample((short) 0x0102, "BitsPerSample"),
        /**
         * The Compression tag.
         */
        Compression((short) 0x0103, "Compression"),
        /**
         * The PhotometricInterpretation tag.
         */
        PhotometricInterpretation((short) 0x0106, "PhotometricInterpretation"),
        /**
         * The StripOffsets tag.
         */
        StripOffsets((short) 0x0111, "StripOffsets"),
        /**
         * The SamplesPerPixel tag.
         */
        SamplesPerPixel((short) 0x0115, "SamplesPerPixel"),
        /**
         * The RowsPerStrip tag.
         */
        RowsPerStrip((short) 0x0116, "RowsPerStrip"),
        /**
         * The StripByteConunts tag.
         */
        StripByteConunts((short) 0x0117, "StripByteConunts"),
        /**
         * The PlanarConfiguration tag.
         */
        PlanarConfiguration((short) 0x011c, "PlanarConfiguration"),
        /**
         * The JpegIFOffset tag.
         */
        JpegIFOffset((short) 0x0201, "JpegIFOffset"),
        /**
         * The JpegIFByteCount tag.
         */
        JpegIFByteCount((short) 0x0202, "JpegIFByteCount"),
        /**
         * The YCbCrSubSampling tag.
         */
        YCbCrSubSampling((short) 0x0212, "YCbCrSubSampling"),
        /**
         * The NewSubfileType tag.
         */
        NewSubfileType((short) 0x00fe, "NewSubfileType"),
        /**
         * The SubfileType tag.
         */
        SubfileType((short) 0x00ff, "SubfileType"),
        /**
         * The TransferFunction tag.
         */
        TransferFunction((short) 0x012d, "TransferFunction"),
        /**
         * The Artist tag.
         */
        Artist((short) 0x013b, "Artist"),
        /**
         * The Predictor tag.
         */
        Predictor((short) 0x013d, "Predictor"),
        /**
         * The TileWidth tag.
         */
        TileWidth((short) 0x0142, "TileWidth"),
        /**
         * The TileLength tag.
         */
        TileLength((short) 0x0143, "TileLength"),
        /**
         * The TileOffsets tag.
         */
        TileOffsets((short) 0x0144, "TileOffsets"),
        /**
         * The TileByteCounts tag.
         */
        TileByteCounts((short) 0x0145, "TileByteCounts"),
        /**
         * The SubIFDs tag.
         */
        SubIFDs((short) 0x014a, "SubIFDs"),
        /**
         * The JPEGTables tag.
         */
        JPEGTables((short) 0x015b, "JPEGTables"),
        /**
         * The CFARepeatPatternDim tag.
         */
        CFARepeatPatternDim((short) 0x828d, "CFARepeatPatternDim"),
        /**
         * The CFAPattern tag.
         */
        CFAPattern((short) 0x828e, "CFAPattern"),
        /**
         * The BatteryLevel tag.
         */
        BatteryLevel((short) 0x828f, "BatteryLevel"),
        /**
         * The IPTC/NAA tag.
         */
        IPTC_NAA((short) 0x83bb, "IPTC/NAA"),
        /**
         * The InterColorProfile tag.
         */
        InterColorProfile((short) 0x8773, "InterColorProfile"),
        /**
         * The SpectralSensitivity tag.
         */
        SpectralSensitivity((short) 0x8824, "SpectralSensitivity"),
        /**
         * The GPSInfo tag.
         */
        GPSInfo((short) 0x8825, "GPSInfo"),
        /**
         * The OECF tag.
         */
        OECF((short) 0x8828, "OECF"),
        /**
         * The Interlace tag.
         */
        Interlace((short) 0x8829, "Interlace"),
        /**
         * The TimeZoneOffset tag.
         */
        TimeZoneOffset((short) 0x882a, "TimeZoneOffset"),
        /**
         * The SelfTimerMode tag.
         */
        SelfTimerMode((short) 0x882b, "SelfTimerMode"),
        /**
         * The FlashEnergy tag.
         */
        FlashEnergy((short) 0x920b, "FlashEnergy"),
        /**
         * The SpatialFrequencyResponse tag.
         */
        SpatialFrequencyResponse((short) 0x920c, "SpatialFrequencyResponse"),
        /**
         * The Noise tag.
         */
        Noise((short) 0x920d, "Noise"),
        /**
         * The ImageNumber tag.
         */
        ImageNumber((short) 0x9211, "ImageNumber"),
        /**
         * The SecurityClassification tag.
         */
        SecurityClassification((short) 0x9212, "SecurityClassification"),
        /**
         * The ImageHistory tag.
         */
        ImageHistory((short) 0x9213, "ImageHistory"),
        /**
         * The SubjectLocation tag.
         */
        SubjectLocation((short) 0x9214, "SubjectLocation"),
        /**
         * The ExposureIndex tag.
         */
        ExposureIndex((short) 0x9215, "ExposureIndex"),
        /**
         * The TIFF/EPStandardID tag.
         */
        TIFF_EPStandardID((short) 0x9216, "TIFF/EPStandardID"),
        /**
         * The SubSecTime tag.
         */
        SubSecTime((short) 0x9290, "SubSecTime"),
        /**
         * The SubSecTimeOriginal tag.
         */
        SubSecTimeOriginal((short) 0x9291, "SubSecTimeOriginal"),
        /**
         * The SubSecTimeDigitized tag.
         */
        SubSecTimeDigitized((short) 0x9292, "SubSecTimeDigitized"),
        /**
         * An other FlashEnergy tag.
         */
        _FlashEnergy((short) 0xa20b, "FlashEnergy"),
        /**
         * An other SpatialFrequencyResponse tag.
         */
        _SpatialFrequencyResponse((short) 0xa20c, "SpatialFrequencyResponse"),
        /**
         * An other SubjectLocation tag.
         */
        _SubjectLocation((short) 0xa214, "SubjectLocation"),
        /**
         * An other ExposureIndex tag.
         */
        _ExposureIndex((short) 0xa215, "ExposureIndex"),
        /**
         * An other CFAPattern tag.
         */
        _CFAPattern((short) 0xa302, "CFAPattern");

        /**
         * The tag id.
         */
        private final short id;

        /**
         * The tag name.
         */
        private final String name;

        /**
         * Construct a new tag having the specified id and name.
         *
         * @param theId The tag id.
         * @param theName The tag name.
         */
        Tag(final short theId, final String theName) {
            this.id = theId;
            this.name = theName;
        }

        /**
         * Returns tag id.
         *
         * @return Tag id.
         */
        public short getId() {
            return id;
        }

        /**
         * Returns tag name.
         *
         * @return Tag name.
         */
        public String getName() {
            return name;
        }

        /**
         * Search and returns the tag having the specified id if any, returns null otherwise.
         *
         * @param anId The id of the tag to search.
         * @return The tag having the specified id if any, returns null otherwise.
         */
        private static Tag getTagById(final short anId) {
            for (final Tag tagNumber : Tag.values()) {
                if (tagNumber.getId() == anId) {
                    return tagNumber;
                }
            }

            return null;
        }

    }

    /**
     * All the data format used in TIFF tags.
     *
     * @author Zleurtor
     */
    private static enum DataFormat {
        /**
         * The unsigned byte data format.
         */
        UNSIGNED_BYTE((short) 1, 1),
        /**
         * The ascii string data format.
         */
        ASCII_STRINGS((short) 2, 1),
        /**
         * The unsigned short data format.
         */
        UNSIGNED_SHORT((short) 3, 2),
        /**
         * The unsigned long data format.
         */
        UNSIGNED_LONG((short) 4, 4),
        /**
         * The unsigned float data format.
         */
        UNSIGNED_RATIONAL((short) 5, 8),
        /**
         * The byte data format.
         */
        SIGNED_BYTE((short) 6, 1),
        /**
         * The undefined data format.
         */
        UNDEFINED((short) 7, 1),
        /**
         * The short data format.
         */
        SIGNED_SHORT((short) 8, 2),
        /**
         * The long data format.
         */
        SIGNED_LONG((short) 9, 4),
        /**
         * The float data format.
         */
        SIGNED_RATIONAL((short) 10, 8),
        /**
         * The float data format.
         */
        SINGLE_FLOAT((short) 11, 4),
        /**
         * The double data format.
         */
        DOUBLE_FLOAT((short) 12, 8);

        /**
         * The id of the data format.
         */
        private final short id;

        /**
         * The number of bytes per component for this data format.
         */
        private final int bytesPerComponent;

        /**
         * Construct a new data format having the specified id and bytesPerComponent.
         *
         * @param theId The id of the data format.
         * @param theBytesPerComponent The number of bytes per component for this data format.
         */
        DataFormat(final short theId, final int theBytesPerComponent) {
            this.id = theId;
            this.bytesPerComponent = theBytesPerComponent;
        }

        /**
         * Returns the id of the data format.
         *
         * @return The id of the data format.
         */
        public final short getId() {
            return id;
        }

        /**
         * Returns the number of bytes per component for this data format.
         *
         * @return The number of bytes per component for this data format.
         */
        public final int getBytesPerComponent() {
            return bytesPerComponent;
        }

        /**
         * Search and returns the data format having the specified id if any, returns null otherwise.
         *
         * @param anId The id of the data format to search.
         * @return The data format having the specified id if any, returns null otherwise.
         */
        private static DataFormat getDataFormatById(final short anId) {
            for (final DataFormat dataFormat : DataFormat.values()) {
                if (dataFormat.getId() == anId) {
                    return dataFormat;
                }
            }

            return null;
        }
    }

    /**
     * An integer used to "unsign" a byte.
     */
    private static final int UNSIGN = 0xff;

    /**
     * The awaited value for the Start Of Image marker.
     */
    private static final int SOI_MARKER = 0xFFD8;

    /**
     * The awaited value for the APP1 marker.
     */
    private static final int APP1_MARKER = 0xFFE1;

    /**
     * The awaited value for the Exif header.
     */
    private static final long EXIF_HEADER = 0x45786966_0000L;

    /**
     * The awaited value for the little endian TIFF header.
     */
    private static final long TIFF_HEADER_LITTLE_ENDIAN = 0x49492A00_08000000L;

    /**
     * The awaited value for the big endian TIFF header.
     */
    private static final long TIFF_HEADER_BIG_ENDIAN = 0x4D4D002A_00000008L;

    /**
     * The awaited value for the little endian tag mark.
     */
    private static final int TAG_MARK_LITTLE_ENDIAN = 0x2A00;

    /**
     * The awaited value for the big endian tag mark.
     */
    private static final int TAG_MARK_BIG_ENDIAN = 0x002A;

    /**
     * The file to read.
     */
    private final File file;

    /**
     * The list of input file directories.
     */
    private List<List<IFDEntry>> ifds;

    /**
     * The image thumbnail (if any, null otherwise).
     */
    private byte[] thumbnail;

    /**
     * Construct a new Exif reader for the specified file.
     *
     * @param theFile The file to read.
     * @throws IOException If an I/O error occurs or if the consistency of the file is not validated.
     */
    public ExifReader(final File theFile) throws IOException {
        super();
        this.file = theFile;

        read();
    }

    /**
     * Returns the image thumbnail (if any, null otherwise).
     *
     * @return The image thumbnail (if any, null otherwise).
     */
    public byte[] getThumbnail() {
        return thumbnail;
    }

    /**
     * Read the file.
     *
     * @throws IOException If an I/O error occurs or if the consistency of the file is not validated.
     */
    private void read() throws IOException {
        try (InputStream in = new FileInputStream(file)) {
            final byte[] app1DataArray = checkConsistency(in);
            ifds = readIFDs(app1DataArray);

            extractThumbnail(app1DataArray);
        }
    }

    /**
     * Read all the input file directories in the app1data (given as an array).
     *
     * @param app1DataArray The array containing all the app1data (without its size).
     * @return The list of input file directories.
     */
    private List<List<IFDEntry>> readIFDs(final byte[] app1DataArray) {
        final int offsetToNextIFDOffset = 4;
        final int offsetToNextIFDLength = 4;
        final int numberOfEntriesLength = 2;
        final int innerValueOffset = 8;
        final int innerValueSize = 4;
        final int ifdEntryLength = 12;

        final List<List<IFDEntry>> theIfds = new ArrayList<>();

        int offsetToNextIFD = toInt(Arrays.
                copyOfRange(app1DataArray, offsetToNextIFDOffset, offsetToNextIFDOffset + offsetToNextIFDLength));

        while (offsetToNextIFD != 0) {
            // Read the next IFD
            final List<IFDEntry> ifd = new ArrayList<>();
            theIfds.add(ifd);

            int offset = offsetToNextIFD;

            final int numberOfEntries = toInt(Arrays.copyOfRange(app1DataArray, offset, offset + numberOfEntriesLength));
            offset += 2;

            for (int i = 0; i < numberOfEntries; i++) {
                // Read the next IFD entry
                final Tag tagNumber = Tag.getTagById(toShort(Arrays.copyOfRange(app1DataArray, offset, offset + 2)));
                final DataFormat dataFormat = DataFormat.getDataFormatById(
                        toShort(Arrays.copyOfRange(app1DataArray, offset + 2, offset + 4)));

                final int numberOfComponents = toInt(Arrays.copyOfRange(app1DataArray, offset + 4, offset + 8));
                final int valueSize = numberOfComponents * dataFormat.getBytesPerComponent();
                final byte[] rawValue;
                if (valueSize > innerValueSize) {
                    // the "tag value" is the offset of the real value
                    final int valueOffset = toInt(Arrays.copyOfRange(app1DataArray, offset + 8, offset + 12));
                    rawValue = Arrays.copyOfRange(app1DataArray, valueOffset, valueOffset + valueSize);
                } else {
                    rawValue = Arrays.copyOfRange(app1DataArray,
                                                  offset + innerValueOffset,
                                                  offset + innerValueOffset + innerValueSize);

                }

                final Object value = evaluateTag(rawValue, dataFormat, numberOfComponents);
                ifd.add(new IFDEntry(tagNumber, dataFormat, value));

                offset += ifdEntryLength;
            }

            offsetToNextIFD = toInt(Arrays.copyOfRange(app1DataArray, offset, offset + offsetToNextIFDLength));
        }

        return theIfds;
    }

    /**
     * Check the consistency of the specified Exif input stream.
     *
     * @param in The input stream for which the consistency shall be checked.
     * @return The data byte buffer for the Exif input stream.
     * @throws IOException If an I/O error occurs or if the consistency is not validated.
     */
    private byte[] checkConsistency(final InputStream in) throws IOException {
        final int exifHeaderLength = 6;
        final int tiffHeaderLength = 8;
        final String littleEndian = "II";
        final String bigEndian = "MM";

        checkMarkers(in);

        final byte[] app1DataSizeArray = new byte[2];
        in.read(app1DataSizeArray);

        final int app1DataSize = toInt(app1DataSizeArray);

        final byte[] exifHeaderArray = new byte[exifHeaderLength];
        in.read(exifHeaderArray);

        if (toLong(exifHeaderArray) != EXIF_HEADER) {
            throw new IOException("Exif Header not correct: 0x"
                    + Long.toHexString(toLong(exifHeaderArray)).toUpperCase()
                    + " (should be 0x457869660000)");
        }

        // We already have read the first eight bytes of the data to get the size and the exif header
        final byte[] app1DataArray = new byte[app1DataSize - app1DataSizeArray.length - exifHeaderArray.length];
        in.read(app1DataArray);

        final byte[] tiffHeaderArray = Arrays.copyOfRange(app1DataArray, 0, tiffHeaderLength);

        final String byteAlign = new String(tiffHeaderArray, 0, 2);
        final short tagMark = (short) toInt(Arrays.copyOfRange(tiffHeaderArray, 2, 4));

        if (byteAlign.equals(littleEndian)) {
            if (toLong(tiffHeaderArray) != TIFF_HEADER_LITTLE_ENDIAN) {
                throw new IOException("TIFF Header not correct: 0x"
                        + Long.toHexString(toLong(tiffHeaderArray)).toUpperCase()
                        + " (should be 0x49492A0008000000)");
            }

            // byteOrder = ByteOrder.LITTLE_ENDIAN;
        } else if (byteAlign.equals(bigEndian)) {
            if (toLong(tiffHeaderArray) != TIFF_HEADER_BIG_ENDIAN) {
                throw new IOException("TIFF Header not correct: 0x"
                        + Long.toHexString(toLong(tiffHeaderArray)).toUpperCase()
                        + " (should be 0x7777002A08000000)");
            }

            // byteOrder = ByteOrder.BIG_ENDIAN;
        }

        if ((byteAlign.equals(littleEndian) && tagMark != TAG_MARK_LITTLE_ENDIAN)) {
            throw new IOException("TIFF Header not correct: 0x"
                    + Integer.toHexString(tagMark).toUpperCase()
                    + " (should be 0x2A00)");
        } else if ((byteAlign.equals(bigEndian) && tagMark != TAG_MARK_BIG_ENDIAN)) {
            throw new IOException("TIFF Header not correct: 0x"
                    + Integer.toHexString(tagMark).toUpperCase()
                    + " (should be 0x002A)");
        }

        return app1DataArray;
    }

    /**
     * Check the markers of the specified Exif input stream.
     *
     * @param in The input stream for which the markers shall be checked.
     * @throws IOException If an I/O error occurs or if the markers are not validated.
     */
    private void checkMarkers(final InputStream in) throws IOException {
        final byte[] soiMarkerArray = new byte[2];
        in.read(soiMarkerArray);

        if (toInt(soiMarkerArray) != SOI_MARKER) {
            throw new IOException("SOI Marker not correct: 0x"
                    + Integer.toHexString(toInt(soiMarkerArray)).toUpperCase()
                    + " (should be 0xFFD8)");
        }

        final byte[] app1MarkerArray = new byte[2];
        in.read(app1MarkerArray);

        if (toInt(app1MarkerArray) != APP1_MARKER) {
            throw new IOException("APP1 Marker not correct: 0x"
                    + Integer.toHexString(toInt(app1MarkerArray)).toUpperCase()
                    + " (should be 0xFFE1)");
        }
    }

    /**
     * Returns the evaluated value of the given byte array for the specified data format and number of components.
     *
     * @param data The raw data to evaluate.
     * @param dataFormat The data format of the value.
     * @param numberOfComponents The number of components the value is composed of.
     * @return The evaluated value of the given byte array.
     */
    private Object evaluateTag(final byte[] data, final DataFormat dataFormat, final int numberOfComponents) {
        final Object[] value = new Object[numberOfComponents];

        for (int component = 0; component < numberOfComponents; component++) {
            final byte[] componentData = Arrays.copyOfRange(data,
                    component * dataFormat.getBytesPerComponent(),
                    (component + 1) * dataFormat.getBytesPerComponent());

            value[component] = evaluateTagComponent(dataFormat, componentData);
        }

        if (dataFormat == DataFormat.ASCII_STRINGS) {
            final byte[] bytes = new byte[value.length];
            for (int i = 0; i < value.length; i++) {
                bytes[i] = (byte) value[i];
            }
            return new String(bytes);
        } else if (numberOfComponents == 1) {
            return value[0];
        } else {
            return value;
        }
    }

    /**
     * Evaluates (and returns) a single component object given as a data format and raw data.
     *
     * @param dataFormat The data format of the component to evaluate.
     * @param componentData The raw data of the component to evaluate.
     * @return The value of a single component.
     */
    //CHECKSTYLE:OFF: CyclomaticComplexity
    private Object evaluateTagComponent(final DataFormat dataFormat, final byte[] componentData) {
        final int integerSise = 4;

        final Object value;

        switch (dataFormat) {
            case UNSIGNED_BYTE:
            case SIGNED_SHORT:
                value = toShort(componentData);
                break;
            case ASCII_STRINGS:
            case SIGNED_BYTE:
            case UNDEFINED:
                value = componentData[0];
                break;
            case UNSIGNED_SHORT:
                value = toInt(componentData);
                break;
            case UNSIGNED_LONG:
            case SIGNED_LONG:
                value = toLong(componentData);
                break;
            case UNSIGNED_RATIONAL:
                value =
                    (double) toLong(Arrays.copyOfRange(componentData, 0, integerSise))
                    / (double) toLong(Arrays.copyOfRange(componentData, integerSise, integerSise * 2));
                break;
            case SIGNED_RATIONAL:
                value =
                    (double) toInt(Arrays.copyOfRange(componentData, 0, integerSise))
                    / (double) toInt(Arrays.copyOfRange(componentData, integerSise, integerSise * 2));
                break;
            case SINGLE_FLOAT:
                value = Float.intBitsToFloat(toInt(componentData));
                break;
            case DOUBLE_FLOAT:
                value = Double.longBitsToDouble(toLong(componentData));
                break;
            default:
                value = null;
        }

        return value;
    }
    //CHECKSTYLE:ON: CyclomaticComplexity

    /**
     * Extracts the thumbnail if any and store it in the thumbnail attribute.
     *
     * @param app1DataArray The buffer containing the TIFF data and from which the thumbnail shall be extracted.
     * @throws IOException If an I/O error occurs.
     */
    private void extractThumbnail(final byte[] app1DataArray) throws IOException {
        final short offsetTag = 0x0201;
        final short sizeTag = 0x0202;

        // Extract the thumbnail
        // Thumbnail information are on the IFD1
        Optional<IFDEntry> optionalOffset = null;
        Optional<IFDEntry> optionalSize = null;
        for (int i = 0; i < ifds.size(); i++) {
            optionalOffset = ifds.get(i).stream().filter(entry -> entry.getTag().getId() == offsetTag).findFirst();
            optionalSize = ifds.get(i).stream().filter(entry -> entry.getTag().getId() == sizeTag).findFirst();
        }

        if (optionalOffset.isPresent() && optionalSize.isPresent()) {
            thumbnail = Arrays.copyOfRange(app1DataArray,
                    ((Long) optionalOffset.get().getValue()).intValue(),
                    ((Long) optionalOffset.get().getValue()).intValue() + ((Long) optionalSize.get().getValue()).intValue());
        }
    }

    /**
     * Convert a byte array to a short value.
     *
     * @param array The array to convert.
     * @return The result of the array conversion to a short value.
     */
    private short toShort(final byte[] array) {
        short integer = 0;
        for (int i = 0; i < array.length; i++) {
            final byte b = array[i];

            integer <<= Byte.SIZE;
            integer |= b & UNSIGN;
        }

        return integer;
    }

    /**
     * Convert a byte array to an int value.
     *
     * @param array The array to convert.
     * @return The result of the array conversion to an int value.
     */
    private int toInt(final byte[] array) {
        int integer = 0;
        for (int i = 0; i < array.length; i++) {
            final byte b = array[i];

            integer <<= Byte.SIZE;
            integer |= b & UNSIGN;
        }

        return integer;
    }

    /**
     * Convert a byte array to a long value.
     *
     * @param array The array to convert.
     * @return The result of the array conversion to a long value.
     */
    private long toLong(final byte[] array) {
        long integer = 0;
        for (int i = 0; i < array.length; i++) {
            final byte b = array[i];

            integer <<= Byte.SIZE;
            integer |= b & UNSIGN;
        }

        return integer;
    }

    @Override
    public final String toString() {
        String s = "ExifReader [file=" + file + "]";

        for (int i = 0; i < ifds.size(); i++) {
            s += "\nIFD" + i + ":";
            for (final IFDEntry entry : ifds.get(i)) {
                s += "\n\t" + entry;
            }
        }

        return s;
    }

    /**
     * The modelling class of an input file directory entry.
     *
     * @author Zleurtor
     */
    private static class IFDEntry {

        /**
         * The tag of this IFD entry.
         */
        private final Tag tag;

        /**
         * The data format of this IFD entry.
         */
        private final DataFormat dataFormat;

        /**
         * The value of this IFD entry.
         */
        private final Object value;

        /**
         * Constructs a new IFD entry with the specified tag, data format, number of components and value.
         *
         * @param theTag The tag of this IFD entry.
         * @param theDataFormat The data format of this IFD entry.
         * @param theValue The value of this IFD entry.
         */
        IFDEntry(final Tag theTag, final DataFormat theDataFormat, final Object theValue) {
            super();
            this.tag = theTag;
            this.dataFormat = theDataFormat;
            this.value = theValue;
        }

        /**
         * Returns the tag of this IFD entry.
         *
         * @return The tag of this IFD entry.
         */
        public Tag getTag() {
            return tag;
        }

        /**
         * Returns the data format of this IFD entry.
         *
         * @return The data format of this IFD entry.
         */
        public DataFormat getDataFormat() {
            return dataFormat;
        }

        /**
         * Returns the value of this IFD entry.
         *
         * @return The value of this IFD entry.
         */
        public Object getValue() {
            return value;
        }

        @Override
        public String toString() {
            return tag.getName() + "[" + getDataFormat().name() + "]: " + value;
        }

    }

}
