/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.URLDecoder;
import java.nio.file.Files;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.SortedMap;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.imageio.ImageIO;

import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.lf5.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.events.MasterEvent;
import zleurtor.scanmanager.model.events.MasterListener;
import zleurtor.scanmanager.model.messages.AbstractMessage;
import zleurtor.scanmanager.model.messages.CloseConnectionMessage;
import zleurtor.scanmanager.model.messages.HeartBeatMessage;
import zleurtor.scanmanager.model.messages.IAmMessage;
import zleurtor.scanmanager.model.messages.LogLevelErrorMessage;
import zleurtor.scanmanager.model.messages.LogLevelMessage;
import zleurtor.scanmanager.model.messages.LogLevelSuccessMessage;
import zleurtor.scanmanager.model.messages.LogMessage;
import zleurtor.scanmanager.model.messages.OptionsMessage;
import zleurtor.scanmanager.model.messages.ShutdownMessage;
import zleurtor.scanmanager.model.messages.SnapshotErrorMessage;
import zleurtor.scanmanager.model.messages.SnapshotMessage;
import zleurtor.scanmanager.model.messages.SnapshotSuccessMessage;
import zleurtor.scanmanager.model.messages.StreamErrorMessage;
import zleurtor.scanmanager.model.messages.StreamMessage;
import zleurtor.scanmanager.model.messages.UpdateMessage;
import zleurtor.scanmanager.model.messages.WhoAreYouMessage;

/**
 * The master's class having in charge to send messages to the slaves and to receive (and treat) the messages from the slaves.
 *
 * @author Zleurtor
 */
public class Master extends AbstractCommunicator {

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(Master.class);

    /**
     * The application's configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    /**
     * The heartbeat period (in milliseconds).
     */
    private static final int HEARTBEAT_PERIOD = CONFIGURATION.getInt("heartbeat.period") * 1_000;

    /**
     * The folder containing the snapshots.
     */
    private static final File SNAPSHOTS_DIRECTORY = new File(CONFIGURATION.getString("snapshot.masterFolder"));

    /**
     * The TCP server socket to which the slaves can connect to to identify themselves.
     */
    private List<ServerSocket> serverSockets;

    /**
     * The TCP server socket to which the slaves can connect to to send heartbeats.
     */
    private ServerSocket heartbeatServerSocket;

    /**
     * The slaves that have identified to the master.
     */
    private final Map<String, Object[]> slaves = new HashMap<>();

    /**
     * The last received heartbeat times.
     */
    private final Map<String, Long> lastHeartbeats = new HashMap<>();

    /**
     * All the objects listening for master events.
     */
    private final Set<MasterListener> listeners = new HashSet<>();

    /**
     * The broadcast address for each available network address.
     */
    private final Map<InetAddress, InetAddress> broadcastAddresses;

    /**
     * Constructs a new master object.
     * @throws SocketException If an error occurs while searching the broadcast addresses.
     */
    public Master() throws SocketException {
        broadcastAddresses = searchBroadcastAddresses();

        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            send(new CloseConnectionMessage());
        }));

        final int delay = 1_000;
        final int period = 1_000;
        final Timer timer = new Timer("Garbage collector");
        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                System.gc();
            }
        }, delay, period);

        final Thread heartbeat = new Thread(() -> {
            final Timer heartbeatTimer = new Timer();
            heartbeatTimer.schedule(new TimerTask() {
                @Override
                public void run() {
                    checkHeartbeats();
                }
            }, HEARTBEAT_PERIOD, HEARTBEAT_PERIOD);

            acceptHeartbeat();
        }, "Heartbeat accepter");
        heartbeat.start();
    }

    /**
     * The method used to "scan" the network.
     */
    public void scan() {
        try {
            // Close previous connections
            send(new CloseConnectionMessage(true));
            slaves.clear();
            fireSlavesClearedEvent();

            // Open the ServerSockets
            if (serverSockets == null) {
                serverSockets = new ArrayList<>();

                for (final InetAddress address : broadcastAddresses.keySet()) {
                    serverSockets.add(new ServerSocket(
                            CONFIGURATION.getInt("discover.port"),
                            CONFIGURATION.getInt("discover.backlog"),
                            address));
                }
            }
            accept();
        } catch (final IOException e) {
            LOGGER.error(e.getMessage(), e);
            return;
        }

        final int period = 1_000; // 1 second
        final Timer scanTimer = new Timer();
        scanTimer.scheduleAtFixedRate(new TimerTask() {
            @Override
            public void run() {
                try {
                    final String content = AbstractMessage.DISCOVER;

                    for (final InetAddress address : broadcastAddresses.keySet()) {
                        try (DatagramSocket socket = new DatagramSocket(CONFIGURATION.getInt("discover.port"), address)) {
                            final DatagramPacket packet =
                                    new DatagramPacket(
                                            content.getBytes(),
                                            content.length(),
                                            broadcastAddresses.get(address),
                                            CONFIGURATION.getInt("discover.port"));
                            try {
                                socket.send(packet);
                            } catch (final SocketException e) {
                                // Do nothing, inconsistency between bound address and destination address
                                //LOGGER.debug(e.getMessage(), e);
                            } catch (final IOException e) {
                                LOGGER.warn(e.getMessage(), e);
                            }
                        }
                    }
                } catch (final SocketException e) {
                    LOGGER.warn(e.getMessage(), e);
                }

                // Send a discover message
            }
        }, 0, period);
        LOGGER.info("Discover message sender ready.");

    }

    /**
     * Search the network address for each network interface then return them.
     *
     * @return All the broadcast addresses.
     * @throws SocketException If an I/O error occurs.
     */
    private Map<InetAddress, InetAddress> searchBroadcastAddresses() throws SocketException {
        final Map<InetAddress, InetAddress> addresses = new HashMap<>();

        // Search broadcast addresses
        LOGGER.debug("Broadcast address search started...");
        final Enumeration<NetworkInterface> networkInterfaces = NetworkInterface.getNetworkInterfaces();
        while (networkInterfaces.hasMoreElements()) {
            final NetworkInterface networkInterface = networkInterfaces.nextElement();
            final Iterator<InterfaceAddress> addressesIterator = networkInterface.getInterfaceAddresses().iterator();

            while (addressesIterator.hasNext()) {
                final InterfaceAddress address = addressesIterator.next();
                if (address.getBroadcast() != null) {
                    LOGGER.debug("Found broadcast address for interface " + networkInterface.getDisplayName()
                        + " -> " + address.getBroadcast());
                    addresses.put(address.getAddress(), address.getBroadcast());
                }
            }
        }
        LOGGER.debug("Broadcast address search finished.");

        return addresses;
    }

    /**
     * The method sending snapshot requests.
     */
    public final void takeSnapshot() {
        // Stop the previews (if any)
        try {
            final Process p = Utils.execute("TASKKILL /T /F /IM mplayer.exe /FI \"STATUS eq RUNNING\"", false);
            p.waitFor();
            p.destroyForcibly();
        } catch (final IOException | InterruptedException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        // Clear the thumbnails
        fireSnapshotsClearedEvent();

        // Create the pictures folder
        final DateFormat format = new SimpleDateFormat("yyyyMMdd_HHmmss");
        final String subFolder = format.format(new Date());

        final SortedMap<String, String> sortedSlaves = new TreeMap<>(Utils.getIpComparator());
        for (final String key : slaves.keySet()) {
            sortedSlaves.put(Utils.extractIp(slaves.get(key)[0].toString()), key);
        }

        final AbstractMessage message = new SnapshotMessage(subFolder,
                Configuration.getInstance().getInt("snapshot.thumbnailWidth") + ":"
                + Configuration.getInstance().getInt("snapshot.thumbnailHeight") + ":"
                + Configuration.getInstance().getInt("snapshot.thumbnailQuality"),
                Boolean.parseBoolean(Configuration.getInstance().getString("polychrome")));
        send(message);

        LOGGER.info("Snapshot message sent.");
    }

    /**
     * The method used to save snapshot files archives.
     *
     * @param inFiles The files to save into a zip archive.
     * @param optionsFile The files containing the snapshot options of each slave.
     * @param outFile The zip archive into which the files shall be saved.
     * @throws IOException If an Input/Output error occurs.
     */
    public final void save(final Map<File, int[]> inFiles, final File optionsFile, final File outFile) throws IOException {
        try (FileOutputStream out = new FileOutputStream(outFile);
                ZipOutputStream outZip = new ZipOutputStream(out)) {

            String parent = "";

            final Map<File, String> colorBalancedFiles = new LinkedHashMap<>();
            for (final Entry<File, int[]> inFile : inFiles.entrySet()) {
                parent = inFile.getKey().getParentFile().getName();

                final int[] colorBalance = inFile.getValue();
                final File colorBalancedFile = File.createTempFile("3d-scan-manager", ".tmp");
                final BufferedImage colorBalancedImage = ImageIO.read(inFile.getKey());

                final int maxColorComponent = 255;
                final int percent = 100;
                for (int x = 0; x < colorBalancedImage.getWidth(); x++) {
                    for (int y = 0; y < colorBalancedImage.getHeight(); y++) {
                        final Color c = new Color(colorBalancedImage.getRGB(x, y));
                        colorBalancedImage.setRGB(x, y,
                                new Color(
                                        Math.min(maxColorComponent, (int) Math.round((c.getRed() / (double) percent) * colorBalance[0])),
                                        Math.min(maxColorComponent, (int) Math.round((c.getGreen() / (double) percent) * colorBalance[1])),
                                        Math.min(maxColorComponent, (int) Math.round((c.getBlue() / (double) percent) * colorBalance[2]))
                                        ).getRGB());
                    }
                }

                ImageIO.write(colorBalancedImage,
                        inFile.getKey().getName().substring(inFile.getKey().getName().lastIndexOf('.') + 1),
                        colorBalancedFile);

                colorBalancedFiles.put(colorBalancedFile, parent + "/" + inFile.getKey().getName());
            }

            int index = 0;
            for (final Entry<File, String> inFile : colorBalancedFiles.entrySet()) {
                outZip.putNextEntry(new ZipEntry(inFile.getValue()));

                try (FileInputStream in = new FileInputStream(inFile.getKey())) {
                    final int bufferSize = 1024;
                    final byte[] buffer = new byte[bufferSize];
                    for (int read = in.read(buffer); read != -1; read = in.read(buffer)) {
                        outZip.write(buffer, 0, read);
                    }
                }

                outZip.closeEntry();

                //Files.delete(inFile.getKey().toPath());

                fireFileSavedEvent(new MasterEvent(this, "Master", null, new int[] { ++index, inFiles.size() }));
            }

            if (inFiles.isEmpty()) {
                fireFileSavedEvent(new MasterEvent(this, "Master", null, new int[] { 0, 0 }));
            }

            outZip.putNextEntry(new ZipEntry(parent + "/options.txt"));

            try (FileInputStream in = new FileInputStream(optionsFile)) {
                final int bufferSize = 1024;
                final byte[] buffer = new byte[bufferSize];
                for (int read = in.read(buffer); read != -1; read = in.read(buffer)) {
                    outZip.write(buffer, 0, read);
                }
            }

            outZip.closeEntry();
        }
    }

    /**
     * Returns the list of snapshot set directories.
     *
     * @return The list of snapshot set directories.
     */
    public final File[] getSnapshotSets() {
        return SNAPSHOTS_DIRECTORY.listFiles(file -> file.isDirectory());
    }

    /**
     * The method used to clear the snapshot files.
     *
     * @param snapshotSets The sets of snapshots to delete.
     */
    public final void trash(final File[] snapshotSets) {
        try {
            for (final File snapshotSet : snapshotSets) {
                Utils.recursiveDelete(snapshotSet);
                Files.delete(snapshotSet.toPath());
            }
            fireSnapshotsClearedEvent();
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * The method used to send a shutdown request to the slaves.
     *
     * @param reboot Whether or not the sent shutdown request shall be a reboot one.
     */
    public final void shutdown(final boolean reboot) {
        final AbstractMessage message = new ShutdownMessage(reboot);
        send(message);

        LOGGER.info("Shutdown message sent.");
    }

    /**
     * The method used to send a shutdown request to one slave.
     *
     * @param slaveName The name of the slave to which the shutdown request shall be sent
     * @param reboot Whether or not the sent shutdown request shall be a reboot one.
     */
    public final void shutdown(final String slaveName, final boolean reboot) {
        final AbstractMessage message = new ShutdownMessage(reboot);

        try {
            writeMessage((OutputStream) slaves.get(slaveName)[2], message);
            LOGGER.info("Shutdown message sent.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * The method used to send a preview request to the slaves.
     *
     * @param slaveName The name of the slave to which the preview request shall be send.
     * @param slaveIp The IP of the slave to which the preview request shall be send.
     */
    public final void stream(final String slaveName, final String slaveIp) {
        try {
            // Create the command
            String command = CONFIGURATION.getString("stream.masterCommand");
            command = MessageFormat.format(command, System.getProperty("user.dir"), slaveIp);

            // Put the command in a shell file
            final File shell = File.createTempFile("stream-", ".bat");
            final PrintWriter pw = new PrintWriter(shell);
            pw.println(command);
            pw.close();

            // Create the process to execute the command
            LOGGER.info("Executing " + shell);
            final Process p = Utils.execute(shell.getAbsolutePath(), false);

            // Execute the command
            final Thread t = new Thread(() -> {
                try {
                    p.waitFor();
                    Files.delete(shell.toPath());
                } catch (final IOException | InterruptedException e) {
                    LOGGER.warn(e.getMessage(), e);
                }
            });
            t.setName("Master stream Thread");
            t.start();

            writeMessage((OutputStream) slaves.get(slaveName)[2], new StreamMessage());
            LOGGER.info("Stream message sent.");
        } catch (final IOException | UnsupportedOperationException e) {
            LOGGER.warn(e.getMessage(), e);
        }

    }

    /**
     * The method used to send a log request to a slave.
     *
     * @param slaveName The slave to which the log request shall be sent.
     */
    public final void log(final String slaveName) {
        final AbstractMessage message = new LogMessage();

        try {
            writeMessage((OutputStream) slaves.get(slaveName)[2], message);
            LOGGER.info("Log message sent.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * The method used to send a log level change request to a slave.
     *
     * @param level The level to which the slave's log level shall be changed.
     * @param slaveName The slave to which the log request shall be sent.
     */
    public final void logLevel(final LogLevel level, final String slaveName) {
        if (slaveName != null) {
            final AbstractMessage message = new LogLevelMessage(level);

            try {
                writeMessage((OutputStream) slaves.get(slaveName)[2], message);
                LOGGER.info("Log level message sent.");
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
            }
        } else {
            try {
                final File log4j = new File(URLDecoder.decode(getClass().getResource("/log4j.properties").getFile(), "UTF-8"));
                final Properties prop = new OrderedProperties();
                try (FileInputStream in = new FileInputStream(log4j)) {
                    prop.load(in);
                }

                final String rootLogger = prop.getProperty("log4j.rootLogger");
                prop.setProperty("log4j.rootLogger", level + rootLogger.substring(rootLogger.indexOf(',')));
                prop.store(new FileOutputStream(log4j), null);

                PropertyConfigurator.configure(log4j.getAbsolutePath());

                fireLogLevelSuccessEvent(new MasterEvent(this, "Master", null, level));
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);

                fireLogLevelErrorEvent(new MasterEvent(this, "Master", null, e.getLocalizedMessage()));
            }

        }
    }

    /**
     * The method used to send a set options request to a slave.
     *
     * @param options The new options for the slave.
     * @param slaveName The slave to which the options request shall be sent.
     */
    public final void setOptions(final String options, final String slaveName) {
        final AbstractMessage message = new OptionsMessage(options);

        try {
            writeMessage((OutputStream) slaves.get(slaveName)[2], message);
            LOGGER.info("Options message sent.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * The method used to send an update request to a slave.
     *
     * @param updateFile The file containing the update for the slave.
     * @param prefix The update file prefix.
     * @param suffix The update file suffix.
     * @param slaveName The slave to which the update request shall be sent.
     */
    public final void updateSlave(final File updateFile, final String prefix, final String suffix, final String slaveName) {
        try {
            final List<UpdateMessage> messages = UpdateMessage.newMessages(updateFile, prefix, suffix);

            for (final AbstractMessage message : messages) {
                writeMessage((OutputStream) slaves.get(slaveName)[2], message);
            }
            LOGGER.info("Update message sent.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * The method accepting heartbeat connections.
     */
    private void acceptHeartbeat() {
        try {
            heartbeatServerSocket = new ServerSocket(0);
            while (true) {
                final Socket heartbeatSocket = heartbeatServerSocket.accept();
                final Thread heartbeatReader = new Thread(() -> {
                    try {
                        readHeartbeat(heartbeatSocket);
                    } catch (final IOException e) {
                        LOGGER.warn("Cannot read heartbeats: " + e.getMessage(), e);
                    }
                });
                heartbeatReader.setName("Heartbeat Reader");
                //heartbeatReader.start();
            }
        } catch (final IOException e) {
            LOGGER.warn("Cannot receive heartbeats: " + e.getMessage(), e);
        } finally {
            try {
                heartbeatServerSocket.close();
            } catch (final IOException e) {
                LOGGER.warn("Cannot close heartbeats accepter: " + e.getMessage(), e);
            }
        }
    }

    /**
     * The method reading heartbeat messages.
     *
     * @param heartbeatSocket The socket from which the heartbeat message shall be read.
     * @throws IOException If an IO error occurs.
     */
    private void readHeartbeat(final Socket heartbeatSocket) throws IOException {
        try (InputStream in = heartbeatSocket.getInputStream();
                OutputStream out = heartbeatSocket.getOutputStream()) {
            boolean first = true;

            while (true) {
                final AbstractMessage message = readMessage(in);
                final String slaveName = ((HeartBeatMessage) message).getName();

                if (first) {
                    LOGGER.info("Heart beat active for {}.", slaveName);

                    first = false;
                }

                if (message instanceof HeartBeatMessage) {
                    heartBeat(message, heartbeatSocket, in, out, slaveName);
                }
            }
        }
    }

    /**
     * Accepts incoming TCP connections.
     */
    private void accept() {
        if (serverSockets != null) {
            for (final ServerSocket serverSocket : serverSockets) {
                final Thread t = new Thread(() -> {
                    LOGGER.info("Waiting incoming connections on " + serverSocket.getLocalSocketAddress() + "...");

                    while (true) {
                        try {
                            final Socket slaveSocket = serverSocket.accept();

                            final OutputStream out = slaveSocket.getOutputStream();
                            final InputStream in = slaveSocket.getInputStream();

                            read(slaveSocket, in, out);
                        } catch (final SocketException e) {
                            // Socket closed in order to rescan
                            break;
                        } catch (final IOException e) {
                            LOGGER.warn(e.getMessage(), e);
                        }
                    }
                });
                t.setName("Accepter Thread");
                t.start();
            }
        }
    }

    /**
     * Executes the handshake between a slave and the master then return the slave's name.
     *
     * @param slaveSocket The socket of the slave with which the handshake shall be made.
     * @param in The input stream receiving slave's handshake information.
     * @param out The output stream to send master's handshake information.
     * @return The name of the slave (if it success to the handshake).
     * @throws IOException If an Input/Output error occurs.
     * @throws ClassNotFoundException If a network object cannot be unmarshalled.
     */
    private String handshake(final Socket slaveSocket, final InputStream in, final OutputStream out)
            throws IOException, ClassNotFoundException {
        writeMessage(out, new WhoAreYouMessage(heartbeatServerSocket.getLocalPort(), HEARTBEAT_PERIOD));
        final Object object = readMessage(in);

        if (object instanceof IAmMessage) {
            final String name = ((IAmMessage) object).getName();
            final String version = ((IAmMessage) object).getVersion();
            final String options = ((IAmMessage) object).getOptions();
            final String ip = Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString());
            LOGGER.info("New slave: {} [{}]", name, ip);

            if (!slaves.containsKey(name)) {
                fireNewSlaveEvent(new MasterEvent(this, ip, name, new String[]{ options, version }));
            }

            return name;
        } else {
            throw new IOException("Bad response");
        }
    }

    /**
     * Reads messages from a slave.
     *
     * @param slaveSocket The socket from which the messages shall be read.
     * @param in The input stream from which incoming messages shall be read.
     * @param out The output stream to use for the handshake.
     */
    private void read(final Socket slaveSocket, final InputStream in, final OutputStream out) {
        final Thread t = new Thread(() -> {
            String name = null;

            try {
                name = handshake(slaveSocket, in, out);
                slaves.put(name, new Object[] { slaveSocket, in, out });

                // Loop to read
                while (true) {
                    final AbstractMessage message = readMessage(in);
                    LOGGER.debug("Incoming message...");

                    if (!dispatchMessage(message, slaveSocket, in, out, name)) {
                        break;
                    }
                }
            } catch (final IOException | ClassNotFoundException e) {
                LOGGER.warn(e.getMessage(), e);
                close(name);
                fireSlaveQuit(new MasterEvent(this, Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()), null, null));
                slaves.remove(name);
            }
        });
        t.setName("Messages reader Thread");
        t.start();
    }

    /**
     * Executes the method corresponding to the received message.
     *
     * @param message The received message.
     * @param slaveSocket The socket from which the message has been received.
     * @param in The input stream from which the message has been received.
     * @param out The ouput stream to send message to the master.
     * @param slaveName The name of the slave which sent the message.
     * @return True if the read loop shall continue, false otherwise.
     */
    private boolean dispatchMessage(final AbstractMessage message, final Socket slaveSocket,
            final InputStream in, final OutputStream out, final String slaveName) {
        if (message instanceof SnapshotSuccessMessage) {
            snapshotSuccessMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()),
                                   ((SnapshotSuccessMessage) message).getSubFolder(),
                                   ((SnapshotSuccessMessage) message).getFileContent());
        } else if (message instanceof SnapshotErrorMessage) {
            snapshotErrorMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()),
                    ((SnapshotErrorMessage) message).getSubFolder(),
                    ((SnapshotErrorMessage) message).getSlaveName(),
                    ((SnapshotErrorMessage) message).getMessage());
        } else if (message instanceof StreamErrorMessage) {
            streamErrorMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()));
        } else if (message instanceof LogMessage) {
            logMessage(slaveSocket, (LogMessage) message);
        } else if (message instanceof LogLevelErrorMessage) {
            logLevelErrorMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()),
                    ((LogLevelErrorMessage) message).getReason());
        } else if (message instanceof LogLevelSuccessMessage) {
            logLevelSuccessMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()),
                    ((LogLevelSuccessMessage) message).getLevel());
        } else if (message instanceof OptionsMessage) {
            optionsMessage(Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()),
                    ((OptionsMessage) message).getOptions());
        } else if (message instanceof CloseConnectionMessage) {
            closeConnection(message, slaveSocket, in, out, slaveName);
            return false;
        }

        return true;
    }

    /**
     * Sends the specified message to all slaves.
     *
     * @param message The message to send to all the slaves.
     */
    private void send(final AbstractMessage message) {
        if (slaves.isEmpty()) {
            return;
        }

        final CyclicBarrier barrier = new CyclicBarrier(slaves.size());

        for (final Entry<String, Object[]> slave : slaves.entrySet()) {
            final Thread t = new Thread(() -> {
                try {
                    barrier.await();
                } catch (final InterruptedException | BrokenBarrierException e) {
                    LOGGER.warn(e.getMessage(), e);
                }

                try {
                    writeMessage((OutputStream) slave.getValue()[2], message);
                } catch (final IOException e) {
                    LOGGER.warn(e.getMessage(), e);
                }
            });
            t.setName("Message sender Thread");
            t.start();
        }
    }

    /**
     * Closes a slave's connections.
     *
     * @param slaveName The name slave to close.
     */
    private void close(final String slaveName) {
        if (!slaves.containsKey(slaveName)) {
            return;
        }

        final Socket socket = (Socket) slaves.get(slaveName)[0];
        final InputStream in = (InputStream) slaves.get(slaveName)[1];
        final OutputStream out = (OutputStream) slaves.get(slaveName)[2];

        close(socket, in, out);

        try {
            final Socket heartbeatSocket = (Socket) slaves.get(slaveName)[0];
            final InputStream heartbeatIn = (InputStream) slaves.get(slaveName)[1];
            final OutputStream heartbeatOut = (OutputStream) slaves.get(slaveName)[2];

            close(heartbeatSocket, heartbeatIn, heartbeatOut);
        } catch (final ArrayIndexOutOfBoundsException e) {
            // No heart beat, do nothing
        }
    }

    /**
     * Closes a slave's socket, input stream and output stream.
     *
     * @param socket The socket of the slave to close.
     * @param in The input stream of the slave to close.
     * @param out The output stream of the slave to close.
     */
    private void close(final Socket socket, final InputStream in, final OutputStream out) {
        try {
            if (socket != null) {
                socket.close();
            }
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        try {
            if (in != null) {
                in.close();
            }
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        try {
            if (out != null) {
                out.close();
            }
        } catch (final IOException e) {
            // Already closed by socket.close(). Do nothing.
        }
    }

    /**
     * Treats the "snapshot success" slave's message.
     *
     * @param ip The IP of the slave which sent the "snapshot success" message.
     * @param subFolder The subFolder where the file shall be written.
     * @param fileContent The content of the image file.
     */
    private void snapshotSuccessMessage(final String ip, final String subFolder, final byte[] fileContent) {
        LOGGER.info("Snapshot message incame.");

        final String sharedDir = Configuration.getInstance().getString("snapshot.masterFolder");
        final File file = new File(sharedDir + (sharedDir.endsWith("/") ? "" : "/") + subFolder + "/" + ip + ".jpg");
        file.getParentFile().mkdirs();

        try (OutputStream out = new FileOutputStream(file)) {
            out.write(fileContent);

            LOGGER.info("Snapshot saved to: " + file + ".");
            fireSnapshotSuccessEvent(new MasterEvent(this, ip, null, subFolder + "/" + file.getName()));
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
            fireSnapshotErrorEvent(new MasterEvent(this, ip, null, e.getMessage() + ". " + subFolder + "/" + file.getName()));
        }
    }

    /**
     * Treats the "snapshot error" slave's message.
     *
     * @param ip The IP of the slave which sent the "snapshot error" message.
     * @param subFolder The sub folder where the snapshot shall be written.
     * @param slaveName The name of the slave which encountered the error.
     * @param message The error message.
     */
    private void snapshotErrorMessage(final String ip, final String subFolder, final String slaveName, final String message) {
        LOGGER.info("Snapshot error message incame.");

        fireSnapshotErrorEvent(new MasterEvent(this, ip, null, message));
    }

    /**
     * Treats the "stream error" slave's message.
     *
     * @param ip The IP of the slave which sent the "stream error" message.
     */
    private void streamErrorMessage(final String ip) {
        LOGGER.info("Stream error message incame.");

        fireStreamErrorEvent(new MasterEvent(this, ip, null, null));
    }

    /**
     * Treats the "log" slave's message.
     *
     * @param slaveSocket The slave socket from which the "log" message incame.
     * @param message The "log" message to treat.
     */
    private void logMessage(final Socket slaveSocket, final LogMessage message) {
        LOGGER.info("Log message incame.");

        try (InputStream gzIn = new GZIPInputStream(new ByteArrayInputStream(message.getContent()));
                ByteArrayOutputStream baOut = new ByteArrayOutputStream()) {
            final int bufferSize = 1024;
            final byte[] buffer = new byte[bufferSize];
            int read;
            while ((read = gzIn.read(buffer)) != -1) {
                baOut.write(buffer, 0, read);
            }

            String slaveName = null;
            for (final Entry<String, Object[]> entry : slaves.entrySet()) {
                if (entry.getValue()[0] == slaveSocket) {
                    slaveName = entry.getKey();
                }
            }

            fireLogEvent(
                    new MasterEvent(this, Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()), slaveName, baOut.toByteArray()),
                    message.getLevel());
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Treats the "log level error" slave's message.
     *
     * @param ip The IP of the slave which sent the "log level error" message.
     * @param reason The error message.
     */
    private void logLevelErrorMessage(final String ip, final String reason) {
        LOGGER.info("Log level error message incame.");

        fireLogLevelErrorEvent(new MasterEvent(this, ip, null, reason));
    }

    /**
     * Treats the "log level success" slave's message.
     *
     * @param ip The IP of the slave which sent the "log level success" message.
     * @param logLevel The level to which the log level has been changed.
     */
    private void logLevelSuccessMessage(final String ip, final LogLevel logLevel) {
        LOGGER.info("Log level success message incame.");

        fireLogLevelSuccessEvent(new MasterEvent(this, ip, null, logLevel));
    }

    /**
     * Treats the "options" slave's message.
     *
     * @param ip The IP of the slave which sent the "options" message.
     * @param options The new slave's options.
     */
    private void optionsMessage(final String ip, final String options) {
        LOGGER.info("Options message incame.");

        fireOptionsEvent(new MasterEvent(this, ip, null, options));
    }

    /**
     * Treats the "close connection" slave's message.
     *
     * @param message The message received from the slave.
     * @param slaveSocket The socket from which the message has been received.
     * @param in The input stream from which the message has been read.
     * @param out The output stream to which the message shall be returned.
     * @param slaveName The name of the slave which closed the connection.
     */
    private void closeConnection(final AbstractMessage message, final Socket slaveSocket, final InputStream in, final OutputStream out,
            final String slaveName) {
        try {
            LOGGER.info("The slave has closed the connection.");

            writeMessage(out, message);

            close(slaveName);
            fireSlaveQuit(new MasterEvent(this, Utils.extractIp(slaveSocket.getRemoteSocketAddress().toString()), null, null));
            slaves.remove(slaveName);
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Treats the "heart beat" slave's message.
     *
     * @param message The message received from the slave.
     * @param heartbeatSocket The socket from which the message has been received.
     * @param in The input stream from which the message has been read.
     * @param out The output stream to which the message shall be returned.
     * @param slaveName The name of the slave which sent a heartbeat.
     */
    private void heartBeat(final AbstractMessage message, final Socket heartbeatSocket, final InputStream in, final OutputStream out,
            final String slaveName) {
        while (!slaves.containsKey(slaveName)) {
            // Waiting for the subscription of the slave...
        }

        slaves.put(slaveName, new Object[] {
                slaves.get(slaveName)[0], slaves.get(slaveName)[1], slaves.get(slaveName)[2], heartbeatSocket, in, out});
        lastHeartbeats.put(slaveName, System.currentTimeMillis());
    }

    /**
     * Treats the absence of "heart beat" slave's message.
     */
    private void checkHeartbeats() {
        for (final Entry<String, Long> lastHeartbeat : lastHeartbeats.entrySet()) {
            if (lastHeartbeat.getValue() - System.currentTimeMillis() > 2 * HEARTBEAT_PERIOD) {
                final String slaveName = lastHeartbeat.getKey();

                LOGGER.warn("{} heart beat has been lost.", slaveName);

                final int heartbeatSocketIndex = 3;

                close(slaveName);
                fireSlaveQuit(new MasterEvent(this,
                        Utils.extractIp(((Socket) slaves.get(slaveName)[heartbeatSocketIndex]).getRemoteSocketAddress().toString()),
                        null,
                        null));
                slaves.remove(slaveName);
            }
        }
    }

    /**
     * Adds the specified object to the list of objects listening for master events.
     *
     * @param listener The object to add to the list of objects listening for master events.
     * @return true If the listeners collection did not already contain the specified element.
     */
    public boolean addMasterListener(final MasterListener listener) {
        return listeners.add(listener);
    }

    /**
     * Fires the "slaves cleared" event.
     */
    private void fireSlavesClearedEvent() {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.slavesCleared();
            });
            t.setName("SlavesClearedEvent dispatcher thread");
            t.start();
        }
    }

    /**
     * Fires the specified "new slave" event.
     *
     * @param evt The "new slave" event to fire to all listeners.
     */
    private void fireNewSlaveEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.newSlave(evt);
            });
            t.setName("NewSlaveEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "slave quit" event.
     *
     * @param evt The "slave quit" event to fire to all listeners.
     */
    private void fireSlaveQuit(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.slaveQuit(evt);
            });
            t.setName("SlaveQuitEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "snapshot success" event.
     *
     * @param evt The "snapshot success" event to fire to all listeners.
     */
    private void fireSnapshotSuccessEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.snapshotSuccess(evt);
            });
            t.setName("SnapshotSuccessEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "snapshot error" event.
     *
     * @param evt The "snapshot error" event to fire to all listeners.
     */
    private void fireSnapshotErrorEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.snapshotError(evt);
            });
            t.setName("SnapshotErrorEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "snapshots cleared" event.
     */
    private void fireSnapshotsClearedEvent() {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.snapshotsCleared();
            });
            t.setName("SnapshotsClearedEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "stream error" event.
     *
     * @param evt The "stream error" event to fire to all listeners.
     */
    private void fireStreamErrorEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.streamError(evt);
            });
            t.setName("StreamErrorEvent dispatcher Thread");
            t.start();
        }
    }

    /**
     * Fires the specified "log" event.
     *
     * @param evt The "log" event to fire to all listeners.
     * @param level The level of the received log message.
     */
    private void fireLogEvent(final MasterEvent evt, final LogLevel level) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.log(evt, level);
            });
            t.setName("LogEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires the specified "log level error" event.
     *
     * @param evt The "log level error" event to fire to all listeners.
     */
    private void fireLogLevelErrorEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.logLevelError(evt);
            });
            t.setName("LogLevelErrorEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires the specified "log level success" event.
     *
     * @param evt The "log level success" event to fire to all listeners.
     */
    private void fireLogLevelSuccessEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.logLevelSuccess(evt);
            });
            t.setName("LogLevelSuccessEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires the specified "options" event.
     *
     * @param evt The "options" event to fire to all listeners.
     */
    private void fireOptionsEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.options(evt);
            });
            t.setName("LogLevelSuccessEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires the specified "fileSaved" event.
     *
     * @param evt The "options" event to fire to all listeners.
     */
    private void fireFileSavedEvent(final MasterEvent evt) {
        for (final MasterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.fileSaved(evt);
            });
            t.setName("LogLevelSuccessEvent dispatcher Thread.");
            t.start();
        }
    }

}
