/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * A Properties object that retains the order of inserted properties.
 *
 * @author Zleurtor
 */
public class OrderedProperties extends Properties {

    private static final long serialVersionUID = 7233152287135923119L;

    /**
     * An ordered set of property keys.
     */
    private final Set<Object> keySet = new LinkedHashSet<>();

    @Override
    public final synchronized Object put(final Object key, final Object value) {
        keySet.add(key);
        return super.put(key, value);
    }

    @Override
    public final synchronized Enumeration<Object> keys() {
        return Collections.enumeration(keySet);
    }

    @Override
    public final Set<Object> keySet() {
        return keySet;
    }

    @Override
    public final Set<Map.Entry<Object, Object>> entrySet() {
        final Set<Map.Entry<Object, Object>> entrySet = new LinkedHashSet<>();

        for (final Object key : keySet) {
            entrySet.add(new Entry(key, super.get(key)));
        }

        return entrySet;
    }

    /**
     * A simple map entry implementation.
     *
     * @author Zleurtor
     */
    private final class Entry implements Map.Entry<Object, Object> {

        /**
         * The entry key.
         */
        private final Object key;

        /**
         * The entry value.
         */
        private Object value;

        /**
         * Creates a new entry with the specified key and value.
         *
         * @param theKey The entry key.
         * @param theValue The entry value.
         */
        private Entry(final Object theKey, final Object theValue) {
            super();
            this.key = theKey;
            this.value = theValue;
        }

        @Override
        public Object getKey() {
            return key;
        }

        @Override
        public Object getValue() {
            return value;
        }

        @Override
        public Object setValue(final Object newValue) {
            final Object formerValue = this.value;
            this.value = newValue;
            return formerValue;
        }
    }
}
