/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.nio.file.Files;
import java.nio.file.attribute.PosixFilePermission;
import java.security.DigestException;
import java.security.NoSuchAlgorithmException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.zip.GZIPOutputStream;

import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.lf5.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.messages.AbstractMessage;
import zleurtor.scanmanager.model.messages.CloseConnectionMessage;
import zleurtor.scanmanager.model.messages.IAmMessage;
import zleurtor.scanmanager.model.messages.LogLevelErrorMessage;
import zleurtor.scanmanager.model.messages.LogLevelMessage;
import zleurtor.scanmanager.model.messages.LogLevelSuccessMessage;
import zleurtor.scanmanager.model.messages.LogMessage;
import zleurtor.scanmanager.model.messages.OptionsMessage;
import zleurtor.scanmanager.model.messages.ShutdownMessage;
import zleurtor.scanmanager.model.messages.SnapshotErrorMessage;
import zleurtor.scanmanager.model.messages.SnapshotMessage;
import zleurtor.scanmanager.model.messages.SnapshotSuccessMessage;
import zleurtor.scanmanager.model.messages.StreamErrorMessage;
import zleurtor.scanmanager.model.messages.StreamMessage;
import zleurtor.scanmanager.model.messages.UpdateMessage;
import zleurtor.scanmanager.model.messages.WhoAreYouMessage;
import zleurtor.scanmanager.model.update.SlaveUpdater;

/**
 * The slaves's class having in charge to receive messages from the master, treat them and return responses.
 *
 * @author Zleurtor
 */
public final class Slave extends AbstractCommunicator {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(Slave.class);

    /**
     * The application's configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    /**
     * A Set containing all the slaves which have been created since the application has started.
     */
    private static Set<Slave> slaves = new HashSet<>();

    /**
     * Whether or not this slave is shutting down (or rebooting).
     */
    private static boolean shuttingDown = false;

    /**
     * The IP address of the master.
     */
    private final InetAddress masterAddress;

    /**
     * The socket used to connect to the master.
     */
    private final Socket socket;

    /**
     * The stream from which the slave receive master's messages.
     */
    private final InputStream in;

    /**
     * The stream to which the slave write messages for the master.
     */
    private final OutputStream out;

    /**
     * The socket used to connect to the master.
     */
    private Socket heartbeatSocket;

    /**
     * The stream to which the slave write messages for the master.
     */
    private OutputStream heartbeatOut;

    /**
     * The slave's name.
     */
    private final String name;

    /**
     * The slave's version.
     */
    private final String version;

    /**
     * The slave's configurator options.
     */
    private String options = "-t 300";

    /**
     * All the update files by their id.
     */
    private Map<String, File> updateFiles = new HashMap<>();

    /**
     * Constructs a new slave, the slave shall automatically connect to the specified address.
     *
     * @param theMasterAddress The address to which the newly created slave shall connect to.
     * @throws IOException If an I/O error occurs while reading stream.
     * @throws ClassNotFoundException If a network object cannot be unmarshalled.
     * @throws AddressNotAccessibleException If the address of the socket receiving the handshake messages is not accessible.
     */
    private Slave(final InetAddress theMasterAddress) throws IOException, ClassNotFoundException, AddressNotAccessibleException {
        masterAddress = theMasterAddress;
        socket = new Socket(theMasterAddress, CONFIGURATION.getInt("discover.port"));
        in = socket.getInputStream();
        out = socket.getOutputStream();

        name = Utils.getMacAddress(socket);
        version = CONFIGURATION.getString("version");
        if (CONFIGURATION.getString("configurator") != null) {
            options = CONFIGURATION.getString("configurator");
        }
        handshake(in, out);

        read();
    }

    /**
     * Executes the handshake process with the master.
     *
     * @param is The stream from which the slave receive master's messages.
     * @param os The stream to which the slave write messages for the master.
     * @throws IOException If an Input/Output error occurs.
     * @throws ClassNotFoundException If a network object cannot be unmarshalled.
     * @throws AddressNotAccessibleException If the address of the socket receiving the handshake messages is not accessible.
     */
    private void handshake(final InputStream is, final OutputStream os)
            throws IOException, ClassNotFoundException, AddressNotAccessibleException {
        // Read the "Who Are You" message from the master
        final AbstractMessage object = readMessage(is);
        if (object instanceof WhoAreYouMessage) {
            final int heartbeatPort = ((WhoAreYouMessage) object).getHeartbeatPort();
            final int heartbeatPeriod = ((WhoAreYouMessage) object).getHeartbeatPeriod();
            startHeartbeat(heartbeatPort, heartbeatPeriod);

            // Respond to the "Who Are You" message from the master
            final AbstractMessage msg = new IAmMessage(name, version, options);
            writeMessage(os, msg);

            LOGGER.info("Mac address sent.");
        } else {
            // Error, not a good daddy
            throw new IOException("Bad master response.");
        }
    }

    /**
     * Start sending heartbeat messages.
     *
     * @param heartbeatPort The port to which connect to to start sending heartbeats.
     * @param heartbeatPeriod The period between two heartbeat messages.
     * @throws IOException If an IO error occurs.
     */
    private void startHeartbeat(final int heartbeatPort, final int heartbeatPeriod) throws IOException {
        heartbeatSocket = new Socket(masterAddress, heartbeatPort);
        heartbeatOut = heartbeatSocket.getOutputStream();

        LOGGER.debug("Heartbeat socket created.");

        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            try {
                heartbeatSocket.close();
            } catch (final IOException e) {
                LOGGER.warn("Cannot close heartbeat socket: " + e.getMessage() + ".", e);
            }
        }));

        // create the heartbeat timer
        heartbeatOut.flush();
//        final Timer timer = new Timer("Heartbeat timer");
//        timer.schedule(new TimerTask() {
//            @Override
//            public void run() {
//                try {
//                    LOGGER.debug("Sending heartbeat.");
//                    writeMessage(heartbeatOut, new HeartBeatMessage(name));
//                } catch (final IOException e) {
//                    LOGGER.warn("Cannot send heartbeat: " + e.getMessage() + ".", e);
//                    timer.cancel();
//                }
//            }
//        }, 0, heartbeatPeriod);
    }

    /**
     * Treat the "Snapshot" message and... Take a snapshot.
     *
     * @param message The message sent by the master to ask the slave to take a snapshot.
     */
    private void takeSnapshot(final SnapshotMessage message) {
        LOGGER.info("Snapshot request incame.");

        try {
            final String filePath = File.createTempFile("scanmanager-", ".jpg").getAbsolutePath();
            final String chromacityOption = message.isPolychrome() ? "" : " -cfx 128:128";
            String command = MessageFormat.format(CONFIGURATION.getString("snapshot.command"), filePath,
                    options + " -th " + message.getThumbnailInformation() + chromacityOption);
            command = command.replace("//", "/");

            LOGGER.debug("Executing command: {}", command);

            new File(filePath).getParentFile().mkdirs();

            Utils.execute(command).waitFor();

            final AbstractMessage snapshotMessage = new SnapshotSuccessMessage(filePath, message.getSubFolder());
            try {
                Files.delete(new File(filePath).toPath());
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage());
            }

            writeMessage(out, snapshotMessage);

            LOGGER.info("Snapshot sent to the master");
        } catch (final IOException | InterruptedException e) {
            LOGGER.warn(e.getMessage(), e);

            try {
                writeMessage(out, new SnapshotErrorMessage(e.getMessage(), message.getSubFolder(), name));
            } catch (final IOException ex) {
                LOGGER.warn(ex.getMessage());
            }
        }
    }

    /**
     * Treat the "Shutdown" message and ... Shutdowns the machine.
     *
     * @param reboot Whether or not the pi(s) shall be rebooted instead of shutted down.
     */
    private void shutdown(final boolean reboot) {
        if (reboot) {
            LOGGER.info("Reboot request incame.");
        } else {
            LOGGER.info("Shutdown request incame.");
        }

        try {
            writeMessage(out, new CloseConnectionMessage());

            final String command = reboot ? CONFIGURATION.getString("reboot.command") : CONFIGURATION.getString("shutdown.command");

            Utils.execute(command).waitFor();
        } catch (final IOException | InterruptedException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Treat the "Stream" message and ... Starts streaming to the server.
     *
     * @param address The address to which the slave shall stream.
     */
    private void stream(final InetAddress address) {
        LOGGER.info("Preview request incame.");

        try {
            String command = CONFIGURATION.getString("stream.slaveCommand");

            command = MessageFormat.format(command, address.getHostAddress(), filterOptions());

            final File shell = File.createTempFile("stream-", ".sh");
            final Set<PosixFilePermission> perms = new HashSet<>();
            perms.add(PosixFilePermission.OWNER_READ);
            perms.add(PosixFilePermission.OWNER_WRITE);
            perms.add(PosixFilePermission.OWNER_EXECUTE);
            perms.add(PosixFilePermission.GROUP_READ);
            perms.add(PosixFilePermission.GROUP_WRITE);
            perms.add(PosixFilePermission.GROUP_EXECUTE);
            perms.add(PosixFilePermission.OTHERS_READ);
            perms.add(PosixFilePermission.OTHERS_WRITE);
            perms.add(PosixFilePermission.OTHERS_EXECUTE);
            Files.setPosixFilePermissions(shell.toPath(), perms);
            final PrintWriter pw = new PrintWriter(shell);
            pw.println(command);
            pw.close();

            LOGGER.info("Executing " + shell);
            final Process p = Utils.execute(shell.getAbsolutePath());

            final Thread t = new Thread(() -> {
                try {
                    p.waitFor();
                    Files.delete(shell.toPath());
                } catch (final IOException | InterruptedException e) {
                    LOGGER.warn(e.getMessage(), e);
                }
            });
            t.setName("Slave stream Thread");
            t.start();

            LOGGER.info("Preview streaming...");
        } catch (final IOException | UnsupportedOperationException e) {
            LOGGER.warn(e.getMessage());

            try {
                String message = e.getMessage();
                if (message == null || message.isEmpty()) {
                    message = e.toString();
                }
                writeMessage(out, new StreamErrorMessage(e.getMessage()));
            } catch (final IOException ex) {
                LOGGER.warn(ex.getMessage());
            }
        }
    }

    /**
     * Filter the slave options by removing the unsupported ones for streaming.
     *
     * @return The slave's option filtered by removing the unsupported ones for streaming.
     */
    private String filterOptions() {
        String filteredOptions = "";

        String lastOption = "";

        final List<String> unAllowedOptions = new ArrayList<>();
        unAllowedOptions.add("-q");
        unAllowedOptions.add("-t");
        unAllowedOptions.add("-e");
        unAllowedOptions.add("-ss");

        for (final String o : options.split("\\s+")) {
            if (o.startsWith("-") && !Utils.isNumber(o)) {
                // An option
                if (!unAllowedOptions.contains(o)) {
                    filteredOptions += o + " ";
                }
                lastOption = o;
            } else {
                // A value
                if (!unAllowedOptions.contains(lastOption)) {
                    filteredOptions += o + " ";
                }
            }
        }
        return filteredOptions;
    }

    /**
     * Treat the "log" message and ... Send the logs to the master.
     */
    private void log() {
        LOGGER.info("Log request incame.");

        final LogMessage message = new LogMessage();

        try (InputStream in = new FileInputStream(Utils.getLogFile());
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                GZIPOutputStream gzOut = new GZIPOutputStream(baos)) {
            final int bufferSize = 1024;
            final byte[] buffer = new byte[bufferSize];
            int read;

            while ((read = in.read(buffer)) != -1) {
                gzOut.write(buffer, 0, read);
            }

            gzOut.finish();
            gzOut.flush();
            baos.flush();

            message.setLevel(Utils.getLogLevel());
            message.setContent(baos.toByteArray());

            writeMessage(out, message);

            LOGGER.info("Log Sent.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Treat the "log level" message.
     *
     * @param level The level to which the log level shall be changed.
     */
    private void logLevel(final LogLevel level) {
        LOGGER.info("Log level change request incame.");

        try {
            final File log4j = new File(getClass().getResource("/log4j.properties").getFile());
            final Properties prop = new OrderedProperties();
            try (FileInputStream in = new FileInputStream(log4j)) {
                prop.load(in);
            }

            final String rootLogger = prop.getProperty("log4j.rootLogger");
            prop.setProperty("log4j.rootLogger", level + rootLogger.substring(rootLogger.indexOf(',')));
            prop.store(new FileOutputStream(log4j), null);

            PropertyConfigurator.configure(log4j.getAbsolutePath());

            writeMessage(out, new LogLevelSuccessMessage(level));

            LOGGER.info("Log level set to " + level + ".");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);

            try {
                writeMessage(out, new LogLevelErrorMessage(e.getLocalizedMessage()));
            } catch (final IOException ex) {
                LOGGER.warn(ex.getMessage());
            }
        }
    }

    /**
     * Treat the "options" message.
     *
     * @param theOptions The new options to set.
     */
    private void setOptions(final String theOptions) {
        LOGGER.info("Set options request incame.");

        try {
            options = theOptions;
            CONFIGURATION.setString("configurator", options);

            writeMessage(out, new OptionsMessage(options));

            LOGGER.info("New options set.");
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Treat the "update" message.
     *
     * @param id The update id.
     * @param updateContent The content of the update file to install.
     * @param prefix The update file prefix.
     * @param suffix The update file suffix.
     * @param totalChunks The update file totalNnumber of chunks.
     * @param chunkNumber The update file current chunk number.
     */
    private synchronized void update(final String id, final byte[] updateContent, final String prefix, final String suffix,
            final int totalChunks, final int chunkNumber) {
        LOGGER.info("Update request incame ({}, chunk {} / {}).", id, chunkNumber, totalChunks);
        File file = null;

        try {
            if (!updateFiles.containsKey(id)) {
                updateFiles.put(id, File.createTempFile(prefix, suffix));
            }
            file = updateFiles.get(id);

            try (FileOutputStream out = new FileOutputStream(file, true)) {
                out.write(updateContent);
                out.flush();
            }

            if (chunkNumber == totalChunks) {
                // The update file is complete
                new SlaveUpdater().update(file);

                if (file != null && !file.delete()) {
                    file.deleteOnExit();
                }

                LOGGER.info("Rebooting...");
                shutdown(true);
            }
        } catch (final IOException | NoSuchAlgorithmException | DigestException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Reads incoming messages from the master.
     */
    private void read() {
        final Thread t = new Thread(() -> {
            try {
                // Loop for reading
                while (true) {
                    final AbstractMessage message = readMessage(in);
                    LOGGER.info("Incoming message...");

                    if (!dispatchMessage(message)) {
                        break;
                    }
                }
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
                close();
                synchronized (slaves) {
                    slaves.remove(this);
                }
            }
        });
        t.setName("Messages reader Thread");
        t.start();
    }

    /**
     * Executes the method corresponding to the received message.
     *
     * @param message The received message.
     * @return True if the read loop shall continue, false otherwise.
     * @throws IOException Any exception thrown by the underlyingOutputStream.
     */
    private boolean dispatchMessage(final AbstractMessage message) throws IOException {
        if (message instanceof SnapshotMessage) {
            takeSnapshot((SnapshotMessage) message);
        } else if (message instanceof ShutdownMessage) {
            shuttingDown = true;
            shutdown(((ShutdownMessage) message).isReboot());
        } else if (message instanceof StreamMessage) {
            stream(socket.getInetAddress());
        } else if (message instanceof LogMessage) {
            log();
        } else if (message instanceof LogLevelMessage) {
            logLevel(((LogLevelMessage) message).getLevel());
        } else if (message instanceof OptionsMessage) {
            setOptions(((OptionsMessage) message).getOptions());
        } else if (message instanceof CloseConnectionMessage) {
            LOGGER.info("The master closed the connection.");
            if (((CloseConnectionMessage) message).isResend()) {
                writeMessage(out, message);
            }
            slaves.remove(this);
            return false;
        } else if (message instanceof UpdateMessage) {
            final UpdateMessage updateMessage = (UpdateMessage) message;
            update(updateMessage.getId(), updateMessage.getFileContent(), updateMessage.getPrefix(), updateMessage.getSuffix(),
                    updateMessage.getTotalChunks(), updateMessage.getChunkNumber());
        }

        return true;
    }

    /**
     * Close the socket, and the streams of the slave.
     */
    private void close() {
        try {
            if (socket != null) {
                socket.close();
            }
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        try {
            if (in != null) {
                in.close();
            }
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        try {
            if (out != null) {
                out.close();
            }
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }
    }



    /**
     * Starts the slave application.
     *
     * @param args The arguments of the slave application (no one is used;-)
     */
    public static void main(final String[] args) {
        if (args.length > 0 && args[0].equals("stop")) {
            stop();
        } else {
            final SlaveUpdater updater = new SlaveUpdater();

            // Hook to delete old license
            final File copyingLesser = new File("COPYING.LESSER");
            if (copyingLesser.exists()) {
                try {
                    Files.delete(copyingLesser.toPath());
                } catch (final IOException e) {
                    LOGGER.warn(e.getMessage(), e);
                }
            }

            // Check if an update is available (and install it)
            if (updater.update()) {
                LOGGER.info("restarting...");
                try {
                    final Process p = Runtime.getRuntime().exec("./slave.sh");
                    p.getInputStream().read();
                } catch (final IOException e) {
                    LOGGER.error("Error while restarting: " + e.getMessage(), e);
                    LOGGER.error("Exiting.");
                    System.exit(-1);
                }
                System.exit(0);
            } else {
                start();
            }
        }
    }

    /**
     * Stop the slave.
     */
    private static void stop() {
        LOGGER.info("Sending stop message...");
        try (DatagramSocket socket = new DatagramSocket()) {
            final DatagramPacket packet = new DatagramPacket(
                    AbstractMessage.STOP.getBytes(),
                    AbstractMessage.STOP.length(),
                    InetAddress.getLocalHost(),
                    CONFIGURATION.getInt("discover.port"));
            socket.send(packet);
        } catch (final IOException e) {
            LOGGER.error(e.getMessage(), e);
            LOGGER.error("Exiting.");
            System.exit(-1);
        }
    }

    /**
     * Start the slave.
     */
    private static void start() {
        // Open the UDP socket
        try {
            final DatagramSocket socket = new DatagramSocket(CONFIGURATION.getInt("discover.port"));
            final Thread t = startDiscoverMessageReader(socket);

            t.join();
        } catch (final IOException | InterruptedException e) {
            LOGGER.error(e.getMessage(), e);
        }
    }

    /**
     * create and starts the discover message reader thread.
     *
     * @param socket The socket from which the discover message shall be read.
     * @return The discover message reader thread.
     */
    private static Thread startDiscoverMessageReader(final DatagramSocket socket) {
        final Thread t = new Thread(() -> {
            // Wait for the discover message
            LOGGER.info("Waiting incoming connections...");

            final int bufferSize = 1024;
            final DatagramPacket packet = new DatagramPacket(new byte[bufferSize], bufferSize);
            while (true) {
                try {
                    socket.receive(packet);

                    if (shuttingDown) {
                        break;
                    }

                    if (packet.getLength() == AbstractMessage.DISCOVER.length()
                            && new String(packet.getData(), 0, AbstractMessage.DISCOVER.length()).equals(AbstractMessage.DISCOVER)) {
                        discoverReceived(packet);
                    } else if (packet.getLength() == AbstractMessage.STOP.length()
                            && new String(packet.getData(), 0, AbstractMessage.STOP.length()).equals(AbstractMessage.STOP)) {
                        stopReceived();
                    }
                } catch (final IOException | ClassNotFoundException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        });
        t.setName("Discover message reader");
        t.start();
        return t;
    }

    /**
     * A "discover" message has been received, treat it.
     *
     * @param packet The datagram packet containing the discover message.
     * @throws IOException If an I/O error occurs while reading stream.
     * @throws ClassNotFoundException If a network object cannot be unmarshalled.
     */
    public static synchronized void discoverReceived(final DatagramPacket packet)
            throws IOException, ClassNotFoundException {
        synchronized (slaves) {
            for (final Slave slave : slaves) {
                if (slave.socket.getInetAddress().equals(packet.getAddress())) {
                    return;
                }
            }

            final byte[] localhost = new byte[] { 127, 0, 0, 1 };

            if (!Arrays.equals(packet.getAddress().getAddress(), localhost)) {
                LOGGER.info("DISCOVER message received.");

                try {
                    final Slave slave = new Slave(packet.getAddress());
                    slaves.add(slave);
                } catch (final AddressNotAccessibleException | SocketException e) {
                    // If SocketException, maybe the user has twitched on the button
                    LOGGER.warn(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * A "stop" message has been received, treat it.
     */
    private static void stopReceived() {
        for (final Slave slave : slaves) {
            try {
                slave.writeMessage(slave.out, new CloseConnectionMessage());
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
            }
        }
        LOGGER.info("STOP message received.");
        LOGGER.info("Exiting.");
        System.exit(0);
    }

}
