/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.net.NetworkInterface;
import java.net.Socket;
import java.nio.file.Files;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.lf5.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A utility class offering convenience methods.
 *
 * @author Zleurtor
 */
public final class Utils {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(Utils.class);

    /**
     * An integer used to "unsign" a byte.
     */
    private static final int UNSIGN = 0xff;

    /**
     * A private empty constructor.
     */
    private Utils() {
        super();
    }

    /**
     * Executes the specified command.
     *
     * @param command The command to execute.
     * @return The process representing the command execution.
     * @throws IOException If an Input/Output error occurs
     */
    public static Process execute(final String command) throws IOException {
        return execute(command, true);
    }

    /**
     * Executes the specified command.
     *
     * @param command The command to execute.
     * @param log Whether the information messages shall be logged.
     * @return The process representing the command execution.
     * @throws IOException If an Input/Output error occurs
     */
    public static Process execute(final String command, final boolean log) throws IOException {
        final Process p = Runtime.getRuntime().exec(command);

        // Read the process's output and error stream
        readStream(p.getInputStream(), log, false);
        readStream(p.getErrorStream(), true, true);

        return p;
    }

    /**
     * Reads the content of the specified input stream.
     *
     * @param in The input stream from which to read.
     * @param log Whether the information messages shall be logged.
     * @param error Whether the information messages shall be logged in error.
     */
    private static void readStream(final InputStream in, final boolean log, final boolean error) {
        final Thread t = new Thread(() -> {
            try {
                String outString = "";

                for (int read = in.read(); read != -1; read = in.read()) {
                    if (read == '\r' || read == '\n') {
                        if (!outString.isEmpty()) {
                            if (log) {
                                if (error) {
                                    LOGGER.error(outString);
                                } else {
                                    LOGGER.info(outString);
                                }
                            }
                            outString = "";
                        }
                    } else {
                        outString += (char) read;
                    }
                }
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
            } finally {
                try {
                    in.close();
                } catch (final IOException e) {
                    LOGGER.warn(e.getMessage(), e);
                }
            }
        });
        t.setName("Stream logger Thread");
        t.start();
    }

    /**
     * Tests whether the specified String represents a number.
     *
     * @param s The String to check whether it represents a number.
     * @return True if the specified String represents a number, false otherwise.
     */
    public static boolean isNumber(final String s) {
        try {
            Double.parseDouble(s);
            return true;
        } catch (final NumberFormatException e) {
            return false;
        }
    }

    /**
     * Extracts the IP address from the specified address String.
     *
     * @param address The address String from which the IP address shall be extracted.
     * @return The IP address extracted from the specified address String.
     */
    public static String extractIp(final String address) {
        final Pattern pattern = Pattern.compile("(.*)/([0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3})[:,](.*)");
        final Matcher matcher = pattern.matcher(address);
        matcher.find();

        final String ip = matcher.group(2);

        return ip;
    }

    /**
     * Get the mac address of the network interface to which the given socket is bound.
     *
     * @param socket A socket bound to the network interface from which we want the mac address.
     * @return The mac address of the network interface on which the specified socket was created.
     * @throws IOException If an Input/Output error occurs.
     * @throws AddressNotAccessibleException If the address of the socket is not accessible.
     */
    public static String getMacAddress(final Socket socket) throws IOException, AddressNotAccessibleException {
        String mac = "";

        final byte[] hardwareAddress = NetworkInterface.getByInetAddress(socket.getLocalAddress()).getHardwareAddress();
        if (hardwareAddress == null) {
            throw new AddressNotAccessibleException("The address " + socket.getLocalAddress() + " doesn't exist or is not accessible.");
        }

        for (final byte b : hardwareAddress) {
            if (!mac.isEmpty()) {
                mac += ":";
            }
            final short unsigned = 0xff;
            String stringByte = Integer.toHexString(b & unsigned);
            if (stringByte.length() < 2) {
                stringByte = "0" + stringByte;
            }
            mac += stringByte;
        }

        return mac;
    }

    /**
     * Delete recursively the directory.
     *
     * @param dir The directory to recursively delete.
     * @throws IOException If an Input/Output error occurs.
     */
    public static void recursiveDelete(final File dir) throws IOException {
        for (final File file : dir.listFiles()) {
            if (file.isDirectory()) {
                recursiveDelete(file);
            }

            Files.delete(file.toPath());
        }
    }

    /**
     * Extract the log file name from the log4j configuration file.
     *
     * @return The log file name.
     * @throws IOException If an input/output error occurs while reading the log4j configuration file.
     */
    public static File getLogFile() throws IOException {
        try (InputStream in = Slave.class.getResource("/log4j.properties").openStream()) {
            final Properties p = new Properties();
            p.load(in);
            return new File(p.getProperty("log4j.appender.file.file"));
        }
    }

    /**
     * Extract the log level from the log4j configuration file.
     *
     * @return The log level.
     * @throws IOException If an input/output error occurs while reading the log4j configuration file.
     */
    public static LogLevel getLogLevel() throws IOException {
        try (InputStream in = Slave.class.getResource("/log4j.properties").openStream()) {
            final Properties p = new Properties();
            p.load(in);

            String logLevel = p.getProperty("log4j.rootLogger");
            logLevel = logLevel.substring(0, logLevel.indexOf(",")).trim();

            for (final Field f : Class.forName(LogLevel.class.getName()).getDeclaredFields()) {
                if (f.getType().equals(Class.forName(LogLevel.class.getName())) && f.getName().equals(logLevel)) {
                    return (LogLevel) f.get(null);
                }
            }

            return null;
        } catch (final SecurityException | ClassNotFoundException | IllegalArgumentException | IllegalAccessException e) {
            LOGGER.warn(e.getMessage(), e);
            return null;
        }
    }

    /**
     * Creates and return a comparator that shall be used to compare String IPs.
     *
     * @return A comparator that shall be used to compare String IPs.
     */
    public static Comparator<String> getIpComparator() {
        return (final String ipString1, final String ipString2) -> {
            final String[] ipStringArray1 = ipString1.split("\\.");
            final String[] ipStringArray2 = ipString2.split("\\.");

            final int[] ip1 = new int[ipStringArray1.length];
            for (int i = 0; i < ipStringArray1.length; i++) {
                ip1[i] = Integer.parseInt(ipStringArray1[i]);
            }

            final int[] ip2 = new int[ipStringArray2.length];
            for (int i = 0; i < ipStringArray2.length; i++) {
                ip2[i] = Integer.parseInt(ipStringArray2[i]);
            }

            for (int i = 0; i < ip1.length && i < ip2.length; i++) {
                if (ip1[i] != ip2[i]) {
                    return ip1[i] - ip2[i];
                }
            }

            return 0;
        };
    }

    /**
     * Compare two version numbers.
     *
     * @param v1 The first version number to compare.
     * @param v2 second first version number to compare.
     * @return A negative integer, zero, or a positive integer as v1 is less than, equal to, or greater than v2.
     */
    public static int compareVersion(final String v1, final String v2) {
        String[] v1Split = v1.split("\\.");
        String[] v2Split = v2.split("\\.");

        v1Split = Arrays.copyOf(v1Split, Math.max(v1Split.length, v2Split.length));
        v2Split = Arrays.copyOf(v2Split, Math.max(v1Split.length, v2Split.length));

        String normalizedV1 = "";
        String normalizedV2 = "";

        for (int i = 0; i < v1Split.length && i < v2Split.length; i++) {
            String s1 = v1Split[i] == null ? "" : v1Split[i];
            String s2 = v2Split[i] == null ? "" : v2Split[i];

            while (s1.length() < s2.length()) {
                s1 = "0" + s1;
            }
            while (s2.length() < s1.length()) {
                s2 = "0" + s2;
            }

            normalizedV1 += (normalizedV1.length() > 0 ? "." : "") + s1;
            normalizedV2 += (normalizedV2.length() > 0 ? "." : "") + s2;
        }

        return normalizedV1.compareTo(normalizedV2);
    }

    /**
     * Convert an array of bytes to an integer.
     *
     * @param array The array of bytes to convert.
     * @return The integer converted from the array of bytes.
     */
    public static int toInt(final byte[] array) {
        int integer = 0;
        for (int i = 0; i < array.length; i++) {
            final byte b = array[i];

            integer <<= Byte.SIZE;
            integer |= b & UNSIGN;
        }

        return integer;
    }

    /**
     * Convert an integer to an array of bytes.
     *
     * @param i The integer to convert.
     * @return The array of bytes converted from the integer.
     */
    public static byte[] toArray(final int i) {
        final byte[] array = new byte[Integer.BYTES];

        int tmp = i;
        for (int j = array.length - 1; j >= 0; j--) {
            array[j] = (byte) (tmp & UNSIGN);
            tmp >>= Byte.SIZE;
        }

        return array;
    }

    /**
     * Convert an ip (an array of 4 bytes) as a string to an integer.
     *
     * @param ip The ip given as a string.
     * @return the integer of the 4 bytes of the given ip.
     */
    public static int ipToInt(final String ip) {
        int i = 0;

        for (final String b : ip.split("\\.")) {
            i <<= Byte.SIZE;
            i |= Integer.parseInt(b);
        }

        return i;
    }

}
