/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.events;

import java.util.EventListener;

import org.apache.log4j.lf5.LogLevel;

/**
 * The interface to implement in order to receive master events.
 *
 * @author Zleurtor
 */
public interface MasterListener extends EventListener {

    /**
     * The event fired when the slaves list is cleared.
     */
    void slavesCleared();

    /**
     * The event fired when a new slave incame.
     *
     * @param evt The event object containing information on the new slave.
     */
    void newSlave(MasterEvent evt);

    /**
     * The event fired when a new slave has gone.
     *
     * @param evt The event object containing information on the slave which has gone.
     */
    void slaveQuit(MasterEvent evt);

    /**
     * The event fired when a snapshot has been successfully processed.
     *
     * @param evt The event object containing information on the slave which processed the snapshot.
     */
    void snapshotSuccess(MasterEvent evt);

    /**
     * The event fired when an error occurred during a snapshot process.
     *
     * @param evt The event object containing information on the slave which failed during the snapshot.
     */
    void snapshotError(MasterEvent evt);

    /**
     * The event fired when the snapshots have been cleared.
     */
    void snapshotsCleared();

    /**
     * The event fired when an error occurred during a preview process.
     *
     * @param evt The event object containing information on the slave which failed during the stream process.
     */
    void streamError(MasterEvent evt);

    /**
     * The event fired when a log file content income.
     *
     * @param evt The event object containing the slave's log file content.
     * @param level The slave's log level.
     */
    void log(MasterEvent evt, LogLevel level);

    /**
     * The event fired when a log level change error income.
     *
     * @param evt The event object containing the slave's log level change error.
     */
    void logLevelError(MasterEvent evt);

    /**
     * The event fired when a log level change success income.
     *
     * @param evt The event object containing the slave's log level success.
     */
    void logLevelSuccess(MasterEvent evt);

    /**
     * The event fired when an options changed income.
     *
     * @param evt The event object containing the slave's options.
     */
    void options(MasterEvent evt);

    /**
     * The event fired when a file has been saved.
     *
     * @param evt The event object containing the saved file.
     */
    void fileSaved(MasterEvent evt);
}
