/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * The super class of all the messages exchanged by the master and the slaves.
 *
 * @author Zleurtor
 */
public abstract class AbstractMessage implements Serializable {

    /**
     * The message to broadcast to all the slaves to discover them.
     */
    public static final String DISCOVER = "WHO IS THERE ?";

    /**
     * The message to send to stop the slave.
     */
    public static final String STOP = "STOP !";

    /**
     * A Logger...
     */
    protected static final Logger LOGGER = LoggerFactory.getLogger(AbstractMessage.class);

    private static final long serialVersionUID = 3273276612817488340L;

    /**
     * Constructs a new empty message.
     */
    protected AbstractMessage() {
        super();
    }

    /**
     * Transform this message to an XML description.
     *
     * @return An XML describing this message.
     * @throws ParserConfigurationException If a DocumentBuildercannot be created which satisfies the configuration requested.
     * @throws TransformerFactoryConfigurationError Thrown in case of service configuration error or ifthe implementation is not available
     *                                              or cannot be instantiated.
     * @throws TransformerException When it is not possible to create a Transformer instance or if an unrecoverable error occurs during the
     *                              course of the transformation.
     * @throws IOException If an I/O error occurs.
     */
    protected abstract String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException;

    /**
     * Set this message values to the ones in the given XML description.
     *
     * @param xml The XML containing the values to set in this message.
     * @return This message.
     * @throws SAXException If any parse errors occur.
     * @throws IOException If any IO errors occur.
     * @throws ParserConfigurationException If a DocumentBuildercannot be created which satisfies the configuration requested.
     * @throws XPathExpressionException If an XPath expression cannot be evaluated.
     */
    protected abstract AbstractMessage fromXml(String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException;

    /**
     * Serialize this object to an array of bytes (avoid the use of ObjectInputStream because of the presence of the
     * handles that spend a lot of memory).
     *
     * @return An array of bytes that is the serialized form of this message.
     * @throws ParserConfigurationException If a DocumentBuildercannot be created which satisfies the configuration requested.
     * @throws TransformerFactoryConfigurationError Thrown in case of service configuration error or ifthe implementation is not available
     *                                              or cannot be instantiated.
     * @throws TransformerException When it is not possible to create a Transformer instance or if an unrecoverable error occurs during the
     *                              course of the transformation.
     * @throws IOException If an I/O error occurs.
     */
    public byte[] serialize()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        return toXml().getBytes();
    }

    /**
     * Unserialize this object from an array of bytes (avoid the use of ObjectInputStream because of the presence of the
     * handles that spend a lot of memory).
     *
     * @param serialized The serialized form of the message to unserialize.
     * @return A message object that is the unserialized form of the given byte array.
     * @throws SAXException If any parse errors occur.
     * @throws IOException If any IO errors occur.
     * @throws ParserConfigurationException If a DocumentBuildercannot be created which satisfies the configuration requested.
     * @throws XPathExpressionException If an XPath expression cannot be evaluated.
     * @throws ClassNotFoundException If the class cannot be located
     * @throws IllegalAccessException If the class or its nullary constructor is not accessible.
     * @throws InstantiationException If this Class represents an abstract class, an interface, an array class, a primitive type, or void;
     *                                or if the class has no nullary constructor or if the instantiation fails for some other reason.
     * @throws SecurityException If a security manager, s, is present and the caller's class loader is not the same as or an ancestor of the
     *                           class loader for the current class and invocation of s.checkPackageAccess() denies access to the package of
     *                           this class.
     * @throws NoSuchMethodException If a matching constructor is not found.
     * @throws InvocationTargetException If the underlying constructor throws an exception.
     * @throws IllegalArgumentException  If this Constructor object is enforcing Java language access control and the underlying constructor
     *                                   is inaccessible.
     */
    public static AbstractMessage unserialize(final byte[] serialized)
            throws XPathExpressionException, SAXException, IOException, ParserConfigurationException, InstantiationException,
                   IllegalAccessException, ClassNotFoundException, IllegalArgumentException, InvocationTargetException,
                   NoSuchMethodException, SecurityException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(serialized)) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        final String type = (String) xPath.evaluate("/message/@type", document, XPathConstants.STRING);

        final AbstractMessage message = (AbstractMessage) Class.forName(type).getConstructor().newInstance();
        message.fromXml(new String(serialized));

        return message;
    }

}
