/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the slave during the discover process.
 *
 * @author Zleurtor
 */
public class IAmMessage extends AbstractMessage {

    private static final long serialVersionUID = 8583803286334162495L;

    /**
     * The name of the slave which sent this message.
     */
    private String name;

    /**
     * The version of the slave which sent this message.
     */
    private String version;

    /**
     * The options used by the slave while taking snapshots or streaming.
     */
    private String options;

    /**
     * Constructs an empty message.
     */
    public IAmMessage() { }

    /**
     * Constructs a new "I am" message using the specified name.
     *
     * @param theName The name of the slave which sent this message.
     * @param theVersion The version of the slave which sent this message.
     * @param theOptions The options used by the slave while taking snapshots.
     */
    public IAmMessage(final String theName, final String theVersion, final String theOptions) {
        super();
        this.name = theName;
        this.version = theVersion;
        this.options = theOptions;
    }

    /**
     * Returns the name of the slave which sent this message.
     *
     * @return The name of the slave which sent this message.
     */
    public final String getName() {
        return name;
    }

    /**
     * Returns the version of the slave which sent this message.
     *
     * @return The version of the slave which sent this message.
     */
    public String getVersion() {
        return version;
    }

    /**
     * Returns the options used by the slave while taking snapshots or streaming.
     *
     * @return The options used by the slave while taking snapshots or streaming.
     */
    public final String getOptions() {
        return options;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element nameElt = document.createElement("name");
        messageElt.appendChild(nameElt);
        nameElt.appendChild(document.createTextNode(name));

        final Element versionElt = document.createElement("version");
        messageElt.appendChild(versionElt);
        versionElt.appendChild(document.createTextNode(version));

        final Element optionsElt = document.createElement("options");
        messageElt.appendChild(optionsElt);
        optionsElt.appendChild(document.createTextNode(options));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        name = (String) xPath.evaluate("/message/name", document, XPathConstants.STRING);
        version = (String) xPath.evaluate("/message/version", document, XPathConstants.STRING);
        options = (String) xPath.evaluate("/message/options", document, XPathConstants.STRING);

        return this;
    }

}
