/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.lf5.LogLevel;
import org.apache.log4j.lf5.LogLevelFormatException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the master to change a slave log level.
 *
 * @author Zleurtor
 */
public class LogLevelMessage extends AbstractMessage {

    private static final long serialVersionUID = 3705677592248987508L;

    /**
     * The new log level to apply.
     */
    private LogLevel level;

    /**
     * Constructs an empty message.
     */
    public LogLevelMessage() { }

    /**
     * Creates a new log level message with the given new log level.
     *
     * @param theLevel The new log level to apply.
     */
    public LogLevelMessage(final LogLevel theLevel) {
        super();
        this.level = theLevel;
    }

    /**
     * Returns the new Log level to apply.
     *
     * @return The new Log level to apply.
     */
    public final LogLevel getLevel() {
        return level;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element levelElt = document.createElement("level");
        messageElt.appendChild(levelElt);
        levelElt.appendChild(document.createTextNode(level != null ? level.getLabel() : ""));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        try {
            final String logLevelLabel = (String) xPath.evaluate("/message/level", document, XPathConstants.STRING);
            level = logLevelLabel.isEmpty() ? null : LogLevel.valueOf(logLevelLabel);
        } catch (final LogLevelFormatException e) {
            throw new IOException(e);
        }

        return this;
    }

}
