/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.Base64;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.lf5.LogLevel;
import org.apache.log4j.lf5.LogLevelFormatException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the master to retrieve the slaves log.
 *
 * @author Zleurtor
 */
public class LogMessage extends AbstractMessage {

    private static final long serialVersionUID = -3351714549848040831L;

    /**
     * The level of the log file.
     */
    private LogLevel level;

    /**
     * The content of the log file.
     */
    private byte[] content;

    /**
     * Constructs an empty message.
     */
    public LogMessage() { }

    /**
     * Returns the level of the log file.
     *
     * @return The level of the log file.
     */
    public final LogLevel getLevel() {
        return level;
    }

    /**
     * Sets the level of the log file.
     *
     * @param newLevel The new level of the log file.
     */
    public final void setLevel(final LogLevel newLevel) {
        this.level = newLevel;
    }

    /**
     * Returns the content of the log file.
     *
     * @return The content of the log file.
     */
    public final byte[] getContent() {
        return content;
    }

    /**
     * Sets the content of the log file.
     *
     * @param newContent The new content of the log file.
     */
    public final void setContent(final byte[] newContent) {
        this.content = newContent;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element levelElt = document.createElement("level");
        messageElt.appendChild(levelElt);
        levelElt.appendChild(document.createTextNode(level != null ? level.getLabel() : ""));

        final Element contentElt = document.createElement("content");
        messageElt.appendChild(contentElt);
        contentElt.appendChild(document.createTextNode(
                Base64.getEncoder().encodeToString(content != null ? content : new byte[0])));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        try {
            final String logLevelLabel = (String) xPath.evaluate("/message/level", document, XPathConstants.STRING);
            level = logLevelLabel.isEmpty() ? null : LogLevel.valueOf(logLevelLabel);
        } catch (final XPathExpressionException | LogLevelFormatException e) {
            throw new IOException(e);
        }
        content = Base64.getDecoder().decode((String) xPath.evaluate("/message/content", document, XPathConstants.STRING));

        return this;
    }

}
