/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the master to shutdown the slaves.
 *
 * @author Zleurtor
 */
public class ShutdownMessage extends AbstractMessage {

    private static final long serialVersionUID = 7412717614771238134L;

    /**
     * Whether or not this message is a reboot message.
     */
    private boolean reboot;

    /**
     * Constructs an empty message.
     */
    public ShutdownMessage() { }

    /**
     * Construct a new shutdown message.
     *
     * @param isReboot Whether or not this message is a reboot message.
     */
    public ShutdownMessage(final boolean isReboot) {
        super();
        this.reboot = isReboot;
    }

    /**
     * Tell whether or not this message is a reboot message.
     *
     * @return True if this message is a reboot one, false otherwise.
     */
    public final boolean isReboot() {
        return reboot;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element rebootElt = document.createElement("reboot");
        messageElt.appendChild(rebootElt);
        rebootElt.appendChild(document.createTextNode(Boolean.toString(reboot)));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        reboot = Boolean.parseBoolean((String) xPath.evaluate("/message/reboot", document, XPathConstants.STRING));

        return this;
    }

}
