/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the server to ask the slaves to take a snapshot.
 *
 * @author Zleurtor
 */
public class SnapshotMessage extends AbstractMessage {

    private static final long serialVersionUID = -5358149642037102141L;

    /**
     * The subfolder in which the snapshot files shall be stored.
     */
    private String subFolder;

    /**
     * The information (width, height and quality) about the thumbnail to produce.
     */
    private String thumbnailInformation;

    /**
     * Whether or not the snapshot shall be polychromatic.
     */
    private boolean polychrome;

    /**
     * Constructs an empty message.
     */
    public SnapshotMessage() { }

    /**
     * Constructs a "snapshot" message using the specified subfolder and options.
     *
     * @param theSubFolder The subfolder in which the snapshot files shall be stored.
     * @param theThumbnailInformation The information (width, height and quality) about the thumbnail to produce.
     * @param thePolychrome Whether or not the snapshot shall be polychromatic.
     */
    public SnapshotMessage(final String theSubFolder, final String theThumbnailInformation, final boolean thePolychrome) {
        super();
        this.subFolder = theSubFolder;
        this.thumbnailInformation = theThumbnailInformation;
        this.polychrome = thePolychrome;
    }

    /**
     * Returns the subfolder in which the snapshot files shall be stored.
     *
     * @return The subfolder in which the snapshot files shall be stored.
     */
    public final String getSubFolder() {
        return subFolder;
    }

    /**
     * Returns the information (width, height and quality) about the thumbnail to produce.
     *
     * @return The information (width, height and quality) about the thumbnail to produce.
     */
    public final String getThumbnailInformation() {
        return thumbnailInformation;
    }

    /**
     * Returns whether or not the snapshot shall be polychromatic.
     *
     * @return Whether or not the snapshot shall be polychromatic.
     */
    public boolean isPolychrome() {
        return polychrome;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element subFolderElt = document.createElement("subFolder");
        messageElt.appendChild(subFolderElt);
        subFolderElt.appendChild(document.createTextNode(subFolder));

        final Element thumbnailInformationElt = document.createElement("thumbnailInformation");
        messageElt.appendChild(thumbnailInformationElt);
        thumbnailInformationElt.appendChild(document.createTextNode(thumbnailInformation));

        final Element polychromeElt = document.createElement("polychrome");
        messageElt.appendChild(polychromeElt);
        polychromeElt.appendChild(document.createTextNode(Boolean.toString(polychrome)));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        subFolder = (String) xPath.evaluate("/message/subFolder", document, XPathConstants.STRING);
        thumbnailInformation =
                (String) xPath.evaluate("/message/thumbnailInformation", document, XPathConstants.STRING);
        polychrome = Boolean.parseBoolean(
                (String) xPath.evaluate("/message/polychrome", document, XPathConstants.STRING));

        return this;
    }

}
