/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.Base64;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by a slave when the snapshot process succeed.
 *
 * @author Zleurtor
 */
public class SnapshotSuccessMessage extends AbstractMessage {

    private static final long serialVersionUID = -7851110296197306960L;

    /**
     * The file content of the snapshot.
     */
    private byte[] fileContent;

    /**
     * The sub folder where the snapshot shall be written (by the client).
     */
    private String subFolder;

    /**
     * Constructs an empty message.
     */
    public SnapshotSuccessMessage() { }

    /**
     * Constructs a new "snapshot success" message using the specified filename.
     *
     * @param theFilename The filename of the snapshot.
     * @param theSubFolder The sub folder where the snapshot shall be written (by the client).
     * @throws IOException If an input/output error occurs during the file reading.
     */
    public SnapshotSuccessMessage(final String theFilename, final String theSubFolder) throws IOException {
        super();
        this.subFolder = theSubFolder;

        if (theFilename != null) {
            try (ByteArrayOutputStream out = new ByteArrayOutputStream();
                    InputStream in = new FileInputStream(theFilename)) {
                final int bufferSize = 1024;
                final byte[] buffer = new byte[bufferSize];
                for (int r = in.read(buffer); r != -1; r = in.read(buffer)) {
                    out.write(buffer, 0, r);
                }
                fileContent = out.toByteArray();
            }
        } else {
            fileContent = null;
        }
    }

    /**
     * Returns the file content of the snapshot.
     *
     * @return The file content of the snapshot.
     */
    public final byte[] getFileContent() {
        return fileContent;
    }

    /**
     * Returns the sub folder where the snapshot shall be written (by the client).
     *
     * @return The sub folder where the snapshot shall be written (by the client).
     */
    public final String getSubFolder() {
        return subFolder;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element fileContentElt = document.createElement("fileContent");
        messageElt.appendChild(fileContentElt);
        fileContentElt.appendChild(document.createTextNode(
                Base64.getEncoder().encodeToString(fileContent != null ? fileContent : new byte[0])));

        final Element subFolderElt = document.createElement("subFolder");
        messageElt.appendChild(subFolderElt);
        subFolderElt.appendChild(document.createTextNode(subFolder));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        fileContent = Base64.getDecoder().decode(
                (String) xPath.evaluate("/message/fileContent", document, XPathConstants.STRING));
        subFolder = (String) xPath.evaluate("/message/subFolder", document, XPathConstants.STRING);

        return this;
    }

}
