/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.List;
import java.util.UUID;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the server to update a slave.
 *
 * @author Zleurtor
 */
public class UpdateMessage extends AbstractMessage {

    private static final long serialVersionUID = -463825257256386613L;

    /**
     * The size of each update message chunk.
     */
    private static final int CHUNK_SIZE = 1024 * 1024;

    /**
     * The update id.
     */
    private String id;

    /**
     * The file content of the update.
     */
    private byte[] fileContent;

    /**
     * The update file prefix.
     */
    private String prefix;

    /**
     * The update file suffix.
     */
    private String suffix;

    /**
     * The total number of chunks for the update file.
     */
    private int totalChunks;

    /**
     * The current chunk number in the update file.
     */
    private int chunkNumber;

    /**
     * Constructs an empty message.
     */
    public UpdateMessage() { }

    /**
     * Constructs a new "update success" message using the specified file content.
     *
     * @param theId The update id.
     * @param theFileContent The update file content.
     * @param thePrefix The update file prefix.
     * @param theSuffix The update file suffix.
     * @param theTotalChunks The total number of chunks for the update file.
     * @param theChunkNumber The current chunk number in the update file.
     */
    private UpdateMessage(final String theId, final byte[] theFileContent, final String thePrefix, final String theSuffix,
            final int theTotalChunks, final int theChunkNumber) {
        super();

        id = theId;
        fileContent = theFileContent;
        prefix = thePrefix;
        suffix = theSuffix;
        totalChunks = theTotalChunks;
        chunkNumber = theChunkNumber;
    }

    /**
     * Create a new list of messages for the given update file.
     *
     * @param theUpdateFile The update file.
     * @param thePrefix The update file prefix.
     * @param theSuffix The update file suffix.
     * @throws IOException if an I/O error occurs while reading the update file.
     * @return A list of messages for the given update file.
     */
    public static final List<UpdateMessage> newMessages(final File theUpdateFile, final String thePrefix, final String theSuffix)
            throws IOException {
        final List<UpdateMessage> messages = new ArrayList<>();

        final String id = UUID.randomUUID().toString();

        final byte[] fileContent;
        if (theUpdateFile != null) {
            try (ByteArrayOutputStream out = new ByteArrayOutputStream();
                    InputStream in = new FileInputStream(theUpdateFile)) {
                final int bufferSize = 1024;
                final byte[] buffer = new byte[bufferSize];
                for (int r = in.read(buffer); r != -1; r = in.read(buffer)) {
                    out.write(buffer, 0, r);
                }
                fileContent = out.toByteArray();
            }
        } else {
            throw new IOException("The update file cannot be null.");
        }

        final int totalChunks = (fileContent.length / CHUNK_SIZE) + ((fileContent.length % CHUNK_SIZE != 0) ? 1 : 0);
        for (int i = 0; i < totalChunks; i++) {
            messages.add(new UpdateMessage(id,
                    Arrays.copyOfRange(fileContent, i * CHUNK_SIZE, Math.min((i + 1) * CHUNK_SIZE, fileContent.length)),
                    thePrefix, theSuffix, totalChunks, i + 1));
        }

        return messages;
    }

    /**
     * Returns the update id.
     *
     * @return The update id.
     */
    public String getId() {
        return id;
    }

    /**
     * Returns the file content of the update.
     *
     * @return The file content of the update.
     */
    public final byte[] getFileContent() {
        return fileContent;
    }

    /**
     * Returns the update file prefix.
     *
     * @return The update file prefix.
     */
    public String getPrefix() {
        return prefix;
    }

    /**
     * Returns the update file suffix.
     *
     * @return The update file suffix.
     */
    public String getSuffix() {
        return suffix;
    }

    /**
     * Returns the total number of chunks for the update file.
     *
     * @return The total number of chunks for the update file.
     */
    public final int getTotalChunks() {
        return totalChunks;
    }

    /**
     * Returns the current chunk number in the update file.
     *
     * @return The current chunk number in the update file.
     */
    public int getChunkNumber() {
        return chunkNumber;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);
        messageElt.setAttribute("id", id);

        final Element fileContentElt = document.createElement("fileContent");
        messageElt.appendChild(fileContentElt);
        fileContentElt.appendChild(document.createTextNode(
                Base64.getEncoder().encodeToString(fileContent != null ? fileContent : new byte[0])));
        fileContentElt.setAttribute("totalChunks", Integer.toString(totalChunks));
        fileContentElt.setAttribute("chunkNumber", Integer.toString(chunkNumber));

        final Element prefixElt = document.createElement("prefix");
        messageElt.appendChild(prefixElt);
        prefixElt.appendChild(document.createTextNode(prefix));

        final Element suffixElt = document.createElement("suffix");
        messageElt.appendChild(suffixElt);
        suffixElt.appendChild(document.createTextNode(suffix));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        id = (String) xPath.evaluate("/message/@id", document, XPathConstants.STRING);
        fileContent = Base64.getDecoder().decode(
                (String) xPath.evaluate("/message/fileContent", document, XPathConstants.STRING));
        totalChunks = ((Double) xPath.evaluate("/message/fileContent/@totalChunks", document, XPathConstants.NUMBER)).intValue();
        chunkNumber = ((Double) xPath.evaluate("/message/fileContent/@chunkNumber", document, XPathConstants.NUMBER)).intValue();
        prefix = (String) xPath.evaluate("/message/prefix", document, XPathConstants.STRING);
        suffix = (String) xPath.evaluate("/message/suffix", document, XPathConstants.STRING);

        return this;
    }

}
