/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.messages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * The message sent by the server to ask a slave to identify itself.
 *
 * @author Zleurtor
 */
public class WhoAreYouMessage extends AbstractMessage {

    private static final long serialVersionUID = 9035154333514936993L;

    /**
     * The port on which heartbeat messages are accepted.
     */
    private int heartbeatPort = -1;

    /**
     * The period between two heartbeat messages.
     */
    private int heartbeatPeriod = 0;

    /**
     * Constructs an empty message.
     */
    public WhoAreYouMessage() { }

    /**
     * Constructs message with the given heartbeat port and period.
     *
     * @param theHeartbeatPort The port on which heartbeat messages are accepted.
     * @param theHeartbeatPeriod The period between two heartbeat messages.
     */
    public WhoAreYouMessage(final int theHeartbeatPort, final int theHeartbeatPeriod) {
        this.heartbeatPort = theHeartbeatPort;
        this.heartbeatPeriod = theHeartbeatPeriod;
    }

    /**
     * The port on which heartbeat messages are accepted.
     *
     * @return The port on which heartbeat messages are accepted.
     */
    public int getHeartbeatPort() {
        return heartbeatPort;
    }

    /**
     * The period between two heartbeat messages.
     *
     * @return The period between two heartbeat messages.
     */
    public int getHeartbeatPeriod() {
        return heartbeatPeriod;
    }

    @Override
    protected final String toXml()
            throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, IOException {
        final Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

        final Element messageElt = document.createElement("message");
        messageElt.setAttribute("type", getClass().getName());
        document.appendChild(messageElt);

        final Element heartbeatPortElt = document.createElement("heartbeatPort");
        messageElt.appendChild(heartbeatPortElt);
        heartbeatPortElt.appendChild(document.createTextNode(Integer.toString(heartbeatPort)));

        final Element heartbeatPeriodElt = document.createElement("heartbeatPeriod");
        messageElt.appendChild(heartbeatPeriodElt);
        heartbeatPeriodElt.appendChild(document.createTextNode(Integer.toString(heartbeatPeriod)));

        try (StringWriter writer = new StringWriter()) {
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            return writer.toString();
        }
    }

    @Override
    protected final AbstractMessage fromXml(final String xml)
            throws SAXException, IOException, ParserConfigurationException, XPathExpressionException {
        final Document document;
        try (InputStream in = new ByteArrayInputStream(xml.getBytes())) {
            document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
        }
        final XPath xPath = XPathFactory.newInstance().newXPath();

        final String heartbeatPortString = (String) xPath.evaluate("/message/heartbeatPort", document, XPathConstants.STRING);
        this.heartbeatPort = Integer.parseInt(heartbeatPortString);

        final String heartbeatPeriodString = (String) xPath.evaluate("/message/heartbeatPeriod", document, XPathConstants.STRING);
        this.heartbeatPeriod = Integer.parseInt(heartbeatPeriodString);

        return this;
    }

}
