/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.update;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.DigestException;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.Properties;

/**
 * The class having in charge to search the updates and apply it (if needed).
 *
 * @author Zleurtor
 */
public abstract class AbstractUpdater {

    /**
     * Check if an update is available (and install it).
     *
     * @return true if an update has been done, false otherwise.
     */
    public abstract boolean update();

    /**
     * Compare some part of the archive and some part of the installed files, then compute the update plan.
     *
     * @param archive The new 3D Scan Manager version archive.
     * @return The collection of files to extract from the archive.
     * @throws IOException If the new version archive file can not be read.
     * @throws FileNotFoundException If the new version archive file does not exist.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws DigestException If a digest error occurs.
     * @throws ClassNotFoundException If the Class of a serialized object cannot be found.
     */
    protected abstract Collection<String> computePlan(File archive)
            throws FileNotFoundException, IOException, NoSuchAlgorithmException, DigestException, ClassNotFoundException;

    /**
     * Merge the configuration contained in the specified input stream with the one contained in the specified file.
     *
     * @param patchIn The input stream from the patch configuration.
     * @param filename The name of the file containing the current configuration.
     * @throws IOException If an input/output error occurs.
     */
    protected void mergeConfigurations(final InputStream patchIn, final String filename)
            throws IOException {
        final Properties properties = new Properties();
        properties.load(patchIn);

        String version = properties.getProperty("version");
        version = version == null ? "" : version;

        try (InputStream currentIn = new FileInputStream(new File(filename))) {
            properties.load(currentIn);
        }

        properties.setProperty("version", version);
        properties.store(new FileOutputStream(filename), null);
    }

}
