/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.update;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Collection;

import javax.net.ssl.SSLContext;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.http.Header;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.conn.ssl.TrustAllStrategy;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicHeader;
import org.apache.http.ssl.SSLContextBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import zleurtor.scanmanager.model.Utils;
import zleurtor.scanmanager.model.events.ArtifactRepositoryManagerConnectorEvent;
import zleurtor.scanmanager.model.events.ArtifactRepositoryManagerConnectorListener;

/**
 * The connector for the Archiva artifact repository manager.
 *
 * @author Zleurtor
 */
public final class ArchivaConnector implements ArtifactRepositoryManagerConnector {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(ArchivaConnector.class);

    /**
     * The kibi unit factor.
     */
    private static final int KIBI = 1024;

    /**
     * The mibi unit factor.
     */
    private static final int MIBI = 1024 * 1024;

    /**
     * The HTTP client to use for executing a request.
     */
    private final CloseableHttpClient client;

    /**
     * The URL to the Archiva repository manager.
     */
    private final String archivaUrl;

    /**
     * All the objects listening for artifact repository manager connector events.
     */
    private final Collection<ArtifactRepositoryManagerConnectorListener> listeners = new ArrayList<>();

    /**
     * Constructs a new artifact repository manager connector for the specified URL.
     *
     * @param theArchivaUrl The URL to the Archiva repository manager.
     * @throws KeyStoreException If the TrustStrategy fails to load.
     * @throws NoSuchAlgorithmException If no Provider supports aTrustManagerFactorySpi implementation for the "trust all" strategy.
     * @throws KeyManagementException If the SSL context builder fails to initialize.
     */
    public ArchivaConnector(final String theArchivaUrl)
            throws KeyManagementException, NoSuchAlgorithmException, KeyStoreException {
        super();
        this.archivaUrl = theArchivaUrl.endsWith("/") ? theArchivaUrl : theArchivaUrl + "/";

        // Create the HTTP client
        final Collection<Header> headers = new ArrayList<>();
        headers.add(new BasicHeader("Authorization", "Basic " + Base64.getEncoder().encodeToString("guest:".getBytes())));
        headers.add(new BasicHeader("Referer", "https://zleurtor.ddns.net:9443/archiva"));
        headers.add(new BasicHeader("Accept", "application/xml"));

        final SSLContext context = SSLContextBuilder
                .create()
                .loadTrustMaterial(new TrustAllStrategy())
                .build();

        client = HttpClientBuilder
                .create()
                .setDefaultHeaders(headers)
                .setSSLContext(context)
                .build();
    }

    @Override
    public String searchLatestVersion(final String groupId, final String artifactId) {
        LOGGER.debug("Search versions for " + groupId + ":" + artifactId);

        String latestVersion = null;

        try {
            Document content = null;

            // Create the request
            final String urlString = archivaUrl + "restServices/archivaServices/browseService/versionsList/" + groupId + "/" + artifactId;
            final HttpGet request = new HttpGet(urlString);

            // Execute the request
            final CloseableHttpResponse response = client.execute(request);
            final InputStream in = response.getEntity().getContent();

            // Parse the response content
            content = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);

            in.close();
            response.close();

            final Element versionsList = content.getDocumentElement();
            final NodeList versions = versionsList.getElementsByTagName("versions");
            for (int i = 0; i < versions.getLength(); i++) {
                final String version = versions.item(i).getTextContent();
                if (!version.endsWith("-SNAPSHOT") && (latestVersion == null || Utils.compareVersion(latestVersion, version) < 0)) {
                    latestVersion = version;
                }
            }
        } catch (final IOException | SAXException | ParserConfigurationException e) {
            LOGGER.warn("Error while searching the " + artifactId + " latest version: " + e.getMessage(), e);
        }

        return latestVersion;
    }

    @Override
    public URL searchArtifactUrl(final String groupId, final String artifactId, final String version, final String classifier,
            final String packaging) throws MalformedURLException {
        LOGGER.debug("Search url for " + groupId + ":" + artifactId + ":" + version + ":" + classifier + ":" + packaging);

        String artifactUrl = null;

        try {
            Document content = null;

            // Create the request
            final String url = archivaUrl + "restServices/archivaServices/browseService/artifactDownloadInfos/"
                    + groupId + "/" + artifactId + "/" + version;
            final HttpGet request = new HttpGet(url);

            // Execute the request
            final CloseableHttpResponse response = client.execute(request);
            final InputStream in = response.getEntity().getContent();

            // Parse the response content
            content = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);

            in.close();
            response.close();

            final Element artifacts = content.getDocumentElement();
            final NodeList artifactList = artifacts.getElementsByTagName("artifact");
            for (int i = 0; i < artifactList.getLength(); i++) {
                final Element artifact = (Element) artifactList.item(i);
                final String artifactClassifier = artifact.getElementsByTagName("classifier").getLength() > 0
                        ? artifact.getElementsByTagName("classifier").item(0).getTextContent() : "";
                final String artifactPackaging = artifact.getElementsByTagName("packaging").getLength() > 0
                        ? artifact.getElementsByTagName("packaging").item(0).getTextContent() : "";

                if (artifactClassifier.equals(classifier) && artifactPackaging.equals(packaging)) {
                    artifactUrl = artifact.getElementsByTagName("url").item(0).getTextContent();
                }
            }
        } catch (final IOException | SAXException | ParserConfigurationException e) {
            LOGGER.warn("Error while searching the " + artifactId + "-" + version + " URL: " + e.getMessage(), e);
        }

        return new URL(artifactUrl);
    }

    @Override
    public int searchArtifactSize(final String groupId, final String artifactId, final String version, final String classifier,
            final String packaging) throws MalformedURLException {
        LOGGER.debug("Search url for " + groupId + ":" + artifactId + ":" + version + ":" + classifier + ":" + packaging);

        int artifactSize = 0;

        try {
            Document content = null;

            // Create the request
            final String url = archivaUrl + "restServices/archivaServices/browseService/artifactDownloadInfos/"
                    + groupId + "/" + artifactId + "/" + version;
            final HttpGet request = new HttpGet(url);

            // Execute the request
            final CloseableHttpResponse response = client.execute(request);
            final InputStream in = response.getEntity().getContent();

            // Parse the response content
            content = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);

            in.close();
            response.close();

            final Element artifacts = content.getDocumentElement();
            final NodeList artifactList = artifacts.getElementsByTagName("artifact");
            for (int i = 0; i < artifactList.getLength(); i++) {
                final Element artifact = (Element) artifactList.item(i);
                final String artifactClassifier = artifact.getElementsByTagName("classifier").getLength() > 0
                        ? artifact.getElementsByTagName("classifier").item(0).getTextContent() : "";
                final String artifactPackaging = artifact.getElementsByTagName("packaging").getLength() > 0
                        ? artifact.getElementsByTagName("packaging").item(0).getTextContent() : "";

                if (artifactClassifier.equals(classifier) && artifactPackaging.equals(packaging)) {
                    String size = artifact.getElementsByTagName("size").item(0).getTextContent();
                    final int factor = size.contains("K") ? KIBI : size.contains("M") ? MIBI : 1;
                    size = size.replaceAll("[^0-9\\.]", "");
                    artifactSize = (int) Math.round(Double.parseDouble(size) * factor);
                }
            }
        } catch (final IOException | SAXException | ParserConfigurationException e) {
            LOGGER.warn("Error while searching the " + artifactId + "-" + version + " size: " + e.getMessage(), e);
        }

        return artifactSize;
    }

    @Override
    public void download(final URL source, final int size, final File dest)
            throws IOException, URISyntaxException {
        LOGGER.debug("Downloading " + source + "...");

        // Create the request
        final HttpGet request = new HttpGet(source.toURI());

        // Execute the request
        final CloseableHttpResponse response = client.execute(request);

        // Download the response content
        try (InputStream in = response.getEntity().getContent();
                OutputStream out = new FileOutputStream(dest)) {
            final int bufferSize = 1024;
            final byte[] buffer = new byte[bufferSize];

            int downloaded = 0;

            for (int r = in.read(buffer); r != -1; r = in.read(buffer)) {
                downloaded += r;
                out.write(buffer, 0, r);
                fireDownloadingArtifactEvent(new ArtifactRepositoryManagerConnectorEvent(this, size, downloaded));
            }
            fireDownloadingArtifactEvent(new ArtifactRepositoryManagerConnectorEvent(this, size, size));
        }

        response.close();
    }

    @Override
    public void close() throws IOException {
        client.close();
    }

    @Override
    public void addArtifactRepositoryManagerConnectorListener(final ArtifactRepositoryManagerConnectorListener listener) {
        listeners.add(listener);
    }

    /**
     * Fire a downloading artifact event to all the listeners.
     *
     * @param evt The event object containing information on the artifact which is being downloaded.
     */
    private void fireDownloadingArtifactEvent(final ArtifactRepositoryManagerConnectorEvent evt) {
        for (final ArtifactRepositoryManagerConnectorListener listener : listeners) {
            new Thread(() -> {
                listener.downloadingArtifact(evt);
            }).start();
        }
    }

}
