/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.update;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;

import zleurtor.scanmanager.model.events.ArtifactRepositoryManagerConnectorListener;

/**
 * The interface which all artifact repository manager connectors shall implements.
 *
 * @author Zleurtor
 */
public interface ArtifactRepositoryManagerConnector extends AutoCloseable {

    /**
     * Searches (and returns) the latest release version for the specified groupId and artifactId.
     *
     * @param groupId The groupId of the artifact for which the versions are queried.
     * @param artifactId The artifactId of the artifact for which the versions are queried.
     * @return All the latest release version for the specified groupId and artifactId.
     */
    String searchLatestVersion(String groupId, String artifactId);

    /**
     * Searches the URL for accessing to the artifact having the specified groupId, artifactId, version and classifier.
     *
     * @param groupId The group id of the queried artifact.
     * @param artifactId The artifact id of the queried artifact.
     * @param version The version of the queried artifact.
     * @param classifier The classifier of the queried artifact.
     * @param packaging The packaging of the queried artifact.
     * @return The URL for accessing to the artifact having the specified groupId, artifactId, version and classifier.
     * @throws MalformedURLException If a URL is malformed.
     */
    URL searchArtifactUrl(String groupId, String artifactId, String version, String classifier, String packaging)
            throws MalformedURLException;

    /**
     * Searches the size of the artifact having the specified groupId, artifactId, version and classifier.
     *
     * @param groupId The group id of the queried artifact.
     * @param artifactId The artifact id of the queried artifact.
     * @param version The version of the queried artifact.
     * @param classifier The classifier of the queried artifact.
     * @param packaging The packaging of the queried artifact.
     * @return The URL for accessing to the artifact having the specified groupId, artifactId, version and classifier.
     * @throws MalformedURLException If a URL is malformed.
     */
    int searchArtifactSize(String groupId, String artifactId, String version, String classifier, String packaging)
            throws MalformedURLException;

    /**
     * Download the content of the specified URL to the given file.
     *
     * @param source The URL of which the content shall be downloaded.
     * @param size The size of the artifact to download.
     * @param dest The file to which the URL content shall be downloaded.
     * @throws IOException If an input/output error occurs during download.
     * @throws URISyntaxException If a URI is malformed.
     */
    void download(URL source, int size, File dest) throws IOException, URISyntaxException;

    @Override
    void close() throws IOException;

    /**
     * Add a new artifact repository manager connector listener.
     *
     * @param listener The new artifact repository manager connector listener.
     */
    void addArtifactRepositoryManagerConnectorListener(ArtifactRepositoryManagerConnectorListener listener);

}
