/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.update;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.jar.JarFile;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.izforge.izpack.api.data.PackFile;
import com.izforge.izpack.api.data.PackInfo;
import com.izforge.izpack.api.data.ParsableFile;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Utils;
import zleurtor.scanmanager.model.events.ArtifactRepositoryManagerConnectorEvent;
import zleurtor.scanmanager.model.events.ArtifactRepositoryManagerConnectorListener;
import zleurtor.scanmanager.model.events.MasterUpdaterEvent;
import zleurtor.scanmanager.model.events.MasterUpdaterListener;

/**
 * The class having in charge to search the updates and apply it (if needed).
 *
 * @author Zleurtor
 */
public class MasterUpdater extends AbstractUpdater implements ArtifactRepositoryManagerConnectorListener {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(MasterUpdater.class);

    /**
     * The application's configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    /**
     * All the objects listening for master updater events.
     */
    private final Collection<MasterUpdaterListener> listeners = new ArrayList<>();

    /**
     * Check if an update is available (and install it).
     *
     * @return true if an update has been done, false otherwise.
     */
    public final boolean update() {
        LOGGER.info("Current version: " + CONFIGURATION.getString("version"));

        final String archivaUrl = CONFIGURATION.getString("artifactRepositoryManager.url");
        final String groupId = CONFIGURATION.getString("master.groupId");
        final String artifactId = CONFIGURATION.getString("master.artifactId");
        final String classifier = CONFIGURATION.getString("master.classifier");
        final String packaging = CONFIGURATION.getString("master.packaging");

        try (ArtifactRepositoryManagerConnector connector = new ArchivaConnector(archivaUrl)) {
            connector.addArtifactRepositoryManagerConnectorListener(this);

            // Search the latest version
            final String latestVersion = connector.searchLatestVersion(groupId, artifactId);
            if (latestVersion != null) {
                LOGGER.info("Latest version found: " + latestVersion);
            } else {
                LOGGER.warn("No version found.");
                return false;
            }

            // Check if an update is needed
            if (Utils.compareVersion(CONFIGURATION.getString("version"), latestVersion) >= 0) {
                LOGGER.info("No update needed.");
                return false;
            }

            fireUpdateFoundEvent(new MasterUpdaterEvent(this));

            // Search the URL of the artifact
            final URL artifactUrl = connector.searchArtifactUrl(groupId, artifactId, latestVersion, classifier, packaging);

            if (artifactUrl != null) {
                LOGGER.info("Url for the master installer found: " + artifactUrl);
            } else {
                LOGGER.warn("No URL found for the master installer.");
                return false;
            }

            // Search the size of the artifact
            final int artifactSize = connector.searchArtifactSize(groupId, artifactId, latestVersion, classifier, packaging);

            if (artifactSize != -1) {
                LOGGER.info("Size of the master installer found: " + artifactSize);
            } else {
                LOGGER.warn("Size of the master installer not found.");
                return false;
            }

            // Download the artifact
            final File dest = File.createTempFile(artifactId + "-" + classifier, "." + packaging);
            connector.download(artifactUrl, artifactSize, dest);
            LOGGER.info("New version downloaded.");

            fireInstallingEvent(new MasterUpdaterEvent(this));

            // Compute patch work plan
            final Collection<String> toExtract = computePlan(dest);
            LOGGER.info("Patch plan computed.");

            // Extract the update files
            final List<File> unpackedFiles = new ArrayList<>();

            try (JarFile installer = new JarFile(dest);
                    ObjectInputStream packsIn = new ObjectInputStream(installer.getInputStream(installer.getEntry("resources/packs.info")));
                    ObjectInputStream varsIn = new ObjectInputStream(installer.getInputStream(installer.getEntry("resources/vars")))) {
                @SuppressWarnings("unchecked") // I know what I do
                final List<PackInfo> packsInfo = (List<PackInfo>) packsIn.readObject();
                final Properties vars = (Properties) varsIn.readObject();
                vars.put("LOG_PATH", Utils.getLogFile().getParent());

                for (final PackInfo packInfo : packsInfo) {
                    if (toExtract.contains(packInfo.getPack().getName())) {
                        unpackedFiles.addAll(unpack(installer.getInputStream(
                                installer.getEntry("resources/packs/pack-" + packInfo.getPack().getName())), packInfo, vars));
                    }
                }
            } finally {
                // Delete the artifact file
                if (!dest.delete()) {
                    dest.deleteOnExit();
                }

                cleanFiles(new File(System.getProperty("user.dir")), unpackedFiles);
            }
        } catch (final KeyManagementException | NoSuchAlgorithmException | KeyStoreException
                     | IOException | URISyntaxException | ClassNotFoundException e) {
            LOGGER.warn("Error while trying to update: " + e.getMessage(), e);
            return false;
        } finally {
            fireFinishedEvent(new MasterUpdaterEvent(this));
        }

        return true;
    }

    /**
     * Compare the content of the installer archive and the installed pack, then compute the packs to unpack.
     *
     * @param archive The new 3D Scan Manager version archive.
     * @return The collection of packs to unpack from the archive.
     * @throws IOException If the new version archive file can not be read.
     * @throws FileNotFoundException If the new version archive file does not exist.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws ClassNotFoundException If the Class of a serialized object cannot be found.
     */
    @Override
    protected final Collection<String> computePlan(final File archive)
            throws FileNotFoundException, IOException, NoSuchAlgorithmException, ClassNotFoundException {
        final Collection<String> plan = new ArrayList<>();

        try (JarFile installer = new JarFile(archive);
            ObjectInputStream in = new ObjectInputStream(installer.getInputStream(installer.getEntry("resources/packs.info")))) {
            @SuppressWarnings("unchecked") // I know what I do
            final List<PackInfo> packsInfo = (List<PackInfo>) in.readObject();

            for (final PackInfo info : packsInfo) {
                if (info.getPack().isRequired() || new File(info.getPack().getName()).getAbsoluteFile().isDirectory()) {
                    plan.add(info.getPack().getName());
                }
            }
        }

        return plan;
    }

    /**
     * Search the latest version on the artifact repository.
     *
     * @return The latest version found on the artifact repository.
     * @throws KeyStoreException If the TrustStrategy fails to load.
     * @throws NoSuchAlgorithmException If no Provider supports aTrustManagerFactorySpi implementation for the "trust all" strategy.
     * @throws KeyManagementException If the SSL context builder fails to initialize.
     * @throws IOException If the artifact repository manager connector fails to close.
     */
    public final String searchLatestVersion() throws KeyManagementException, NoSuchAlgorithmException, KeyStoreException, IOException {
        final String archivaUrl = CONFIGURATION.getString("artifactRepositoryManager.url");
        final String groupId = CONFIGURATION.getString("master.groupId");
        final String artifactId = CONFIGURATION.getString("master.artifactId");

        try (ArtifactRepositoryManagerConnector connector = new ArchivaConnector(archivaUrl)) {
            return connector.searchLatestVersion(groupId, artifactId);
        }
    }

    /**
     * Unpack the pack denoted by the specified pack info and using the specified vars if needed.
     *
     * @param in The input stream pointing on the pack to unpack.
     * @param packInfo The information on the pack to unpack.
     * @param vars The variables to use during unpacking process.
     * @return The list of unpacked files.
     * @throws IOException If an input/output error occurs during unpacking.
     * @throws FileNotFoundException If the target file cannot be created.
     */
    private List<File> unpack(final InputStream in, final PackInfo packInfo, final Properties vars)
            throws FileNotFoundException, IOException {
        final List<File> unpackedFiles = new ArrayList<>();

        for (final PackFile packFile : packInfo.getPackFilesMap().keySet()) {
            String targetPath = packFile.getTargetPath();

            final long length = packFile.length();

            // Check if the file is Parsable
            boolean parsable = false;
            for (final ParsableFile parsableFile : packInfo.getParsables()) {
                if (parsableFile.getPath().equals(targetPath)) {
                    parsable = true;
                }
            }

            // Evaluate vars in the file name
            targetPath = targetPath.replace("${INSTALL_PATH}", new File(System.getProperty("user.dir")).getAbsolutePath());
            for (final Object key : vars.keySet()) {
                targetPath = targetPath.replace("${" + key + "}", vars.getProperty(key.toString()));
            }

            if (packFile.isDirectory()) {
                LOGGER.debug("Creating directory" + targetPath);

                new File(targetPath).mkdirs();
            } else if (targetPath.contains("/conf/") && targetPath.endsWith(".properties")) {
                LOGGER.debug("Merging " + targetPath);

                // Read the configuration line by line, checking for the vars at each line then merge the configuration files
                try (ByteArrayOutputStream byteArrayOut = new ByteArrayOutputStream()) {
                    parse(in, length, byteArrayOut, vars);
                    try (ByteArrayInputStream byteArrayIn = new ByteArrayInputStream(byteArrayOut.toByteArray())) {
                        mergeConfigurations(byteArrayIn, targetPath);
                    }
                }
            } else if (parsable) {
                LOGGER.debug("Parsing " + targetPath);

                // Copy as a text file, line by line, checking for the vars at each line.
                try (OutputStream out = new FileOutputStream(targetPath)) {
                    parse(in, length, out, vars);
                }
            } else {
                LOGGER.debug("Copying " + targetPath);

                // Copy as binary file
                try (OutputStream out = new FileOutputStream(targetPath)) {
                    copy(in, length, out);
                }
            }

            unpackedFiles.add(new File(targetPath));
        }

        return unpackedFiles;
    }

    /**
     * Copy the content of the specified input stream (having the specified length) into the given output stream.
     *
     * @param in The input stream of which the content shall be copied.
     * @param length The length of the input stream.
     * @param out The output stream to which the data shall be copied.
     * @throws IOException If an input/output error occurs.
     */
    private void copy(final InputStream in, final long length, final OutputStream out)
            throws IOException {
        final int bufferSize = 1024;
        final byte[] buffer = new byte[bufferSize];

        int totalRead = 0;
        do {
            final int read = in.read(buffer, 0, (int) Math.min(length - totalRead, bufferSize));
            totalRead += read;
            out.write(buffer, 0, read);
        } while (totalRead < length);
    }

    /**
     * Parse the content of the specified input stream (having the specified length) then copy it into the given output stream.
     *
     * @param in The input stream of which the content shall be parsed.
     * @param length The length of the input stream.
     * @param out The output stream to which the parsed data shall be writen.
     * @param vars The vars used to parse the input data.
     * @throws IOException If an input/output error occurs.
     */
    private void parse(final InputStream in, final long length, final OutputStream out, final Properties vars)
            throws IOException {
        try (ByteArrayOutputStream byteArrayOut = new ByteArrayOutputStream()) {
            copy(in, length, byteArrayOut);

            try (ByteArrayInputStream byteArrayIn = new ByteArrayInputStream(byteArrayOut.toByteArray());
                    BufferedReader reader = new BufferedReader(new InputStreamReader(byteArrayIn));
                    PrintWriter writer = new PrintWriter(out)) {
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    line = line.replace("${INSTALL_PATH}", new File(System.getProperty("user.dir")).getAbsolutePath());
                    for (final Object key : vars.keySet()) {
                        line = line.replace("${" + key + "}", vars.getProperty(key.toString()));
                    }
                    writer.println(line);
                }
            }
        }
    }

    /**
     * Clean the installation directory deleting the old unused files.
     *
     * @param installationFile The installation directory.
     * @param unpackedFiles All the unpoacked files during update.
     */
    private void cleanFiles(final File installationFile, final List<File> unpackedFiles) {
        if (installationFile.isDirectory()) {
            for (final File f : installationFile.listFiles()) {
                cleanFiles(f, unpackedFiles);
            }
        } else if (!unpackedFiles.contains(installationFile)
                && installationFile.getName().endsWith(".jar")
                && !installationFile.getName().equals("uninstaller.jar")) {
            installationFile.delete();
        }
    }

    /**
     * Registers a new object listening for master updater events.
     *
     * @param listener The new object listening for master updater events to register.
     */
    public final void addMasterUpdaterListener(final MasterUpdaterListener listener) {
        listeners.add(listener);
    }

    /**
     * Fires a new update found event.
     *
     * @param evt The event object representing the master updater event.
     */
    private void fireUpdateFoundEvent(final MasterUpdaterEvent evt) {
        for (final MasterUpdaterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.updateFound(evt);
            });
            t.setName("UpdateFoundEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires a new update download start event.
     *
     * @param evt The event object representing the master updater event.
     */
    private void fireDownloadingEvent(final MasterUpdaterEvent evt) {
        for (final MasterUpdaterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.downloading(evt);
            });
            t.setName("DownloadingEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires a new update install start event.
     *
     * @param evt The event object representing the master updater event.
     */
    private void fireInstallingEvent(final MasterUpdaterEvent evt) {
        for (final MasterUpdaterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.installing(evt);
            });
            t.setName("InstallingEvent dispatcher Thread.");
            t.start();
        }
    }

    /**
     * Fires a new update finished event.
     *
     * @param evt The event object representing the master updater event.
     */
    private void fireFinishedEvent(final MasterUpdaterEvent evt) {
        for (final MasterUpdaterListener listener : listeners) {
            final Thread t = new Thread(() -> {
                listener.finished(evt);
            });
            t.setName("FinishedEvent dispatcher Thread.");
            t.start();
        }
    }

    @Override
    public final void downloadingArtifact(final ArtifactRepositoryManagerConnectorEvent evt) {
        fireDownloadingEvent(new MasterUpdaterEvent(this, evt.getArtifactSize(), evt.getDownloadedSize()));

    }

}
