/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.model.update;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.security.DigestException;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.GZIPInputStream;

import org.kamranzafar.jtar.TarEntry;
import org.kamranzafar.jtar.TarInputStream;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Utils;

/**
 * The class having in charge to search the updates and apply it (if needed).
 *
 * @author Zleurtor
 */
public class SlaveUpdater extends AbstractUpdater {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(SlaveUpdater.class);

    /**
     * The application's configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    @Override
    public final boolean update() {
        LOGGER.info("Current version: " + CONFIGURATION.getString("version"));

        final String archivaUrl = CONFIGURATION.getString("artifactRepositoryManager.url");
        final String groupId = CONFIGURATION.getString("slave.groupId");
        final String artifactId = CONFIGURATION.getString("slave.artifactId");
        final String classifier = CONFIGURATION.getString("slave.classifier");
        final String packaging = CONFIGURATION.getString("slave.packaging");

        try (ArtifactRepositoryManagerConnector connector = new ArchivaConnector(archivaUrl)) {
            // Search the latest version
            final String latestVersion = connector.searchLatestVersion(groupId, artifactId);
            if (latestVersion != null) {
                LOGGER.info("Latest version found: " + latestVersion);
            } else {
                LOGGER.warn("No version found.");
                return false;
            }

            // Check if an update is needed
            if (Utils.compareVersion(CONFIGURATION.getString("version"), latestVersion) >= 0) {
                LOGGER.info("No update needed.");
                return false;
            }

            // Search the URL of the artifact
            final URL artifactUrl = connector.searchArtifactUrl(groupId, artifactId, latestVersion, classifier, packaging);
            if (artifactUrl != null) {
                LOGGER.info("Url for the slave archive found: " + artifactUrl);
            } else {
                LOGGER.warn("No URL found for the slave archive.");
                return false;
            }

            // Download the artifact
            final File dest = File.createTempFile(artifactId + "-" + classifier, ".tar.gz");
            connector.download(artifactUrl, -1, dest);
            LOGGER.info("New version downloaded.");

            update(dest);
        } catch (final KeyManagementException | NoSuchAlgorithmException | KeyStoreException
                     | IOException | URISyntaxException | DigestException e) {
            LOGGER.warn("Error while trying to update: " + e.getMessage(), e);
            return false;
        }

        return true;
    }

    /**
     * Install the update contained in the specified update file.
     *
     * @param updateFile The file containing the update to install.
     * @throws FileNotFoundException If the update file cannot be found.
     * @throws IOException If an I/O error occurs during the update process.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws DigestException If a digest error occurs.
     */
    public final void update(final File updateFile)
            throws FileNotFoundException, IOException, NoSuchAlgorithmException, DigestException {
        // Compute patch work plan
        final Collection<String> toExtract = computePlan(updateFile);
        LOGGER.info("Patch plan computed.");

        // Extract the update files
        try (TarInputStream in = new TarInputStream(new GZIPInputStream(new FileInputStream(updateFile)))) {
            for (TarEntry entry = in.getNextEntry(); entry != null; entry = in.getNextEntry()) {
                if (toExtract.contains(entry.getName())) {
                    final File file = new File(entry.getName());
                    file.getAbsoluteFile().getParentFile().mkdirs();
                    try (OutputStream out = new FileOutputStream(file, false)) {
                        final int bufferSize = 1024;
                        final byte[] buffer = new byte[bufferSize];
                        for (int r = in.read(buffer); r != -1; r = in.read(buffer)) {
                            out.write(buffer, 0, r);
                        }
                    }
                } else if (entry.getName().equals("conf/configuration.properties")) {
                    // Merge the configuration files
                    mergeConfigurations(in, entry.getName());
                    LOGGER.info("Configurations merged.");
                }
            }
        }
        LOGGER.info("Update extracted.");

        // Delete the artifact
        Files.delete(updateFile.toPath());
    }

    /**
     * Compare the content of the archive and the actual files, then compute the files to extract.
     *
     * @param archive The new 3D Scan Manager version archive.
     * @return The collection of files to extract from the archive.
     * @throws IOException If the new version archive file can not be read.
     * @throws FileNotFoundException If the new version archive file does not exist.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws DigestException If a digest error occurs.
     */
    @Override
    protected final Collection<String> computePlan(final File archive)
            throws FileNotFoundException, IOException, NoSuchAlgorithmException, DigestException {
        final Collection<String> plan = new ArrayList<>();

        // Compute the archive content checksums
        final Map<String, String> archiveChecksums = computeArchiveChecksums(archive);
        final Map<String, String> currentChecksums = computeCurrentChecksums(archiveChecksums.keySet());

        for (final String entryname : archiveChecksums.keySet()) {
            // Exception: conf files shall not be patched
            if (!entryname.startsWith("conf/")) {
                if (!currentChecksums.containsKey(entryname)
                        || !currentChecksums.get(entryname).equals(archiveChecksums.get(entryname))) {
                    plan.add(entryname);
                }
            }
        }

        return plan;
    }

    /**
     * Compute the checksum of each entry in the specified archive.
     *
     * @param archive The archive of which the checksum of each entry shall be computed.
     * @return A map containing, as keys the entry names, as values the checksums.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws IOException If the new version archive file can not be read.
     * @throws FileNotFoundException If a digest error occurs.
     */
    protected Map<String, String> computeArchiveChecksums(final File archive)
            throws NoSuchAlgorithmException, IOException, FileNotFoundException {

        final Map<String, String> archiveChecksums = new HashMap<>();

        try (TarInputStream in = new TarInputStream(new GZIPInputStream(new FileInputStream(archive)))) {
            for (TarEntry entry = in.getNextEntry(); entry != null; entry = in.getNextEntry()) {
                final String checksum = computeChecksum(in);

                if (!entry.isDirectory()) {
                    archiveChecksums.put(entry.getName(), checksum);
                }
            }
        }

        return archiveChecksums;
    }

    /**
     * Compute the checksum of the specified files in the current installation.
     *
     * @param filenames The name of the files of which the checksum shall be computed.
     * @return A map containing, as keys the entry names, as values the checksums.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     * @throws IOException If the new version archive file can not be read.
     * @throws FileNotFoundException If a digest error occurs.
     */
    protected Map<String, String> computeCurrentChecksums(final Collection<String> filenames)
            throws NoSuchAlgorithmException, IOException, FileNotFoundException {

        final Map<String, String> currentChecksums = new HashMap<>();

        for (final String filename : filenames) {
            final File file = new File(filename);

            if (file.exists()) {
                try (InputStream in = new FileInputStream(file)) {
                    final String checksum = computeChecksum(in);

                    currentChecksums.put(filename, checksum);
                }
            }
        }

        return currentChecksums;
    }

    /**
     * Compute the checksum of the content of the specified input stream.
     *
     * @param in The input stream of which the checksum of the content shall be computed.
     * @return The checksum of the content of the specified input stream.
     * @throws IOException If the input stream can not be read.
     * @throws NoSuchAlgorithmException If no Provider supports MD5 or SHA-1 implementation.
     */
    private String computeChecksum(final InputStream in)
            throws IOException, NoSuchAlgorithmException {
        final int unsign = 0xff;
        final int bufferSize = 1024;
        final byte[] buffer = new byte[bufferSize];

        final MessageDigest md5 = MessageDigest.getInstance("MD5");
        final MessageDigest sha1 = MessageDigest.getInstance("SHA-1");

        final StringBuilder checksumBuilder = new StringBuilder();

        for (int read = in.read(buffer); read != -1; read = in.read(buffer)) {
            md5.update(buffer, 0, read);
            sha1.update(buffer, 0, read);
        }

        final byte[] md5Digest = md5.digest();
        final byte[] sha1Digest = sha1.digest();

        for (final byte b : md5Digest) {
            String bb = Integer.toHexString(b & unsign);
            if (bb.length() < 2) {
                bb = "0" + bb;
            }
            checksumBuilder.append(bb);
        }
        for (final byte b : sha1Digest) {
            String bb = Integer.toHexString(b & unsign);
            if (bb.length() < 2) {
                bb = "0" + bb;
            }
            checksumBuilder.append(bb);
        }
        return checksumBuilder.toString();
    }

}
