/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.management.ManagementFactory;
import java.lang.management.RuntimeMXBean;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.control.ConfiguratorAction;
import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.model.update.MasterUpdater;
import zleurtor.scanmanager.view.gui.AboutButton;
import zleurtor.scanmanager.view.gui.ColorsButton;
import zleurtor.scanmanager.view.gui.ContentPanel;
import zleurtor.scanmanager.view.gui.GuiUtils;
import zleurtor.scanmanager.view.gui.LogButton;
import zleurtor.scanmanager.view.gui.SaveButton;
import zleurtor.scanmanager.view.gui.ShutdownButton;
import zleurtor.scanmanager.view.gui.SlavesPanel;
import zleurtor.scanmanager.view.gui.SnapshotButton;
import zleurtor.scanmanager.view.gui.SoundButton;
import zleurtor.scanmanager.view.gui.TrashButton;
import zleurtor.scanmanager.view.gui.UpdateButton;
import zleurtor.scanmanager.view.gui.UpdateDialog;

/**
 * The master application launcher.
 *
 * @author Zleurtor
 */
public final class Launcher {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(Launcher.class);

    /**
     * The default width of the main frame.
     */
    private static final int DEFAULT_WIDTH = 960;

    /**
     * The default height of the main frame.
     */
    private static final int DEFAULT_HEIGHT = 720;

    /**
     * The master.
     */
    private static Master master;

    /**
     * A private constructor for this utility class.
     */
    private Launcher() {
        super();
    }

    /**
     * Check if the application is running with elevated privileges. If no, restart with elevated privileges.
     *
     * @param args The application arguments (used to restart the application).
     */
    private void checkPrivilege(final String[] args) {
        if (!isPrivileged()) {
            try {
                final RuntimeMXBean runtimeMxBean = ManagementFactory.getRuntimeMXBean();
                final List<String> jvmArgs = runtimeMxBean.getInputArguments();

                final List<String> jvmArgsToRemove = new ArrayList<>();
                for (final String arg : jvmArgs) {
                    if (arg.startsWith("-agentlib") || arg.startsWith("-javaagent")) {
                        jvmArgsToRemove.add(arg);
                    }
                }

                final List<String> command = new ArrayList<>();
                command.add("wscript.exe");
                command.add(extractElevateScript());
                command.add(javaExecutable());
                command.addAll(jvmArgs);
                command.add("-Dprivileged=true");
                command.add("-classpath");
                command.add(javaClassPath());
                command.add("-splash:resources/" + ResourcesManager.getInstance().getString(ResourcesManager.KEY_ICON));
                command.add(this.getClass().getName());
                for (final String arg : args) {
                    command.add(arg);
                }

                command.removeAll(jvmArgsToRemove);

                final ProcessBuilder builder = new ProcessBuilder(command);
                final Process p = builder.start();
                p.waitFor();
            } catch (final IOException | InterruptedException e) {
                LOGGER.error("Cannot elevate privileges: " + e.getMessage(), e);
            }
            System.exit(0);
        }
    }

    /**
     * Checks whether or not the application is running with elevated privileges.
     *
     * @return True if the application is running with elevated privileges, false otherwise.
     */
    private boolean isPrivileged() {
        return Boolean.parseBoolean(System.getProperty("privileged"));
    }

    /**
     * Delete the old (LGPL) license file.
     */
    private void deleteOldLicense() {
        final File copyingLesser = new File("COPYING.LESSER");
        if (copyingLesser.exists()) {
            try {
                Files.delete(copyingLesser.toPath());
            } catch (final IOException e) {
                LOGGER.warn(e.getMessage(), e);
            }
        }
    }

    /**
     * Extract the elevate.js script to a temporary directory.
     *
     * @return The absolute path of the extracted elevate.js script.
     * @throws IOException If an Input/Output error occurs.
     */
    private String extractElevateScript() throws IOException {
        final File script = File.createTempFile("elevate", ".js");
        script.deleteOnExit();

        try (InputStream in = this.getClass().getResourceAsStream("/scripts/elevate.js");
                OutputStream out = new FileOutputStream(script)) {
            final int bufferSize = 2048;
            final byte[] buffer = new byte[bufferSize];
            int bytesRead;
            while ((bytesRead = in.read(buffer)) >= 0) {
                out.write(buffer, 0, bytesRead);
            }
        }

        return script.getAbsolutePath();
    }

    /**
     * Returns the java executable absolute path.
     *
     * @return The java executable absolute path.
     */
    private String javaExecutable() {
        return System.getProperty("java.home") + File.separator + "bin" + File.separator + "javaw.exe";
    }

    /**
     * Returns the actual java class path.
     *
     * @return The actual java class path.
     */
    private String javaClassPath() {
        return System.getProperty("java.class.path");
    }

    /**
     * Restarts the master.
     */
    private void restart() {
        try {
            String command = System.getProperty("user.dir").replace("/", File.separator);
            if (!command.endsWith(File.separator)) {
                command += File.separator;
            }
            command += "master.bat";

            final Process p = Runtime.getRuntime().exec(command, null, new File(System.getProperty("user.dir")));
            p.waitFor();
        } catch (final IOException | InterruptedException e) {
            LOGGER.error("Cannot elevate privileges: " + e.getMessage(), e);
        }
        System.exit(0);
    }

    /**
     * The master application entry point.
     *
     * @param args The master application arguments.
     */
    public static void main(final String[] args) {
        LOGGER.info("Logger initialized");

        // Check the privileges
        final Launcher launcher = new Launcher();
        launcher.checkPrivilege(args);

        // Hook to delete old license
        launcher.deleteOldLicense();

        // Set UI
        try {
            UIManager.setLookAndFeel("zleurtor.scanmanager.view.plaf.ScanManagerLookAndFeel");
        } catch (final ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }

        // Check for updates
        final UpdateDialog updateDialog = new UpdateDialog();
        final MasterUpdater updater = new MasterUpdater();
        updater.addMasterUpdaterListener(updateDialog);
        if (updater.update()) {
            launcher.restart();
        }

        // Create the master
        try {
            master = new Master();

            final Configuration configuration = Configuration.getInstance();
            int width = DEFAULT_WIDTH;
            int height = DEFAULT_HEIGHT;
            if (configuration.getString("frame.width") != null) {
                width = configuration.getInt("frame.width");
            }
            if (configuration.getString("frame.height") != null) {
                height = configuration.getInt("frame.height");
            }

            final int gap = 5;

            final JFrame frame = new JFrame(ResourcesManager.getInstance().getString(
                    ResourcesManager.KEY_TITLE,
                    Configuration.getInstance().getString("version")));
            frame.addComponentListener(new FrameListener());

            // initialize options field
            final JTextField optionsField = new JTextField();

            // Initialize slaves panel
            final SlavesPanel slavesPanel = new SlavesPanel(master, optionsField);
            final JScrollPane scrollPane =
                    new JScrollPane(slavesPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
            scrollPane.addComponentListener(slavesPanel);
            scrollPane.getVerticalScrollBar().setOpaque(false);
            scrollPane.getHorizontalScrollBar().setOpaque(false);
            master.addMasterListener(slavesPanel);
            master.scan();

            // Initialize configuration button
            final JPanel fieldPanel = new JPanel(new GridBagLayout());
            final JButton configuratorButton = new JButton(new ConfiguratorAction(master, slavesPanel, optionsField));
            fieldPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_SNAPSHOTOPTIONS)),
                    new GridBagConstraints(0, 0, 1, 1, 0, 0,
                            GridBagConstraints.WEST, GridBagConstraints.NONE,
                            new Insets(0, 0, 0, gap / 2), 0, 0));
            fieldPanel.add(optionsField,
                    new GridBagConstraints(1, 0, 1, 1, 1, 0,
                            GridBagConstraints.WEST, GridBagConstraints.BOTH,
                            new Insets(0, gap / 2, 0, gap / 2), 0, 0));
            fieldPanel.add(configuratorButton,
                    new GridBagConstraints(2, 0, 1, 1, 0, 0,
                            GridBagConstraints.EAST, GridBagConstraints.BOTH,
                            new Insets(0, gap / 2, 0, 0), 0, 0));
            fieldPanel.setOpaque(false);

            // Initialize buttons
            final JButton snapshotButton = new SnapshotButton(master, slavesPanel);
            final SaveButton saveButton = new SaveButton(master);
            final JButton trashButton = new TrashButton(master, slavesPanel);
            final JButton shutdownButton = new ShutdownButton(master, slavesPanel);
            final JButton updateButton = new UpdateButton(master, slavesPanel);
            final JButton aboutButton = new AboutButton();
            final JButton logButton = new LogButton(slavesPanel);
            final JButton soundButton = new SoundButton();
            final JButton colorsButton = new ColorsButton();
            saveButton.setSlavesPanel(slavesPanel);

            // Initialize toolbar
            final JToolBar toolbar = new JToolBar(JToolBar.HORIZONTAL);
            final Dimension separatorSize = new Dimension(
                    ResourcesManager.getInstance().getInt(ResourcesManager.KEY_SEPARATOR_WIDTH),
                    (int) Math.round(toolbar.getPreferredSize().height
                            * ResourcesManager.getInstance().getDouble(ResourcesManager.KEY_SEPARATOR_RATIO)));
            toolbar.setFloatable(false);
            toolbar.setLayout(new FlowLayout(FlowLayout.LEADING, gap, 0));
            toolbar.add(snapshotButton);
            toolbar.add(saveButton);
            toolbar.add(trashButton);
            toolbar.add(shutdownButton);
            toolbar.add(updateButton);
            toolbar.add(aboutButton);
            JSeparator separator = new JSeparator(JSeparator.VERTICAL);
            separator.setPreferredSize(separatorSize);
            separator.setMinimumSize(separatorSize);
            separator.setMaximumSize(separatorSize);
            toolbar.add(separator);
            toolbar.add(logButton);
            separator = new JSeparator(JSeparator.VERTICAL);
            separator.setPreferredSize(separatorSize);
            separator.setMinimumSize(separatorSize);
            separator.setMaximumSize(separatorSize);
            toolbar.add(separator);
            toolbar.add(soundButton);
            toolbar.add(colorsButton);

            // Initialize frame
            frame.setIconImage(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_ICON));
            frame.setContentPane(new ContentPanel());
            frame.setLayout(new GridBagLayout());
            frame.add(toolbar,
                    new GridBagConstraints(0, 0, 1, 1, 1, 0,
                            GridBagConstraints.CENTER, GridBagConstraints.NONE,
                            new Insets(-gap, 0, gap / 2, gap / 2), 0, 0));
            frame.add(fieldPanel,
                    new GridBagConstraints(0, 1, 1, 1, 1, 0,
                            GridBagConstraints.CENTER, GridBagConstraints.BOTH,
                            new Insets(gap / 2, gap, gap / 2, gap), 0, 0));
            frame.add(scrollPane,
                    new GridBagConstraints(0, 2, 1, 1, 1, 1,
                            GridBagConstraints.CENTER, GridBagConstraints.BOTH,
                            new Insets(gap / 2, gap, gap, gap), 0, 0));

            frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
            frame.setSize(new Dimension(width, height));
            frame.setLocationRelativeTo(null);
            frame.setVisible(true);

            addWindowListsner(frame);

            slavesPanel.requestFocus();
        } catch (final IOException e) {
            JOptionPane.showMessageDialog(null,
                    e.getMessage(),
                    "Error",
                    JOptionPane.ERROR_MESSAGE);
            System.exit(-1);
        }
    }

    /**
     * Add a window listener to the main frame.
     *
     * @param frame The frame to which a listener shall be added.
     */
    private static void addWindowListsner(final JFrame frame) {
        frame.addWindowListener(new WindowListener() {

            @Override
            public void windowOpened(final WindowEvent evt) { }

            @Override
            public void windowIconified(final WindowEvent evt) { }

            @Override
            public void windowDeiconified(final WindowEvent evt) { }

            @Override
            public void windowDeactivated(final WindowEvent evt) { }

            @Override
            public void windowClosing(final WindowEvent evt) {
                if (JOptionPane.showConfirmDialog(frame,
                        ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_REALLYQUIT),
                        ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_CONFIRM),
                        JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
                    System.exit(0);
                }
            }

            @Override
            public void windowClosed(final WindowEvent evt) { }

            @Override
            public void windowActivated(final WindowEvent evt) { }
        });
    }

    /**
     * The frame listener listening for dimension modification.
     *
     * @author Zleurtor
     */
    private static class FrameListener implements ComponentListener {

        @Override
        public void componentResized(final ComponentEvent evt) {
            Configuration.getInstance().setString("frame.width", Integer.toString(evt.getComponent().getWidth()));
            Configuration.getInstance().setString("frame.height", Integer.toString(evt.getComponent().getHeight()));
        }

        @Override
        public void componentMoved(final ComponentEvent evt) { }

        @Override
        public void componentShown(final ComponentEvent evt) { }

        @Override
        public void componentHidden(final ComponentEvent evt) { }
    }

}
