/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view;

import java.awt.Color;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * The class to handle the UI resources.
 *
 * @author Zleurtor
 */
public class ResourcesManager {

    /**
     * The title key.
     */
    public static final String KEY_TITLE = "title";

    /**
     * The save dialog title key.
     */
    public static final String KEY_TITLE_SAVEDIALOG = "title.saveDialog";

    /**
     * The configurator title key.
     */
    public static final String KEY_TITLE_CONFIGURATOR = "title.configurator";

    /**
     * The information title key.
     */
    public static final String KEY_TITLE_INFO = "title.info";

    /**
     * The warning title key.
     */
    public static final String KEY_TITLE_WARNING = "title.warning";

    /**
     * The trash title key.
     */
    public static final String KEY_TITLE_TRASH = "title.trash";

    /**
     * The error title key.
     */
    public static final String KEY_TITLE_ERROR = "title.error";

    /**
     * The confirm title key.
     */
    public static final String KEY_TITLE_CONFIRM = "title.confirm";

    /**
     * The input title key.
     */
    public static final String KEY_TITLE_INPUT = "title.input";

    /**
     * The about title key.
     */
    public static final String KEY_TITLE_ABOUT = "title.about";

    /**
     * The snapshot options title key.
     */
    public static final String KEY_TITLE_SNAPSHOTOPTIONS = "title.snapshotOptions";

    /**
     * The configurator image part title key.
     */
    public static final String KEY_TITLE_CONFIGURATOR_IMAGE = "title.configurator.image";

    /**
     * The configurator colors part title key.
     */
    public static final String KEY_TITLE_CONFIGURATOR_COLORS = "title.configurator.colors";

    /**
     * The log panel title key.
     */
    public static final String KEY_TITLE_LOGDIALOG = "title.logDialog";

    /**
     * The update dialog title key.
     */
    public static final String KEY_TITLE_UPDATEDIALOG = "title.updateDialog";

    /**
     * The snapshots dialog title key.
     */
    public static final String KEY_TITLE_SNAPSHOTSDIALOG = "title.snapshotsDialog";

    /**
     * The wait dialog title key.
     */
    public static final String KEY_TITLE_WAITUPDATEDIALOG = "title.waitUpdateDialog";

    /**
     * The main icon key.
     */
    public static final String KEY_ICON = "icon";

    /**
     * The scan icon key.
     */
    public static final String KEY_ICON_SCAN = "icon.scan";

    /**
     * The snapshot icon key.
     */
    public static final String KEY_ICON_SNAPSHOT = "icon.snapshot";

    /**
     * The modelize icon key.
     */
    public static final String KEY_ICON_MODELIZE = "icon.modelize";

    /**
     * The save icon key.
     */
    public static final String KEY_ICON_SAVE = "icon.save";

    /**
     * The trash icon key.
     */
    public static final String KEY_ICON_TRASH = "icon.trash";

    /**
     * The shutdown icon key.
     */
    public static final String KEY_ICON_SHUTDOWN = "icon.shutdown";

    /**
     * The about icon key.
     */
    public static final String KEY_ICON_ABOUT = "icon.about";

    /**
     * The update icon key.
     */
    public static final String KEY_ICON_UPDATE = "icon.update";

    /**
     * The log icon key.
     */
    public static final String KEY_ICON_LOGBUTTON = "icon.logButton";

    /**
     * The shutdown icon key.
     */
    public static final String KEY_ICON_SHUTDOWNBUTTON = "icon.shutdownButton";

    /**
     * The log icon key.
     */
    public static final String KEY_ICON_LOG = "icon.log";

    /**
     * The sound on icon key.
     */
    public static final String KEY_ICON_SOUNDON = "icon.soundOnButton";

    /**
     * The sound off icon key.
     */
    public static final String KEY_ICON_SOUNDOFF = "icon.soundOffButton";

    /**
     * The polychrome icon key.
     */
    public static final String KEY_ICON_POLYCHROME = "icon.polychrome";

    /**
     * The monochrome icon key.
     */
    public static final String KEY_ICON_MONOCHROME = "icon.monochrome";

    /**
     * The previous icon key.
     */
    public static final String KEY_BUTTON_PREVIOUS = "icon.previous";

    /**
     * The next icon key.
     */
    public static final String KEY_BUTTON_NEXT = "icon.next";

    /**
     * The background color key.
     */
    public static final String KEY_COLOR_BACKGROUND = "color.background";

    /**
     * The scan tooltip key.
     */
    public static final String KEY_TOOLTIP_SCAN = "tooltip.scan";

    /**
     * The snapshot tooltip key.
     */
    public static final String KEY_TOOLTIP_SNAPSHOT = "tooltip.snapshot";

    /**
     * The modelize tooltip key.
     */
    public static final String KEY_TOOLTIP_MODELIZE = "tooltip.modelize";

    /**
     * The save tooltip key.
     */
    public static final String KEY_TOOLTIP_SAVE = "tooltip.save";

    /**
     * The trash tooltip key.
     */
    public static final String KEY_TOOLTIP_TRASH = "tooltip.trash";

    /**
     * The shutdown tooltip key.
     */
    public static final String KEY_TOOLTIP_SHUTDOWN = "tooltip.shutdown";

    /**
     * The about tooltip key.
     */
    public static final String KEY_TOOLTIP_ABOUT = "tooltip.about";

    /**
     * The update tooltip key.
     */
    public static final String KEY_TOOLTIP_UPDATE = "tooltip.update";

    /**
     * The sound tooltip key.
     */
    public static final String KEY_TOOLTIP_SOUND = "tooltip.sound";

    /**
     * The log tooltip key.
     */
    public static final String KEY_TOOLTIP_LOG = "tooltip.log";

    /**
     * The chromacity tooltip key.
     */
    public static final String KEY_TOOLTIP_CHROMACITY = "tooltip.chromacity";

    /**
     * The snapshot error image key.
     */
    public static final String KEY_IMAGE_SNAPSHOTERROR = "image.snapshotError";

    /**
     * The snapshot empty image key.
     */
    public static final String KEY_IMAGE_SNAPSHOTEMPTY = "image.snapshotEmpty";

    /**
     * The stream image key.
     */
    public static final String KEY_IMAGE_STREAM = "image.stream";

    /**
     * The wait image key.
     */
    public static final String KEY_IMAGE_WAIT = "image.wait";

    /**
     * The background image key.
     */
    public static final String KEY_IMAGE_BACKGROUND = "image.background";

    /**
     * The border top image key.
     */
    public static final String KEY_IMAGE_BORDER_TOP = "image.border.top";

    /**
     * The border bottom image key.
     */
    public static final String KEY_IMAGE_BORDER_BOTTOM = "image.border.bottom";

    /**
     * The border left image key.
     */
    public static final String KEY_IMAGE_BORDER_LEFT = "image.border.left";

    /**
     * The border right image key.
     */
    public static final String KEY_IMAGE_BORDER_RIGHT = "image.border.right";

    /**
     * The border top left image key.
     */
    public static final String KEY_IMAGE_BORDER_TOPLEFT = "image.border.topLeft";

    /**
     * The border top right image key.
     */
    public static final String KEY_IMAGE_BORDER_TOPRIGHT = "image.border.topRight";

    /**
     * The border bottom left image key.
     */
    public static final String KEY_IMAGE_BORDER_BOTTOMLEFT = "image.border.bottomLeft";

    /**
     * The border bottom right image key.
     */
    public static final String KEY_IMAGE_BORDER_BOTTOMRIGHT = "image.border.bottomRight";

    /**
     * The toolbar border bottom image key.
     */
    public static final String KEY_IMAGE_TOOLBAR_BORDER_BOTTOM = "image.toolbar.border.bottom";

    /**
     * The toolbar border left image key.
     */
    public static final String KEY_IMAGE_TOOLBAR_BORDER_LEFT = "image.toolbar.border.left";

    /**
     * The toolbar border right image key.
     */
    public static final String KEY_IMAGE_TOOLBAR_BORDER_RIGHT = "image.toolbar.border.right";

    /**
     * The GPL v3 image key.
     */
    public static final String KEY_IMAGE_GPLV3 = "image.gplv3";

    /**
     * The trash confirm text key.
     */
    public static final String KEY_CONFIRM_TRASH = "confirm.trash";

    /**
     * The overwrite confirm text key.
     */
    public static final String KEY_CONFIRM_OVERWRITE = "confirm.overwrite";

    /**
     * The saving file text key.
     */
    public static final String KEY_TEXT_SAVINGFILES = "text.savingFiles";

    /**
     * The zip description text key.
     */
    public static final String KEY_TEXT_ZIPDESCRIPTION = "text.zipDescription";

    /**
     * The snapshot option text key.
     */
    public static final String KEY_TEXT_SNAPSHOTOPTIONS = "text.snapshotOptions";

    /**
     * The stream error text key.
     */
    public static final String KEY_TEXT_STREAMERROR = "text.streamError";

    /**
     * The quality text key.
     */
    public static final String KEY_TEXT_QUALITY = "text.quality";

    /**
     * The timeout text key.
     */
    public static final String KEY_TEXT_TIMEOUT = "text.timeout";

    /**
     * The encoding text key.
     */
    public static final String KEY_TEXT_ENCODING = "text.encoding";

    /**
     * The sharpness text key.
     */
    public static final String KEY_TEXT_SHARPNESS = "text.sharpness";

    /**
     * The contrast text key.
     */
    public static final String KEY_TEXT_CONTRAST = "text.contrast";

    /**
     * The brightness text key.
     */
    public static final String KEY_TEXT_BRIGHTNESS = "text.brightness";

    /**
     * The saturation text key.
     */
    public static final String KEY_TEXT_SATURATION = "text.saturation";

    /**
     * The EV compensation text key.
     */
    public static final String KEY_TEXT_EVCOMPENSATION = "text.evCompensation";

    /**
     * The exposure text key.
     */
    public static final String KEY_TEXT_EXPOSURE = "text.exposure";

    /**
     * The AWB text key.
     */
    public static final String KEY_TEXT_AWB = "text.awb";

    /**
     * The shutter speed text key.
     */
    public static final String KEY_TEXT_SHUTTERSPEED = "text.shutterspeed";

    /**
     * The red shift text key.
     */
    public static final String KEY_TEXT_RED = "text.red";

    /**
     * The green shift text key.
     */
    public static final String KEY_TEXT_GREEN = "text.green";

    /**
     * The blue shift text key.
     */
    public static final String KEY_TEXT_BLUE = "text.blue";

    /**
     * The unsupported options text key.
     */
    public static final String KEY_TEXT_UNSUPPORTEDOPTIONS = "text.unsupportedOptions";

    /**
     * The downloading update text key.
     */
    public static final String KEY_TEXT_DOWNLOADINGUPDATE = "text.downloadingUpdate";

    /**
     * The installing update text key.
     */
    public static final String KEY_TEXT_INSTALLINGUPDATE = "text.installingUpdate";

    /**
     * The log level success text key.
     */
    public static final String KEY_TEXT_LOGLEVELERROR = "text.logLevelError";

    /**
     * The log level error text key.
     */
    public static final String KEY_TEXT_LOGLEVELSUCCESS = "text.logLevelSuccess";

    /**
     * The really quit confirm text key.
     */
    public static final String KEY_TEXT_REALLYQUIT = "text.reallyQuit";

    /**
     * The loading log text key.
     */
    public static final String KEY_TEXT_LOADINGLOG = "text.loadingLog";

    /**
     * The awaiting snapshots text key.
     */
    public static final String KEY_TEXT_AWAITINGSNAPSHOTS = "text.awaitingSnapshots";

    /**
     * The no slave text key.
     */
    public static final String KEY_TEXT_NOSLAVE = "text.noSlave";

    /**
     * The update error text key.
     */
    public static final String KEY_TEXT_UPDATEERROR = "text.updateError";

    /**
     * The no update found text key.
     */
    public static final String KEY_TEXT_NOUPDATE = "text.noUpdate";

    /**
     * The update slaves text key.
     */
    public static final String KEY_TEXT_UPDATESLAVES = "text.updateSlaves";

    /**
     * The wait updates text key.
     */
    public static final String KEY_TEXT_WAITUPDATEDIALOG = "text.waitUpdateDialog";

    /**
     * The profiles text key.
     */
    public static final String KEY_TEXT_PROFILES = "text.profiles";

    /**
     * The profile not found text key.
     */
    public static final String KEY_TEXT_PROFILENOTFOUND = "text.profileNotFound";

    /**
     * The no pipe in profile name text key.
     */
    public static final String KEY_TEXT_NOPIPEINPROFILE = "text.noPipeInProfile";

    /**
     * The download/local update file input message key.
     */
    public static final String KEY_TEXT_INPUTDOWNLOADLOCAL = "text.inputDownloadLocal";

    /**
     * The about text key.
     */
    public static final String KEY_TEXT_ABOUT = "text.about";

    /**
     * The reboot/restart input message key.
     */
    public static final String KEY_TEXT_INPUTREBOOTSHUTDOWN = "text.inputRebootShutdown";

    /**
     * The enter profile name message key.
     */
    public static final String KEY_TEXT_ENTERPROFILENAME = "text.enterProfileName";

    /**
     * The no snapshot message key.
     */
    public static final String KEY_TEXT_NOSNAPSHOT = "text.noSnapshot";

    /**
     * The image reading error message key.
     */
    public static final String KEY_TEXT_IMAGEREADINGERROR = "text.imageReadingError";

    /**
     * The saving message key.
     */
    public static final String KEY_TEXT_SAVING = "text.saving";

    /**
     * The configurator button text key.
     */
    public static final String KEY_BUTTON_CONFIGURATOR = "button.configurator";

    /**
     * The cancel button text key.
     */
    public static final String KEY_BUTTON_CANCEL = "button.cancel";

    /**
     * The delete button text key.
     */
    public static final String KEY_BUTTON_DELETE = "button.delete";

    /**
     * The apply to selected button text key.
     */
    public static final String KEY_BUTTON_APPLYTOSELECTED = "button.applyToSelected";

    /**
     * The apply all button text key.
     */
    public static final String KEY_BUTTON_APPLYALL = "button.applyAll";

    /**
     * The reset button text key.
     */
    public static final String KEY_BUTTON_RESET = "button.reset";

    /**
     * The select all button text key.
     */
    public static final String KEY_BUTTON_SELECTALL = "button.selectAll";

    /**
     * The deselect all button text key.
     */
    public static final String KEY_BUTTON_DESELECTALL = "button.deselectAll";

    /**
     * The copy to clipboard button text key.
     */
    public static final String KEY_BUTTON_COPYTOCLIPBOARD = "button.copyToClipboard";

    /**
     * The load profile button text key.
     */
    public static final String KEY_BUTTON_LOAD_PROFILE = "button.loadProfile";

    /**
     * The save profile button text key.
     */
    public static final String KEY_BUTTON_SAVEPROFILE = "button.saveProfile";

    /**
     * The delete profile button text key.
     */
    public static final String KEY_BUTTON_DELETEPROFILE = "button.deleteProfile";

    /**
     * The reboot option text key.
     */
    public static final String KEY_INPUT_REBOOT = "input.reboot";

    /**
     * The shutdown option text key.
     */
    public static final String KEY_INPUT_SHUTDOWN = "input.shutdown";

    /**
     * The reboot selected option text key.
     */
    public static final String KEY_INPUT_REBOOTSELECTED = "input.rebootSelected";

    /**
     * The shutdown selected option text key.
     */
    public static final String KEY_INPUT_SHUTDOWNSELECTED = "input.shutdownSelected";

    /**
     * The download update option text key.
     */
    public static final String KEY_INPUT_DOWNLOAD_UPDATE = "input.downloadUpdate";

    /**
     * The local update file option text key.
     */
    public static final String KEY_INPUT_LOCAL_UPDATE_FILE = "input.localUpdateFile";

    /**
     * quack sound key.
     */
    public static final String KEY_SOUND_QUACK = "sound.quack";

    /**
     * The panel background UI key.
     */
    public static final String KEY_PANEL_BACKGROUND = "Panel.background";

    /**
     * The panel foreground UI key.
     */
    public static final String KEY_PANEL_FOREGROUND = "Panel.foreground";

    /**
     * The button foreground UI key.
     */
    public static final String KEY_BUTTON_BACKGROUND = "Button.background";

    /**
     * The button foreground UI key.
     */
    public static final String KEY_BUTTON_FOREGROUND = "Button.foreground";

    /**
     * The button border UI key.
     */
    public static final String KEY_BUTTON_BORDER = "Button.border";

    /**
     * The button border insets top UI key.
     */
    public static final String KEY_BUTTON_BORDERINSETS_TOP = "Button.borderInsets.top";

    /**
     * The button border insets left UI key.
     */
    public static final String KEY_BUTTON_BORDERINSETS_LEFT = "Button.borderInsets.left";

    /**
     * The button border insets bottom UI key.
     */
    public static final String KEY_BUTTON_BORDERINSETS_BOTTOM = "Button.borderInsets.bottom";

    /**
     * The button border insets right UI key.
     */
    public static final String KEY_BUTTON_BORDERINSETS_RIGHT = "Button.borderInsets.right";

    /**
     * The button active color UI key.
     */
    public static final String KEY_BUTTON_ACTIVE = "Button.active";

    /**
     * The button foreground UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BACKGROUND = "ToggleButton.background";

    /**
     * The button foreground UI key.
     */
    public static final String KEY_TOGGLEBUTTON_FOREGROUND = "ToggleButton.foreground";

    /**
     * The button border UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BORDER = "ToggleButton.border";

    /**
     * The button border insets top UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BORDERINSETS_TOP = "ToggleButton.borderInsets.top";

    /**
     * The button border insets left UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BORDERINSETS_LEFT = "ToggleButton.borderInsets.left";

    /**
     * The button border insets bottom UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BORDERINSETS_BOTTOM = "ToggleButton.borderInsets.bottom";

    /**
     * The button border insets right UI key.
     */
    public static final String KEY_TOGGLEBUTTON_BORDERINSETS_RIGHT = "ToggleButton.borderInsets.right";

    /**
     * The label background UI key.
     */
    public static final String KEY_LABEL_BACKGROUND = "Label.background";

    /**
     * The label foreground UI key.
     */
    public static final String KEY_LABEL_FOREGROUND = "Label.foreground";

    /**
     * The text field background UI key.
     */
    public static final String KEY_TEXTFIELD_BACKGROUND = "TextField.background";

    /**
     * The text field foreground UI key.
     */
    public static final String KEY_TEXTFIELD_FOREGROUND = "TextField.foreground";

    /**
     * The text field selection background UI key.
     */
    public static final String KEY_TEXTFIELD_SELECTIONBACKGROUND = "TextField.selectionBackground";

    /**
     * The text field selection foreground UI key.
     */
    public static final String KEY_TEXTFIELD_SELECTIONFOREGROUND = "TextField.selectionForeground";

    /**
     * The text field caret foreground UI key.
     */
    public static final String KEY_TEXTFIELD_CARETFOREGROUND = "TextField.caretForeground";

    /**
     * The text field border UI key.
     */
    public static final String KEY_TEXTFIELD_BORDER = "TextField.border";

    /**
     * The text field border insets top UI key.
     */
    public static final String KEY_TEXTFIELD_BORDERINSETS_TOP = "TextField.borderInsets.top";

    /**
     * The text field border insets left UI key.
     */
    public static final String KEY_TEXTFIELD_BORDERINSETS_LEFT = "TextField.borderInsets.left";

    /**
     * The text field border insets bottom UI key.
     */
    public static final String KEY_TEXTFIELD_BORDERINSETS_BOTTOM = "TextField.borderInsets.bottom";

    /**
     * The text field border insets right UI key.
     */
    public static final String KEY_TEXTFIELD_BORDERINSETS_RIGHT = "TextField.borderInsets.right";

    /**
     * The tool bar background UI key.
     */
    public static final String KEY_TOOLBAR_BACKGROUND = "ToolBar.background";

    /**
     * The tool bar foreground UI key.
     */
    public static final String KEY_TOOLBAR_FOREGROUND = "ToolBar.foreground";

    /**
     * The option pane background UI key.
     */
    public static final String KEY_OPTIONPANE_BACKGROUND = "OptionPane.background";

    /**
     * The option pane foreground UI key.
     */
    public static final String KEY_OPTIONPANE_FOREGROUND = "OptionPane.foreground";

    /**
     * The option pane question icon UI key.
     */
    public static final String KEY_OPTIONPANE_QUESTIONICON = "OptionPane.questionIcon";

    /**
     * The option pane information icon UI key.
     */
    public static final String KEY_OPTIONPANE_INFORMATIONICON = "OptionPane.informationIcon";

    /**
     * The option pane error icon UI key.
     */
    public static final String KEY_OPTIONPANE_ERRORICON = "OptionPane.errorIcon";

    /**
     * The option pane warning icon UI key.
     */
    public static final String KEY_OPTIONPANE_WARNINGICON = "OptionPane.warningIcon";

    /**
     * The scroll bar background UI key.
     */
    public static final String KEY_SCROLLBAR_BACKGROUND = "ScrollBar.background";

    /**
     * The scroll bar foreground UI key.
     */
    public static final String KEY_SCROLLBAR_FOREGROUND = "ScrollBar.foreground";

    /**
     * The scroll bar arc UI key.
     */
    public static final String KEY_SCROLLBAR_ARC = "ScrollBar.arc";

    /**
     * The scroll bar thickness UI key.
     */
    public static final String KEY_SCROLLBAR_THICKNESS = "ScrollBar.thickness";

    /**
     * The scroll bar thickness UI key.
     */
    public static final String KEY_SCROLLBAR_INCREMENT = "ScrollBar.increment";

    /**
     * The spinner foreground UI key.
     */
    public static final String KEY_SPINNER_FOREGROUND = "Spinner.foreground";

    /**
     * The spinner border UI key.
     */
    public static final String KEY_SPINNER_BORDER = "Spinner.border";

    /**
     * The spinner border insets top UI key.
     */
    public static final String KEY_SPINNER_BORDERINSETS_TOP = "Spinner.borderInsets.top";

    /**
     * The spinner border insets left UI key.
     */
    public static final String KEY_SPINNER_BORDERINSETS_LEFT = "Spinner.borderInsets.left";

    /**
     * The spinner border insets bottom UI key.
     */
    public static final String KEY_SPINNER_BORDERINSETS_BOTTOM = "Spinner.borderInsets.bottom";

    /**
     * The spinner border insets right UI key.
     */
    public static final String KEY_SPINNER_BORDERINSETS_RIGHT = "Spinner.borderInsets.right";

    /**
     * The formatted text field background UI key.
     */
    public static final String KEY_FORMATTEDTEXTFIELD_BACKGROUND = "FormattedTextField.background";

    /**
     * The formatted text field foreground UI key.
     */
    public static final String KEY_FORMATTEDTEXTFIELD_FOREGROUND = "FormattedTextField.foreground";

    /**
     * The formatted text field selection background UI key.
     */
    public static final String KEY_FORMATTEDTEXTFIELD_SELECTIONBACKGROUND = "FormattedTextField.selectionBackground";

    /**
     * The formatted text field selection foreground UI key.
     */
    public static final String KEY_FORMATTEDTEXTFIELD_SELECTIONFOREGROUND = "FormattedTextField.selectionForeground";

    /**
     * The formatted text field caret foreground UI key.
     */
    public static final String KEY_FORMATTEDTEXTFIELD_CARETFOREGROUND = "FormattedTextField.caretForeground";

    /**
     * The combo box background UI key.
     */
    public static final String KEY_COMBOBOX_BACKGROUND = "ComboBox.background";

    /**
     * The combo box foreground UI key.
     */
    public static final String KEY_COMBOBOX_FOREGROUND = "ComboBox.foreground";

    /**
     * The combo box selection background UI key.
     */
    public static final String KEY_COMBOBOX_SELECTIONBACKGROUND = "ComboBox.selectionBackground";

    /**
     * The combo box selection foreground UI key.
     */
    public static final String KEY_COMBOBOX_SELECTIONFOREGROUND = "ComboBox.selectionForeground";

    /**
     * The combo box border UI key.
     */
    public static final String KEY_COMBOBOX_BORDER = "ComboBox.border";

    /**
     * The combo box border insets top UI key.
     */
    public static final String KEY_COMBOBOX_BORDERINSETS_TOP = "ComboBox.borderInsets.top";

    /**
     * The combo box border insets left UI key.
     */
    public static final String KEY_COMBOBOX_BORDERINSETS_LEFT = "ComboBox.borderInsets.left";

    /**
     * The combo box border insets bottom UI key.
     */
    public static final String KEY_COMBOBOX_BORDERINSETS_BOTTOM = "ComboBox.borderInsets.bottom";

    /**
     * The combo box border insets right UI key.
     */
    public static final String KEY_COMBOBOX_BORDERINSETS_RIGHT = "ComboBox.borderInsets.right";

    /**
     * The check box background UI key.
     */
    public static final String KEY_CHECKBOX_BACKGROUND = "CheckBox.background";

    /**
     * The check box foreground UI key.
     */
    public static final String KEY_CHECKBOX_FOREGROUND = "CheckBox.foreground";

    /**
     * The check box selected icon UI key.
     */
    public static final String KEY_CHECKBOX_SELECTEDICON = "CheckBox.selectedIcon";

    /**
     * The check box unselected icon UI key.
     */
    public static final String KEY_CHECKBOX_UNSELECTEDICON = "CheckBox.unselectedIcon";

    /**
     * The slaves panel selection UI key.
     */
    public static final String KEY_SLAVESPANEL_SELECTION = "SlavesPanel.selection";

    /**
     * The slider rule color UI key.
     */
    public static final String KEY_RULE_TRACKCOLOR = "Slider.trackColor";

    /**
     * The slider cursor color UI key.
     */
    public static final String KEY_SLIDER_THUMBCOLOR = "Slider.thumbColor";

    /**
     * The slider foreground color UI key.
     */
    public static final String KEY_SLIDER_FOREGROUND = "Slider.foreground";

    /**
     * The List background UI key.
     */
    public static final String KEY_LIST_BACKGROUND = "List.background";

    /**
     * The List foreground UI key.
     */
    public static final String KEY_LIST_FOREGROUND = "List.foreground";

    /**
     * The List selection background UI key.
     */
    public static final String KEY_LIST_SELECTIONBACKGROUND = "List.selectionBackground";

    /**
     * The List selection foreground UI key.
     */
    public static final String KEY_LIST_SELECTIONFOREGROUND = "List.selectionForeground";

    /**
     * The titled border color UI key.
     */
    public static final String KEY_TITLEDBORDER_COLOR = "TitledBorder.color";

    /**
     * The file view file icon UI key.
     */
    public static final String KEY_FILEVIEW_FILEICON = "FileView.fileIcon";

    /**
     * The file view floppy drive icon UI key.
     */
    public static final String KEY_FILEVIEW_FLOPPYDRIVEICON = "FileView.floppyDriveIcon";

    /**
     * The file view computer icon UI key.
     */
    public static final String KEY_FILEVIEW_COMPUTERICON = "FileView.computerIcon";

    /**
     * The file view hard drive icon UI key.
     */
    public static final String KEY_FILEVIEW_HARDDRIVEICON = "FileView.hardDriveIcon";

    /**
     * The file view directory icon UI key.
     */
    public static final String KEY_FILEVIEW_DIRECTORYICON = "FileView.directoryIcon";

    /**
     * The file view file icon UI key.
     */
    public static final String KEY_FILECHOOSER_UPFOLDERICON = "FileChooser.upFolderIcon";

    /**
     * The file view floppy drive icon UI key.
     */
    public static final String KEY_FILECHOOSER_NEWFOLDERICON = "FileChooser.newFolderIcon";

    /**
     * The file view computer icon UI key.
     */
    public static final String KEY_FILECHOOSER_HOMEFOLDERICON = "FileChooser.homeFolderIcon";

    /**
     * The file view hard drive icon UI key.
     */
    public static final String KEY_FILECHOOSER_LISTVIEWICON = "FileChooser.listViewIcon";

    /**
     * The file view directory icon UI key.
     */
    public static final String KEY_FILECHOOSER_DETAILSVIEWICON = "FileChooser.detailsViewIcon";

    /**
     * The log panel background UI key.
     */
    public static final String KEY_LOGPANEL_BAKCGROUND = "LogPanel.background";

    /**
     * The log panel info foreground UI key.
     */
    public static final String KEY_LOGPANEL_FOREGROUND_INFO = "LogPanel.foreground.info";

    /**
     * The log panel warn foreground UI key.
     */
    public static final String KEY_LOGPANEL_FOREGROUND_WARN = "LogPanel.foreground.warn";

    /**
     * The log panel error foreground UI key.
     */
    public static final String KEY_LOGPANEL_FOREGROUND_ERROR = "LogPanel.foreground.error";

    /**
     * The log panel debug foreground UI key.
     */
    public static final String KEY_LOGPANEL_FOREGROUND_DEBUG = "LogPanel.foreground.debug";

    /**
     * The log panel other foreground UI key.
     */
    public static final String KEY_LOGPANEL_FOREGROUND_OTHER = "LogPanel.foreground.other";

    /**
     * The log panel width UI key.
     */
    public static final String KEY_LOGPANEL_WIDTH = "LogPanel.width";

    /**
     * The log panel height UI key.
     */
    public static final String KEY_LOGPANEL_HEIGHT = "LogPanel.height";

    /**
     * The wait panel foreground UI key.
     */
    public static final String KEY_WAITPANEL_FOREGROUND = "waitPanel.foreground";

    /**
     * The wait panel margin UI key.
     */
    public static final String KEY_WAITPANEL_MARGIN = "waitPanel.margin";

    /**
     * The wait panel FPS UI key.
     */
    public static final String KEY_WAITPANEL_FPS = "waitPanel.fps";

    /**
     * The separator ratio UI key.
     */
    public static final String KEY_SEPARATOR_RATIO = "separator.ratio";

    /**
     * The separator foreground UI key.
     */
    public static final String KEY_SEPARATOR_FOREGROUND = "separator.foreground";

    /**
     * The separator width UI key.
     */
    public static final String KEY_SEPARATOR_WIDTH = "separator.width";

    /**
     * The tooltip background UI key.
     */
    public static final String KEY_TOOLTIP_BACKGROUND = "tooltip.background";

    /**
     * The tooltip foreground UI key.
     */
    public static final String KEY_TOOLTIP_FOREGROUND = "tooltip.foreground";

    /**
     * The tooltip border UI key.
     */
    public static final String KEY_TOOLTIP_BORDER = "tooltip.border";

    /**
     * The progress bar border UI key.
     */
    public static final String KEY_PROGRESSBAR_BORDER = "progressBar.border";

    /**
     * The progress bar foreground UI key.
     */
    public static final String KEY_PROGRESSBAR_FOREGROUND = "progressBar.foreground";

    /**
     * The progress bar border insets top UI key.
     */
    public static final String KEY_PROGRESSBAR_BORDERINSETS_TOP = "progressBar.borderInsets.top";

    /**
     * The progress bar border insets left UI key.
     */
    public static final String KEY_PROGRESSBAR_BORDERINSETS_LEFT = "progressBar.borderInsets.left";

    /**
     * The progress bar border insets bottom UI key.
     */
    public static final String KEY_PROGRESSBAR_BORDERINSETS_BOTTOM = "progressBar.borderInsets.bottom";

    /**
     * The progress bar border insets right UI key.
     */
    public static final String KEY_PROGRESSBAR_BORDERINSETS_RIGHT = "progressBar.borderInsets.right";

    /**
     * The tabbed pane background UI key.
     */
    public static final String KEY_TABBEDPANE_BACKGROUND = "tabbedPane.background";

    /**
     * The tabbed pane foreground UI key.
     */
    public static final String KEY_TABBEDPANE_FOREGROUND = "tabbedPane.foreground";

    /**
     * The tabbed pane highlight UI key.
     */
    public static final String KEY_TABBEDPANE_HIGHLIGHT = "tabbedPane.highlight";

    /**
     * The tabbed pane light UI key.
     */
    public static final String KEY_TABBEDPANE_LIGHT = "tabbedPane.light";

    /**
     * The tabbed pane shadow UI key.
     */
    public static final String KEY_TABBEDPANE_SHADOW = "tabbedPane.shadow";

    /**
     * The tabbed pane dark shadow UI key.
     */
    public static final String KEY_TABBEDPANE_DARKSHADOW = "tabbedPane.darkShadow";

    /**
     * The tabbed pane selected tab color UI key.
     */
    public static final String KEY_TABBEDPANE_TABSELECTEDCOLOR = "tabbedPane.tabSelectedColor";

    /**
     * The default name of the resource files.
     */
    private static final String DEFAULT_BASE_NAME = "resources";

    /**
     * The manager's unique instance.
     */
    private static final ResourcesManager INSTANCE = new ResourcesManager();

    /**
     * The hexadecimal basis.
     */
    private static final int HEXA = 16;

    /**
     * The bundle containing the properties of the resources file.
     */
    private final ResourceBundle bundle = PropertyResourceBundle.getBundle(DEFAULT_BASE_NAME);

    /**
     * Returns the manager's unique instance.
     *
     * @return The manager's unique instance.
     */
    public static final ResourcesManager getInstance() {
        return INSTANCE;
    }

    /**
     * Gets a string for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired string.
     * @param args The arguments to format the message.
     * @return The string for the given key.
     */
    public final String getString(final String key, final Object... args) {
        return MessageFormat.format(bundle.getString(key), args);
    }

    /**
     * Gets an image for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired image.
     * @return The image for the given key.
     * @throws IOException If an Input/Output error occurs while reading the image.
     */
    public final Image getImage(final String key) throws IOException {
        return ImageIO.read(getClass().getClassLoader().getResourceAsStream(getString(key)));
    }

    /**
     * Gets a color for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired color.
     * @return The color for the given key.
     */
    public final Color getColor(final String key) {
        return new Color((int) Long.parseLong(getString(key), HEXA) & 0xffffffff, true);
    }

    /**
     * Gets an integer for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired integer.
     * @return The integer for the given key.
     */
    public final int getInt(final String key) {
        return Integer.parseInt(getString(key));
    }

    /**
     * Gets a double for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired double.
     * @return The double for the given key.
     */
    public final double getDouble(final String key) {
        return Double.parseDouble(getString(key));
    }

    /**
     * Gets an audio input stream for the given key from the resource bundle or one of its parents.
     *
     * @param key The key for the desired audio input stream.
     * @return The audio input stream for the given key.
     * @throws UnsupportedAudioFileException If the stream does not point to valid audiofile data recognized by the system
     * @throws IOException If an input/output error occurs.
     */
    public final AudioInputStream getAudio(final String key) throws UnsupportedAudioFileException, IOException {
        return AudioSystem.getAudioInputStream(getClass().getClassLoader().getResourceAsStream(getString(key)));
    }

    /**
     * Colorize the given image with the given color. New pixel color has the same alpha value than  the same pixel in the image and his
     * liminance is multiplied by the luminance of the same pixel in the image
     *
     * @param i The image to colorize.
     * @param c The color to use to colorize the image.
     * @return A colorized copy of the specified image.
     */
    public static final Image colorize(final Image i, final Color c) {
        final BufferedImage colorized = new BufferedImage(i.getWidth(null), i.getHeight(null), BufferedImage.TYPE_INT_ARGB);
        colorized.getGraphics().drawImage(i, 0, 0, null);

        for (int x = 0; x < colorized.getWidth(null); x++) {
            for (int y = 0; y < colorized.getWidth(null); y++) {
                Color cc = new Color(colorized.getRGB(x, y), true);
                final double l = luminance(cc);

                cc = new Color(
                        (int) Math.round(l * c.getRed()),
                        (int) Math.round(l * c.getGreen()),
                        (int) Math.round(l * c.getBlue()),
                        cc.getAlpha());

                colorized.setRGB(x, y, cc.getRGB());
            }
        }

        return colorized;
    }

    /**
     * Compute the luminance (in [0,1]) of the specified color.
     *
     * @param c The color of which the luminance shall be computed.
     * @return The luminance of the specified color.
     */
    private static double luminance(final Color c) {
        final int maxValue = 255;

        final double r = 0.2126;
        final double g = 0.7152;
        final double b = 0.0722;

        final double l = r * c.getRed() + g * c.getGreen() + b * c.getBlue();

        return l / maxValue;
    }

}
