/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view;

import java.awt.Image;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Utils;

/**
 * The model object containing the slave information.
 *
 * @author Zleurtor
 */
public class SlaveInformation implements Comparable<SlaveInformation> {

    /**
     * The default color balance (100%).
     */
    public static final int DEFAULT_COLOR_BALANCE = 100;

    /**
     * The slave name (usually mac address).
     */
    private String name;

    /**
     * The slave version.
     */
    private String version;

    /**
     * The slave IP.
     */
    private String ip;

    /**
     * The slave options.
     */
    private String options;

    /**
     * The slave snapshot file name.
     */
    private String snapshotFilename;

    /**
     * The slave thumbnail image.
     */
    private Image thumbnail;

    /**
     * The slave color balance.
     */
    private int[] colorBalance = new int[] { DEFAULT_COLOR_BALANCE, DEFAULT_COLOR_BALANCE, DEFAULT_COLOR_BALANCE };

    /**
     * Constructs a new slave information with specified information.
     *
     * @param theName The slave name.
     * @param theIp The slave IP.
     * @param theThumbnail The slave thumbnail image.
     */
    public SlaveInformation(final String theName, final String theIp, final Image theThumbnail) {
        super();
        this.name = theName;
        this.ip = theIp;
        this.thumbnail = theThumbnail;

        final String theOptions = Configuration.getInstance().getString("configurator." + ip);
        if (theOptions != null) {
            options = theOptions;
        }

        try {
            colorBalance[0] = Configuration.getInstance().getInt("configurator." + ip + ".red");
            colorBalance[1] = Configuration.getInstance().getInt("configurator." + ip + ".green");
            colorBalance[2] = Configuration.getInstance().getInt("configurator." + ip + ".blue");
        } catch (final NumberFormatException e) {
            // Do nothing...
        }
    }

    /**
     * Returns the slave name(usually mac address).
     *
     * @return The slave name.
     */
    public final String getName() {
        return name;
    }

    /**
     * Returns the slave version.
     *
     * @return The slave version.
     */
    public final String getVersion() {
        return version;
    }

    /**
     * Sets new slave version.
     *
     * @param newVersion The new slave version.
     */
    public final void setVersion(final String newVersion) {
        this.version = newVersion;
    }

    /**
     * Returns the slave IP.
     *
     * @return The slave IP.
     */
    public final String getIp() {
        return ip;
    }

    /**
     * Returns the slave options.
     *
     * @return The slave options.
     */
    public final String getOptions() {
        return options;
    }

    /**
     * Sets new slave options.
     *
     * @param newOptions The new slave options.
     */
    public final void setOptions(final String newOptions) {
        this.options = newOptions;
        Configuration.getInstance().setString("configurator." + ip, options);
    }

    /**
     * Returns the slave snapshot file name.
     *
     * @return The slave snapshot file name.
     */
    public final String getSnapshotFilename() {
        return snapshotFilename;
    }

    /**
     * Sets a new slave snapshot file name.
     *
     * @param newSnapshotFilename The new slave snapshot file name.
     */
    public final void setSnapshotFilename(final String newSnapshotFilename) {
        this.snapshotFilename = newSnapshotFilename;
    }

    /**
     * Returns the slave thumbnail image.
     *
     * @return The slave thumbnail image.
     */
    public final Image getThumbnail() {
        return thumbnail;
    }

    /**
     * Sets a new slave thumbnail image.
     *
     * @param theThumbnail The new slave thumbnail image.
     */
    public final void setThumbnail(final Image theThumbnail) {
        this.thumbnail = theThumbnail;
    }

    /**
     * Returns the slave color balance.
     *
     * @return The slave color balance.
     */
    public final int[] getColorBalance() {
        return colorBalance;
    }

    /**
     * Sets a new slave color balance.
     *
     * @param newColorBalance The new slave color balance.
     */
    public final void setColorBalance(final int[] newColorBalance) {
        this.colorBalance = newColorBalance;
        Configuration.getInstance().setString("configurator." + ip + ".red", Integer.toString(colorBalance[0]));
        Configuration.getInstance().setString("configurator." + ip + ".green", Integer.toString(colorBalance[1]));
        Configuration.getInstance().setString("configurator." + ip + ".blue", Integer.toString(colorBalance[2]));
    }

    @Override
    public final int compareTo(final SlaveInformation other) throws ClassCastException {
        return Utils.getIpComparator().compare(getIp(), other.getIp());
    }

}
