/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Image;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;

import zleurtor.scanmanager.view.ResourcesManager;

/**
 * A specific button class customized for the 3D scan manager.
 *
 * @author Zleurtor
 */
public abstract class AbstractToolBarButton extends JButton implements MouseListener {

    private static final long serialVersionUID = 8227963340844360461L;

    /**
     * The default icon of the button.
     */
    private Icon defaultIcon;

    /**
     * The icon of the button when overflown by the cursor.
     */
    private Icon overflyIcon;

    /**
     * The icon of the button when clicked by the user.
     */
    private Icon activeIcon;

    /**
     * Constructs a new button withe specified parameters.
     *
     * @param icon The default icon of the button.
     * @param tooltip The tooltip of the button.
     */
    public AbstractToolBarButton(final Image icon, final String tooltip) {
        setForeground(ResourcesManager.getInstance().getColor(ResourcesManager.KEY_TOOLBAR_FOREGROUND));
        setBorder(BorderFactory.createEmptyBorder(
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_BUTTON_BORDERINSETS_TOP),
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_BUTTON_BORDERINSETS_LEFT),
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_BUTTON_BORDERINSETS_BOTTOM),
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_BUTTON_BORDERINSETS_RIGHT)));
        setContentAreaFilled(false);
        setOpaque(false);

        defaultIcon = new ImageIcon(ResourcesManager.colorize(icon, getForeground()));
        overflyIcon = new ImageIcon(ResourcesManager.colorize(icon,
                ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE)));
        activeIcon = new ImageIcon(ResourcesManager.colorize(icon,
                ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE).brighter()));
        setIcon(defaultIcon);
        setToolTipText(tooltip);

        addMouseListener(this);
    }

    /**
     * Returns the default icon of the button.
     *
     * @param newDefaultIcon The default icon of the button.
     */
    public final void setDefaultIcon(final Icon newDefaultIcon) {
        this.defaultIcon = newDefaultIcon;
    }

    /**
     * Returns the icon of the button when overflown by the cursor.
     *
     * @param newOverflyIcon The icon of the button when overflown by the cursor.
     */
    public final void setOverflyIcon(final Icon newOverflyIcon) {
        this.overflyIcon = newOverflyIcon;
    }

    /**
     * Returns the icon of the button when clicked by the user.
     *
     * @param newActiveIcon The icon of the button when clicked by the user.
     */
    public final void setActiveIcon(final Icon newActiveIcon) {
        this.activeIcon = newActiveIcon;
    }

    @Override
    public final void mouseEntered(final MouseEvent evt) {
        if (isEnabled()) {
            setIcon(overflyIcon);
        }
    }

    @Override
    public final void mouseExited(final MouseEvent evt) {
        if (isEnabled()) {
            setIcon(defaultIcon);
        }
    }

    @Override
    public final void mouseClicked(final MouseEvent evt) { }

    @Override
    public final void mousePressed(final MouseEvent evt) {
        if (isEnabled()) {
            setIcon(activeIcon);
        }
    }

    @Override
    public final void mouseReleased(final MouseEvent evt) {
        if (isEnabled()) {
            if (contains(evt.getPoint())) {
                setIcon(overflyIcon);
            } else {
                setIcon(defaultIcon);
            }
        }
    }

}
