/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;
import javax.swing.ComboBoxModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.JSpinner.NumberEditor;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Utils;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;

/**
 * The dialog displaying all the options configurable for the snapshot and the preview processes.
 *
 * @author Zleurtor
 */
public class ConfiguratorDialog extends JDialog {

    private static final long serialVersionUID = -5300033032252933077L;

    /**
     * All the option names of supported options.
     */
    private static final List<String> SUPPORTED_OPTION_NAMES =
            Arrays.asList(new String[] { "-q", "-t", "-e", "-sh", "-co", "-br", "-sa", "-ev", "-ex", "-awb", "-ss" });

    /**
     * The distance, measured in values, between each minor tick mark.
     */
    private static final int MINOR_TICK_SPACING = 1;

    /**
     * The distance, measured in values, between each major tick mark.
     */
    private static final int MAJOR_TICK_SPACING = 5;

    /**
     * The quality default value.
     */
    private static final int QUALITY_DEFAULT_VALUE = 100;

    /**
     * The timeout default value.
     */
    private static final int TIMEOUT_DEFAULT_VALUE = 300;

    /**
     * The encoding default value.
     */
    private static final String ENCODING_DEFAULT_VALUE = "jpg";

    /**
     * The sharpness default value.
     */
    private static final int SHARPNESS_DEFAULT_VALUE = 0;

    /**
     * The contrast default value.
     */
    private static final int CONTRAST_DEFAULT_VALUE = 0;

    /**
     * The brightness default value.
     */
    private static final int BRIGHTNESS_DEFAULT_VALUE = 50;

    /**
     * The saturation default value.
     */
    private static final int SATURATION_DEFAULT_VALUE = 0;

    /**
     * The EV compensation default value.
     */
    private static final int EVCOMPENSATION_DEFAULT_VALUE = 0;

    /**
     * The exposure default value.
     */
    private static final String EXPOSURE_DEFAULT_VALUE = "auto";

    /**
     * The AWB default value.
     */
    private static final String AWB_DEFAULT_VALUE = "auto";

    /**
     * The shutter speed default value.
     */
    private static final int SHUTTERSPEED_DEFAULT_VALUE = 125_000;

    /**
     * All the options names.
     */
    private static final Map<String, String> OPTIONS_NAMES = new HashMap<>();

    /**
     * The profiles combo box.
     */
    private final JComboBox<String> profilesComboBox = new JComboBox<>();

    /**
     * The load profile button.
     */
    private final JButton loadProfileButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_LOAD_PROFILE));

    /**
     * The save profile button.
     */
    private final JButton saveProfileButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_SAVEPROFILE));

    /**
     * The delete profile button.
     */
    private final JButton deleteProfileButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_DELETEPROFILE));

    /**
     * The quality option slider.
     */
    private final JSlider qualitySlider = new JSlider(0, 100, 100);

    /**
     * The quality option's value label.
     */
    private final JLabel qualityValue = new JLabel();

    /**
     * The timeout option spinner.
     */
    private final JSpinner timeoutSpinner = new JSpinner(new SpinnerNumberModel(300, 1, Integer.MAX_VALUE, 1));

    /**
     * The encoding option combo box.
     */
    private final JComboBox<String> encodingComboBox = new JComboBox<>();

    /**
     * The sharpness option slider.
     */
    private final JSlider sharpnessSlider = new JSlider(-100, 100, 0);

    /**
     * The sharpness option's value label.
     */
    private final JLabel sharpnessValue = new JLabel();

    /**
     * The contrast option slider.
     */
    private final JSlider contrastSlider = new JSlider(-100, 100, 0);

    /**
     * The contrast option's value label.
     */
    private final JLabel contrastValue = new JLabel();

    /**
     * The brightness option slider.
     */
    private final JSlider brightnessSlider = new JSlider(0, 100, 50);

    /**
     * The brightness option's value label.
     */
    private final JLabel brightnessValue = new JLabel();

    /**
     * The saturation option slider.
     */
    private final JSlider saturationSlider = new JSlider(-100, 100, 0);

    /**
     * The saturation option's value label.
     */
    private final JLabel saturationValue = new JLabel();

    /**
     * The EV compensation option slider.
     */
    private final JSlider evCompensationSlider = new JSlider(-10, 10, 0);

    /**
     * The EV compensation option's value label.
     */
    private final JLabel evCompensationValue = new JLabel();

    /**
     * The exposure option combo box.
     */
    private final JComboBox<String> exposureComboBox = new JComboBox<>();

    /**
     * The AWB option combo box.
     */
    private final JComboBox<String> awbComboBox = new JComboBox<>();

    /**
     * The shutter speed option slider.
     */
    private final JSlider shutterSpeedSlider = new JSlider(1_000, 1_000_000, 125_000);

    /**
     * The shutter speed option's value label.
     */
    private final JLabel shutterSpeedValue = new JLabel();

    /**
     * A list of files containing snapshots used as sample for color shift setting.
     */
    private List<File> colorsSamples = new ArrayList<>();

    /**
     * The panel displaying the samples.
     */
    private ColorsSamplePanel colorsSamplePanel = new ColorsSamplePanel();

    /**
     * The current displayed sample index.
     */
    private int colorsSampleIndex = 0;

    /**
     * The label displaying an arrow to display the previous sample.
     */
    private final JLabel previousLabel = new JLabel();

    /**
     * The label displaying an arrow to display the next sample.
     */
    private final JLabel nextLabel = new JLabel();

    /**
     * The red shift option slider.
     */
    private final JSlider redSlider = new JSlider(0, 200, 100);

    /**
     * The red shift option's value label.
     */
    private final JLabel redValue = new JLabel();

    /**
     * The green shift option slider.
     */
    private final JSlider greenSlider = new JSlider(0, 200, 100);

    /**
     * The green shift option's value label.
     */
    private final JLabel greenValue = new JLabel();

    /**
     * The blue shift option slider.
     */
    private final JSlider blueSlider = new JSlider(0, 200, 100);

    /**
     * The blue shift option's value label.
     */
    private final JLabel blueValue = new JLabel();

    /**
     * The unsupported options text field.
     */
    private final JTextField unsupportedOptions = new JTextField();

    /**
     * All the controls.
     */
    private final Map<String, JComponent> controls = new HashMap<>();

    /**
     * The reset button (to use to reinit the controls to their default values).
     */
    private final JButton resetButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_RESET));

    /**
     * The cancel button (to use to quit the dialog without changing the options value).
     */
    private final JButton cancelButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_CANCEL));

    /**
     * The button to validate the modified options and apply them to the selected slaves.
     */
    private final JButton applySelectedButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_APPLYTOSELECTED));

    /**
     * The button to validate the modified options and apply them to all the slaves.
     */
    private final JButton applyAllButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_APPLYALL));

    /**
     * The used close option (cancel, apply to selected or apply to all).
     */
    private CloseOption closeOption = CloseOption.CANCEL;

    /**
     * The options string (evaluated at the dialog closure).
     */
    private String options;

    static {
        OPTIONS_NAMES.put("-q", "quality");
        OPTIONS_NAMES.put("-t", "timeout");
        OPTIONS_NAMES.put("-e", "encoding");
        OPTIONS_NAMES.put("-sh", "sharpness");
        OPTIONS_NAMES.put("-co", "contrast");
        OPTIONS_NAMES.put("-br", "brightness");
        OPTIONS_NAMES.put("-sa", "saturation");
        OPTIONS_NAMES.put("-ev", "evCompensation");
        OPTIONS_NAMES.put("-ex", "exposure");
        OPTIONS_NAMES.put("-awb", "awb");
        OPTIONS_NAMES.put("-ss", "shutterSpeed");
    }

    /**
     * Constructs a new dialog having the specified window as owner.
     *
     * @param window The Window from which the dialog is displayed.
     * @throws IOException If an Input/Output error occurs while reading an image.
     */
    public ConfiguratorDialog(final Window window) throws IOException {
        super(window, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_SNAPSHOTOPTIONS), ModalityType.APPLICATION_MODAL);

        controls.put("-q", qualitySlider);
        controls.put("-t", timeoutSpinner);
        controls.put("-e", encodingComboBox);
        controls.put("-sh", sharpnessSlider);
        controls.put("-co", contrastSlider);
        controls.put("-br", brightnessSlider);
        controls.put("-sa", saturationSlider);
        controls.put("-ev", evCompensationSlider);
        controls.put("-ex", exposureComboBox);
        controls.put("-awb", awbComboBox);
        controls.put("-ss", shutterSpeedSlider);

        setContentPane(new ContentPanel());
        getContentPane().setLayout(new GridBagLayout());
        initializeComponents();
        reset();

        final int gap = 6;
        final int profilesPanelRow = 0;
        final int tabbedPanelRow = 1;
        final int unsupportedOptionsPanelRow = 2;
        final int buttonsPanelRow = 3;

        add(createProfilesPanel(), new GridBagConstraints(0, profilesPanelRow, 1, 1, 1, 0,
                GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
                new Insets(gap, gap, gap / 2, gap), 0, 0));
        final JTabbedPane tabbedPanel = new JTabbedPane();
        tabbedPanel.add(createImagePanel(), ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_CONFIGURATOR_IMAGE));
        tabbedPanel.add(createColorsPanel(), ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_CONFIGURATOR_COLORS));
        add(tabbedPanel, new GridBagConstraints(0, tabbedPanelRow, 1, 1, 1, 1,
                GridBagConstraints.CENTER, GridBagConstraints.BOTH,
                new Insets(gap / 2, gap, gap / 2, gap), 0, 0));
        add(createUnsupportedOptionsPanel(), new GridBagConstraints(0, unsupportedOptionsPanelRow, 1, 1, 1, 0,
                GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
                new Insets(gap / 2, gap, gap / 2, gap), 0, 0));
        add(createButtonsPanel(), new GridBagConstraints(0, buttonsPanelRow, 1, 1, 1, 0,
                GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
                new Insets(gap, gap, gap / 2, gap), 0, 0));
    }

    /**
     * Returns the list of files containing snapshots used as sample for color shift setting.
     *
     * @return The list of files containing snapshots used as sample for color shift setting.
     */
    public final List<File> getColorsSamples() {
        return colorsSamples;
    }

    /**
     * Sets a new list of files containing snapshots used as sample for color shift setting.
     *
     * @param newColorsSamples The new list of files containing snapshots used as sample for color shift setting.
     */
    public void setColorsSamples(final List<File> newColorsSamples) {
        try {
            colorsSampleIndex = 0;

            this.colorsSamples = newColorsSamples;
            if (colorsSamples != null && !colorsSamples.isEmpty()) {
                colorsSamplePanel.setColorsSample(colorsSamples.get(colorsSampleIndex));
            } else {
                colorsSamplePanel.setColorsSample(null);
            }
        } catch (final IOException e) {
            GuiUtils.showError(e.getMessage(), this);
        }
    }

    /**
     * Returns the used close option (cancel, apply to selected or apply to all).
     *
     * @return The used close option.
     */
    public final CloseOption getCloseOption() {
        return closeOption;
    }

    /**
     * Returns the options string (evaluated at the dialog closure).
     *
     * @return The options string.
     */
    public final String getOptions() {
        return options;
    }

    /**
     * Returns the color balance array.
     *
     * @return The color balance array.
     */
    public final int[] getColorBalance() {
        return new int[] {
                redSlider.getValue(),
                greenSlider.getValue(),
                blueSlider.getValue(),
        };
    }

    /**
     * Creates the profiles panel.
     *
     * @return The newly created profiles panel.
     */
    private JPanel createProfilesPanel() {
        final JPanel profilesPanel = new JPanel(new GridBagLayout());
        profilesPanel.setOpaque(false);

        final int gap = 6;
        final int loadProfileButtonColumn = 2;
        final int saveProfileButtonColumn = 3;
        final int deleteProfileButtonColumn = 4;

        profilesPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_PROFILES)),
                new GridBagConstraints(0, 0, 1, 1, 0, 1,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(0, gap, 0, gap / 2), 0, 0));
        profilesPanel.add(profilesComboBox,
                new GridBagConstraints(1, 0, 1, 1, 1, 1,
                        GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                        new Insets(0, gap / 2, 0, gap), 0, 0));
        profilesPanel.add(loadProfileButton,
                new GridBagConstraints(loadProfileButtonColumn, 0, 1, 1, 0, 1,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(0, gap / 2, 0, gap), 0, 0));
        profilesPanel.add(saveProfileButton,
                new GridBagConstraints(saveProfileButtonColumn, 0, 1, 1, 0, 1,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(0, gap / 2, 0, gap), 0, 0));
        profilesPanel.add(deleteProfileButton,
                new GridBagConstraints(deleteProfileButtonColumn, 0, 1, 1, 0, 1,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(0, gap / 2, 0, gap), 0, 0));

        return profilesPanel;
    }

    /**
     * Creates the image panel.
     *
     * @return The newly created image panel.
     */
    private JPanel createImagePanel() {
        final JPanel imagePanel = new JPanel(new TableLayout(11, 7, 6, 6));
        imagePanel.setOpaque(false);

        qualitySlider.setPaintTicks(true);
        qualitySlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        qualitySlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        sharpnessSlider.setPaintTicks(true);
        sharpnessSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        sharpnessSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        contrastSlider.setPaintTicks(true);
        contrastSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        contrastSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        brightnessSlider.setPaintTicks(true);
        brightnessSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        brightnessSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        saturationSlider.setPaintTicks(true);
        saturationSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        saturationSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        evCompensationSlider.setPaintTicks(true);
        evCompensationSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        evCompensationSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        shutterSpeedSlider.setPaintTicks(true);
        shutterSpeedSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        shutterSpeedSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        final int controlsWidth = 3;
        final int labelsColum = 5;

        final int qualityRow = 0;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_QUALITY), JLabel.RIGHT),
                new TableConstraints(qualityRow, 0, 2, 1));
        imagePanel.add(qualitySlider,
                new TableConstraints(qualityRow, 2, controlsWidth, 1));
        imagePanel.add(qualityValue,
                new TableConstraints(qualityRow, labelsColum, 2, 1));

        final int timeoutRow = 1;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_TIMEOUT), JLabel.RIGHT),
                new TableConstraints(timeoutRow, 0, 2, 1));
        imagePanel.add(timeoutSpinner,
                new TableConstraints(timeoutRow, 2, controlsWidth, 1));

        final int encodingRow = 2;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_ENCODING), JLabel.RIGHT),
                new TableConstraints(encodingRow, 0, 2, 1));
        imagePanel.add(encodingComboBox,
                new TableConstraints(encodingRow, 2, controlsWidth, 1));

        final int sharpnessRow = 3;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_SHARPNESS), JLabel.RIGHT),
                new TableConstraints(sharpnessRow, 0, 2, 1));
        imagePanel.add(sharpnessSlider,
                new TableConstraints(sharpnessRow, 2, controlsWidth, 1));
        imagePanel.add(sharpnessValue,
                new TableConstraints(sharpnessRow, labelsColum, 2, 1));

        final int contrastRow = 4;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_CONTRAST), JLabel.RIGHT),
                new TableConstraints(contrastRow, 0, 2, 1));
        imagePanel.add(contrastSlider,
                new TableConstraints(contrastRow, 2, controlsWidth, 1));
        imagePanel.add(contrastValue,
                new TableConstraints(contrastRow, labelsColum, 2, 1));

        final int brightnessRow = 5;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_BRIGHTNESS), JLabel.RIGHT),
                new TableConstraints(brightnessRow, 0, 2, 1));
        imagePanel.add(brightnessSlider,
                new TableConstraints(brightnessRow, 2, controlsWidth, 1));
        imagePanel.add(brightnessValue,
                new TableConstraints(brightnessRow, labelsColum, 2, 1));

        final int saturationRow = 6;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_SATURATION), JLabel.RIGHT),
                new TableConstraints(saturationRow, 0, 2, 1));
        imagePanel.add(saturationSlider,
                new TableConstraints(saturationRow, 2, controlsWidth, 1));
        imagePanel.add(saturationValue,
                new TableConstraints(saturationRow, labelsColum, 2, 1));

        final int evCompensationRow = 7;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_EVCOMPENSATION), JLabel.RIGHT),
                new TableConstraints(evCompensationRow, 0, 2, 1));
        imagePanel.add(evCompensationSlider,
                new TableConstraints(evCompensationRow, 2, controlsWidth, 1));
        imagePanel.add(evCompensationValue,
                new TableConstraints(evCompensationRow, labelsColum, 2, 1));

        final int exposureRow = 8;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_EXPOSURE), JLabel.RIGHT),
                new TableConstraints(exposureRow, 0, 2, 1));
        imagePanel.add(exposureComboBox,
                new TableConstraints(exposureRow, 2, controlsWidth, 1));

        final int awbRow = 9;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_AWB), JLabel.RIGHT),
                new TableConstraints(awbRow, 0, 2, 1));
        imagePanel.add(awbComboBox,
                new TableConstraints(awbRow, 2, controlsWidth, 1));

        final int shutterSpeedRow = 10;
        imagePanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_SHUTTERSPEED), JLabel.RIGHT),
                new TableConstraints(shutterSpeedRow, 0, 2, 1));
        imagePanel.add(shutterSpeedSlider,
                new TableConstraints(shutterSpeedRow, 2, controlsWidth, 1));
        imagePanel.add(shutterSpeedValue,
                new TableConstraints(shutterSpeedRow, labelsColum, 2, 1));

        return imagePanel;
    }

    /**
     * Creates the colors panel.
     *
     * @return The newly created colors panel.
     */
    private JPanel createColorsPanel() {
        final JPanel colorsPanel = new JPanel(new TableLayout(11, 7, 6, 6));
        colorsPanel.setOpaque(false);
        colorsSamplePanel.setOpaque(false);

        redSlider.setPaintTicks(true);
        redSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        redSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        greenSlider.setPaintTicks(true);
        greenSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        greenSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        blueSlider.setPaintTicks(true);
        blueSlider.setMinorTickSpacing(MINOR_TICK_SPACING);
        blueSlider.setMajorTickSpacing(MAJOR_TICK_SPACING);

        final int controlsWidth = 3;
        final int labelsColum = 5;

        final int sampleWidth = 5;
        final int sampleHeight = 8;
        final int previousColumn = 0;
        final int nextColumn = 6;
        colorsPanel.add(colorsSamplePanel,
                new TableConstraints(0, 1, sampleWidth, sampleHeight));
        colorsPanel.add(previousLabel,
                new TableConstraints(0, previousColumn, 1, sampleHeight));
        colorsPanel.add(nextLabel,
                new TableConstraints(0, nextColumn, 1, sampleHeight));

        final int redRow = 8;
        colorsPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_RED), JLabel.RIGHT),
                new TableConstraints(redRow, 0, 2, 1));
        colorsPanel.add(redSlider,
                new TableConstraints(redRow, 2, controlsWidth, 1));
        colorsPanel.add(redValue,
                new TableConstraints(redRow, labelsColum, 2, 1));

        final int greenRow = 9;
        colorsPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_GREEN), JLabel.RIGHT),
                new TableConstraints(greenRow, 0, 2, 1));
        colorsPanel.add(greenSlider,
                new TableConstraints(greenRow, 2, controlsWidth, 1));
        colorsPanel.add(greenValue,
                new TableConstraints(greenRow, labelsColum, 2, 1));

        final int blueRow = 10;
        colorsPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_BLUE), JLabel.RIGHT),
                new TableConstraints(blueRow, 0, 2, 1));
        colorsPanel.add(blueSlider,
                new TableConstraints(blueRow, 2, controlsWidth, 1));
        colorsPanel.add(blueValue,
                new TableConstraints(blueRow, labelsColum, 2, 1));

        return colorsPanel;
    }

    /**
     * Creates the unsupported options panel.
     *
     * @return The newly created options panel.
     */
    private JPanel createUnsupportedOptionsPanel() {
        final JPanel unsupportedOptionsPanel = new JPanel(new GridBagLayout());

        final int gap = 6;

        unsupportedOptionsPanel.add(new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_UNSUPPORTEDOPTIONS)),
                new GridBagConstraints(0, 0, 1, 1, 0, 1,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(0, gap, 0, gap / 2), 0, 0));
        unsupportedOptionsPanel.add(unsupportedOptions,
                new GridBagConstraints(1, 0, 1, 1, 1, 1,
                        GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                        new Insets(0, gap / 2, 0, gap), 0, 0));

        return unsupportedOptionsPanel;
    }

    /**
     * Creates the buttons panel.
     *
     * @return The newly created buttons panel.
     */
    private JPanel createButtonsPanel() {
        final int gap = 5;

        final JPanel buttonsPanel = new JPanel();
        buttonsPanel.setLayout(new GridBagLayout());

        final int resetColumn = 0;
        final int cancelColumn = 1;
        final int applySelectedColumn = 2;
        final int applyAllColumn = 3;

        buttonsPanel.add(resetButton,
                new GridBagConstraints(resetColumn, 0, 1, 1, 1, 0, GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap, gap, gap / 2), 0, 0));
        buttonsPanel.add(cancelButton,
                new GridBagConstraints(cancelColumn, 0, 1, 1, 0, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap / 2), 0, 0));
        buttonsPanel.add(applySelectedButton,
                new GridBagConstraints(applySelectedColumn, 0, 1, 1, 0, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap / 2), 0, 0));
        buttonsPanel.add(applyAllButton,
                new GridBagConstraints(applyAllColumn, 0, 1, 1, 0, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap), 0, 0));

        return buttonsPanel;
    }

    /**
     * Initializes the components (items in combo boxes, labels values and listeners).
     *
     * @throws IOException If an Input/Output error occurs while reading an image.
     */
    private void initializeComponents() throws IOException {
        String profiles = Configuration.getInstance().getString("configurator.profiles");
        if (profiles == null) {
            profiles = "";
        }
        for (final String profile : profiles.split("\\|")) {
            if (!profile.trim().isEmpty()) {
                profilesComboBox.addItem(profile);
            }
        }
        profilesComboBox.setEditable(true);

        encodingComboBox.addItem("jpg");
        encodingComboBox.addItem("bmp");
        encodingComboBox.addItem("gif");
        encodingComboBox.addItem("png");
        encodingComboBox.setEditable(false);

        exposureComboBox.addItem("auto");
        exposureComboBox.addItem("off");
        exposureComboBox.addItem("night");
        exposureComboBox.addItem("nightpreview");
        exposureComboBox.addItem("backlight");
        exposureComboBox.addItem("spotlight");
        exposureComboBox.addItem("sports");
        exposureComboBox.addItem("snow");
        exposureComboBox.addItem("beach");
        exposureComboBox.addItem("verylong");
        exposureComboBox.addItem("fixedfps");
        exposureComboBox.addItem("antishake");
        exposureComboBox.addItem("fireworks");
        exposureComboBox.setEditable(false);

        awbComboBox.addItem("auto");
        awbComboBox.addItem("off");
        awbComboBox.addItem("sun");
        awbComboBox.addItem("cloud");
        awbComboBox.addItem("shade");
        awbComboBox.addItem("tungsten");
        awbComboBox.addItem("fluorescent");
        awbComboBox.addItem("incandescent");
        awbComboBox.addItem("flash");
        awbComboBox.addItem("horizon");
        awbComboBox.addItem("greyworld");
        awbComboBox.setEditable(false);

        previousLabel.setIcon(new ImageIcon(
                ResourcesManager.colorize(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_BUTTON_PREVIOUS),
                        previousLabel.getForeground())));
        previousLabel.setOpaque(false);
        previousLabel.setHorizontalAlignment(SwingConstants.LEFT);
        previousLabel.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));

        nextLabel.setIcon(new ImageIcon(
                ResourcesManager.colorize(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_BUTTON_NEXT),
                        previousLabel.getForeground())));
        nextLabel.setOpaque(false);
        nextLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        nextLabel.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));

        initComponentsListeners();
    }

    /**
     * Initializes the components' listeners.
     *
     * @throws IOException If an Input/Output error occurs while reading an image.
     */
    private void initComponentsListeners() throws IOException {
        loadProfileButton.addActionListener((final ActionEvent evt) -> {
            loadProfile();
        });
        saveProfileButton.addActionListener((final ActionEvent evt) -> {
            saveProfile();
        });
        deleteProfileButton.addActionListener((final ActionEvent evt) -> {
            deleteProfile();
        });

        qualitySlider.addChangeListener((ChangeEvent evt) -> {
            qualityValue.setText(Integer.toString((int) qualitySlider.getValue()));
        });
        sharpnessSlider.addChangeListener((ChangeEvent evt) -> {
            sharpnessValue.setText(Integer.toString((int) sharpnessSlider.getValue()));
        });
        contrastSlider.addChangeListener((ChangeEvent evt) -> {
            contrastValue.setText(Integer.toString((int) contrastSlider.getValue()));
        });
        brightnessSlider.addChangeListener((ChangeEvent evt) -> {
            brightnessValue.setText(Integer.toString((int) brightnessSlider.getValue()));
        });
        saturationSlider.addChangeListener((ChangeEvent evt) -> {
            saturationValue.setText(Integer.toString((int) saturationSlider.getValue()));
        });
        evCompensationSlider.addChangeListener((ChangeEvent evt) -> {
            evCompensationValue.setText(Integer.toString((int) evCompensationSlider.getValue()));
        });
        shutterSpeedSlider.addChangeListener((ChangeEvent evt) -> {
            shutterSpeedValue.setText(formatShutterSpeed(shutterSpeedSlider.getValue()));
        });
        redSlider.addChangeListener((ChangeEvent evt) -> {
            redValue.setText(Integer.toString((int) redSlider.getValue()));
        });
        greenSlider.addChangeListener((ChangeEvent evt) -> {
            greenValue.setText(Integer.toString((int) greenSlider.getValue()));
        });
        blueSlider.addChangeListener((ChangeEvent evt) -> {
            blueValue.setText(Integer.toString((int) blueSlider.getValue()));
        });

        final MouseListener mouseListener = new MouseAdapter() {
            private final Icon defaultPreviousIcon = previousLabel.getIcon();

            private final Icon overflyPreviousIcon = new ImageIcon(
                    ResourcesManager.colorize(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_BUTTON_PREVIOUS),
                            ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE)));

            private final Icon defaultNextIcon = nextLabel.getIcon();

            private final Icon overflyNextIcon = new ImageIcon(
                    ResourcesManager.colorize(ResourcesManager.getInstance().getImage(ResourcesManager.KEY_BUTTON_NEXT),
                            ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE)));

            @Override
            public void mouseClicked(final MouseEvent evt) {
                try {
                    if (evt.getSource() == previousLabel && !colorsSamples.isEmpty()) {
                        if (colorsSampleIndex <= 0) {
                            colorsSampleIndex = colorsSamples.size() - 1;
                        } else {
                            colorsSampleIndex = (colorsSampleIndex - 1) % colorsSamples.size();
                        }
                        colorsSamplePanel.setColorsSample(colorsSamples.get(colorsSampleIndex));
                    } else if (evt.getSource() == nextLabel && !colorsSamples.isEmpty()) {
                        colorsSampleIndex = (colorsSampleIndex + 1) % colorsSamples.size();
                        colorsSamplePanel.setColorsSample(colorsSamples.get(colorsSampleIndex));
                    }
                } catch (final IOException e) {
                    GuiUtils.showError(e.getMessage(), colorsSamplePanel);
                }
            }

            @Override
            public void mouseEntered(final MouseEvent evt) {
                if (evt.getSource() == previousLabel) {
                    previousLabel.setIcon(overflyPreviousIcon);
                } else if (evt.getSource() == nextLabel) {
                    nextLabel.setIcon(overflyNextIcon);
                }
            }

            @Override
            public void mouseExited(final MouseEvent evt) {
                if (evt.getSource() == previousLabel) {
                    previousLabel.setIcon(defaultPreviousIcon);
                } else if (evt.getSource() == nextLabel) {
                    nextLabel.setIcon(defaultNextIcon);
                }
            }
        };

        previousLabel.addMouseListener(mouseListener);
        nextLabel.addMouseListener(mouseListener);

        resetButton.addActionListener((final ActionEvent evt) -> {
            reset();
            repaint();
        });
        cancelButton.addActionListener((final ActionEvent evt) -> {
            closeOption = CloseOption.CANCEL;
            computeOptions();
            dispose();
        });
        applySelectedButton.addActionListener((final ActionEvent evt) -> {
            validateTimeoutSpinner();
            closeOption = CloseOption.APPLY_SELECTED;
            computeOptions();
            dispose();
        });
        applyAllButton.addActionListener((final ActionEvent evt) -> {
            validateTimeoutSpinner();
            closeOption = CloseOption.APPLY_ALL;
            computeOptions();
            dispose();
        });

        final ColorsChangeListener colorsChangeListener = new ColorsChangeListener(colorsSamplePanel);
        redSlider.addChangeListener(colorsChangeListener);
        greenSlider.addChangeListener(colorsChangeListener);
        blueSlider.addChangeListener(colorsChangeListener);
    }

    /**
     * Validate the current modified value in the timeout spinner.
     */
    private void validateTimeoutSpinner() {
        if (timeoutSpinner.getEditor() instanceof NumberEditor) {
            final NumberEditor editor = (NumberEditor) timeoutSpinner.getEditor();
            if (Utils.isNumber(editor.getTextField().getText())) {
                final double timeout = Double.parseDouble(editor.getTextField().getText());
                timeoutSpinner.setValue(timeout);
            }
        }
    }

    /**
     * Format the shutter speed value (given as microseconds) as a "1/n s".
     *
     * @param shutterSpeed The shutter speed value (given as microseconds).
     * @return The formatted shutter speed.
     */
    private String formatShutterSpeed(final double shutterSpeed) {
        final double micro = 1_000_000d;
        return NumberFormat.getInstance().format(Math.round(shutterSpeed))
                + " µs  (1/"
                + NumberFormat.getInstance().format(Math.round(micro / shutterSpeed))
                + " s)";
    }

    /**
     * Computes the options string.
     */
    private void computeOptions() {
        options = "";

        options += "-q " + qualitySlider.getValue();
        options += " -t " + timeoutSpinner.getValue();
        options += " -e " + encodingComboBox.getSelectedItem();
        options += " -sh " + sharpnessSlider.getValue();
        options += " -co " + contrastSlider.getValue();
        options += " -br " + brightnessSlider.getValue();
        options += " -sa " + saturationSlider.getValue();
        options += " -ev " + evCompensationSlider.getValue();
        options += " -ex " + exposureComboBox.getSelectedItem();
        options += " -awb " + awbComboBox.getSelectedItem();
        options += " -ss " + shutterSpeedSlider.getValue();
        if (unsupportedOptions.getText() != null && !unsupportedOptions.getText().isEmpty()) {
            options += " " + unsupportedOptions.getText();
        }
    }

    /**
     * Reset the controls to their default values.
     */
    private void reset() {
        closeOption = CloseOption.CANCEL;
        options = "";

        qualitySlider.setValue(QUALITY_DEFAULT_VALUE);
        timeoutSpinner.setValue(TIMEOUT_DEFAULT_VALUE);
        encodingComboBox.setSelectedItem(ENCODING_DEFAULT_VALUE);
        sharpnessSlider.setValue(SHARPNESS_DEFAULT_VALUE);
        contrastSlider.setValue(CONTRAST_DEFAULT_VALUE);
        brightnessSlider.setValue(BRIGHTNESS_DEFAULT_VALUE);
        saturationSlider.setValue(SATURATION_DEFAULT_VALUE);
        evCompensationSlider.setValue(EVCOMPENSATION_DEFAULT_VALUE);
        exposureComboBox.setSelectedItem(EXPOSURE_DEFAULT_VALUE);
        awbComboBox.setSelectedItem(AWB_DEFAULT_VALUE);
        shutterSpeedSlider.setValue(SHUTTERSPEED_DEFAULT_VALUE);
        unsupportedOptions.setText(null);

        qualityValue.setText(Integer.toString((int) qualitySlider.getValue()));
        sharpnessValue.setText(Integer.toString((int) sharpnessSlider.getValue()));
        contrastValue.setText(Integer.toString((int) contrastSlider.getValue()));
        brightnessValue.setText(Integer.toString((int) brightnessSlider.getValue()));
        saturationValue.setText(Integer.toString((int) saturationSlider.getValue()));
        evCompensationValue.setText(Integer.toString((int) evCompensationSlider.getValue()));
        shutterSpeedValue.setText(formatShutterSpeed(shutterSpeedSlider.getValue()));
        redValue.setText(Integer.toString((int) redSlider.getValue()));
        greenValue.setText(Integer.toString((int) greenSlider.getValue()));
        blueValue.setText(Integer.toString((int) blueSlider.getValue()));

        redSlider.setValue(SlaveInformation.DEFAULT_COLOR_BALANCE);
        greenSlider.setValue(SlaveInformation.DEFAULT_COLOR_BALANCE);
        blueSlider.setValue(SlaveInformation.DEFAULT_COLOR_BALANCE);
    }

    /**
     * Parses the specified string and initializes the controls depending on the parsed values.
     *
     * @param toParse The options strings to parse.
     */
    public final void parse(final String toParse) {
        reset();

        if (toParse == null || toParse.isEmpty()) {
            reset();
        } else {
            final String[] split = toParse.split("\\s+");
            String unsupported = "";
            String formerOption = "";

            for (final String o : split) {
                if (o.startsWith("-") && !Utils.isNumber(o)) {
                    // An option
                    if (isOption(o)) {
                        formerOption = o;
                    } else {
                        if (!unsupported.isEmpty()) {
                            unsupported += " ";
                        }
                        unsupported += o;
                        formerOption = "";
                    }
                } else {
                    if (!setValues(formerOption, o)) {
                        if (!unsupported.isEmpty()) {
                            unsupported += " ";
                        }
                        unsupported += o;
                    }
                    formerOption = "";
                }
            }

            qualityValue.setText(Integer.toString((int) qualitySlider.getValue()));
            sharpnessValue.setText(Integer.toString((int) sharpnessSlider.getValue()));
            contrastValue.setText(Integer.toString((int) contrastSlider.getValue()));
            brightnessValue.setText(Integer.toString((int) brightnessSlider.getValue()));
            saturationValue.setText(Integer.toString((int) saturationSlider.getValue()));
            evCompensationValue.setText(Integer.toString((int) evCompensationSlider.getValue()));
            shutterSpeedValue.setText(formatShutterSpeed(shutterSpeedSlider.getValue()));
            unsupportedOptions.setText(unsupported);
        }
    }

    /**
     * Actualize the values of the color balance sliders with given ones.
     *
     * @param colors The new color balance to set.
     */
    public void setColorBalance(final int[] colors) {
        redSlider.setValue(colors[0]);
        greenSlider.setValue(colors[1]);
        blueSlider.setValue(colors[2]);
    }

    /**
     * Parses the specified string and return it as a readable text.
     *
     * @param toParse The options strings to parse.
     * @return The specified string as a readable text.
     */
    public static final String parseToReadableFormat(final String toParse) {
        final Map<String, String> values = new LinkedHashMap<>();

        // Set default values
        values.put("-q", Integer.toString(BRIGHTNESS_DEFAULT_VALUE));
        values.put("-t", Integer.toString(TIMEOUT_DEFAULT_VALUE));
        values.put("-e", ENCODING_DEFAULT_VALUE);
        values.put("-sh", Integer.toString(SHARPNESS_DEFAULT_VALUE));
        values.put("-co", Integer.toString(CONTRAST_DEFAULT_VALUE));
        values.put("-br", Integer.toString(BRIGHTNESS_DEFAULT_VALUE));
        values.put("-sa", Integer.toString(SATURATION_DEFAULT_VALUE));
        values.put("-ev", Integer.toString(EVCOMPENSATION_DEFAULT_VALUE));
        values.put("-ex", EXPOSURE_DEFAULT_VALUE);
        values.put("-awb", AWB_DEFAULT_VALUE);
        values.put("-ss", Integer.toString(SHUTTERSPEED_DEFAULT_VALUE));

        // Set the argument values
        final String[] split = toParse.split("\\s+");
        String formerOption = "";

        for (final String o : split) {
            if (o.startsWith("-") && !Utils.isNumber(o)) {
                // An option
                formerOption = isOption(o) ? o : "";
            } else if (values.containsKey(formerOption)) {
                values.put(formerOption, o);
                formerOption = "";
            }
        }

        final StringBuilder readableFormat = new StringBuilder();
        for (final String key : values.keySet()) {
            if (readableFormat.length() > 0) {
                readableFormat.append("\n");
            }
            readableFormat.append("\t" + OPTIONS_NAMES.get(key) + ": " + values.get(key));
        }

        return readableFormat.toString();
    }

    /**
     * Test whether or not the specified string is a snapshot/preview supported option name.
     *
     * @param option The option name to test.
     * @return true if the specified string is a snapshot/preview supported option name, false otherwise.
     */
    private static boolean isOption(final String option) {
        return SUPPORTED_OPTION_NAMES.contains(option);
    }

    /**
     * Set the values of a control depending on the specified option.
     *
     * @param formerOption The former option name.
     * @param value The option value.
     * @return true if the control corresponding to the specified has been found (and its value has been changed).
     */
    private boolean setValues(final String formerOption, final String value) {
        final JComponent control = controls.get(formerOption);

        if (control == null) {
            return false;
        }

        if (control instanceof JSlider) {
            ((JSlider) control).setValue((int) Math.round(Double.parseDouble(value)));
        } else if (control instanceof JComboBox) {
            ((JComboBox<?>) control).setSelectedItem(value);
        } else {
            ((JSpinner) control).setValue(Double.parseDouble(value));
        }

        return true;
    }

    /**
     * Load the selected profile.
     */
    private void loadProfile() {
        profilesComboBox.setSelectedItem(profilesComboBox.getEditor().getItem());
        final Object selectedProfile = profilesComboBox.getSelectedItem();
        if (selectedProfile == null || selectedProfile.toString().isEmpty()) {
            JOptionPane.showMessageDialog(this, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_ENTERPROFILENAME));
        } else {
            final String profile = selectedProfile.toString();

            final String newOptions = Configuration.getInstance().getString("configurator.profile." + profile);
            final String newRed = Configuration.getInstance().getString("configurator.profile." + profile + ".red");
            final String newGreen = Configuration.getInstance().getString("configurator.profile." + profile + ".green");
            final String newBlue = Configuration.getInstance().getString("configurator.profile." + profile + ".blue");

            if (newOptions == null) {
                JOptionPane.showMessageDialog(this, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_PROFILENOTFOUND));
            }

            final int defaultBalance = 100;
            redSlider.setValue(newRed != null ? Integer.parseInt(newRed) : defaultBalance);
            greenSlider.setValue(newGreen != null ? Integer.parseInt(newGreen) : defaultBalance);
            blueSlider.setValue(newBlue != null ? Integer.parseInt(newBlue) : defaultBalance);

            parse(newOptions);
            repaint();
        }
    }

    /**
     * Save the selected profile.
     */
    private void saveProfile() {
        profilesComboBox.setSelectedItem(profilesComboBox.getEditor().getItem());
        final Object selectedProfile = profilesComboBox.getSelectedItem();
        if (selectedProfile == null || selectedProfile.toString().isEmpty()) {
            JOptionPane.showMessageDialog(this, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_ENTERPROFILENAME));
        } else if (selectedProfile.toString().contains("|")) {
            JOptionPane.showMessageDialog(this, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_NOPIPEINPROFILE));
        } else {
            final String profile = selectedProfile.toString();
            String profiles = "";

            final ComboBoxModel<String> model = profilesComboBox.getModel();
            boolean found = false;
            for (int i = 0; i < model.getSize(); i++) {
                if (model.getElementAt(i).trim().equals(profile.trim())) {
                    found = true;
                }
                if (!profiles.isEmpty()) {
                    profiles += "|";
                }
                profiles += model.getElementAt(i).trim();
            }
            if (!found) {
                profilesComboBox.addItem(profile.trim());

                if (!profiles.isEmpty()) {
                    profiles += "|";
                }
                profiles += profile.trim();
            }

            computeOptions();

            Configuration.getInstance().setString("configurator.profiles", profiles);
            Configuration.getInstance().setString("configurator.profile." + profile, options);
            Configuration.getInstance().setString("configurator.profile." + profile + ".red", Integer.toString(redSlider.getValue()));
            Configuration.getInstance().setString("configurator.profile." + profile + ".green", Integer.toString(greenSlider.getValue()));
            Configuration.getInstance().setString("configurator.profile." + profile + ".blue", Integer.toString(blueSlider.getValue()));
        }
    }

    /**
     * Delete the selected profile.
     */
    private void deleteProfile() {
        profilesComboBox.setSelectedItem(profilesComboBox.getEditor().getItem());
        final Object selectedProfile = profilesComboBox.getSelectedItem();
        if (selectedProfile == null || selectedProfile.toString().isEmpty()) {
            JOptionPane.showMessageDialog(this, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_ENTERPROFILENAME));
        } else {
            final String profile = selectedProfile.toString();
            String profiles = "";

            final ComboBoxModel<String> model = profilesComboBox.getModel();
            for (int i = 0; i < model.getSize(); i++) {
                if (!model.getElementAt(i).trim().equals(profile.trim())) {
                    if (!profiles.isEmpty()) {
                        profiles += "|";
                    }
                    profiles += model.getElementAt(i).trim();
                }
            }

            computeOptions();

            Configuration.getInstance().setString("configurator.profiles", profiles);
            profilesComboBox.removeItem(profile);
        }
    }

    /**
     * The enum defining all the close options.
     *
     * @author Zleurtor
     */
    public static enum CloseOption {
        /**
         * The option selected when the user close the dialog using the "cancel" button.
         */
        CANCEL,
        /**
         * The option selected when the user close the dialog using the "apply to selected" button.
         */
        APPLY_SELECTED,
        /**
         * The option selected when the user close the dialog using the "apply to all" button.
         */
        APPLY_ALL
    }

    /**
     * A class displaying an image.
     *
     * @author Zleurtor
     */
    private final class ColorsSamplePanel extends JPanel implements ComponentListener {

        private static final long serialVersionUID = -3810245420909334443L;

        /**
         * The currently displayed sample image file.
         */
        private File colorsSample;

        /**
         * The currently displayed sample image.
         */
        private BufferedImage colorsSampleImage;

        /**
         * The currently displayed sample image.
         */
        private BufferedImage scaledImage;

        /**
         * Creates a new ColorsSamplePanel.
         */
        private ColorsSamplePanel() {
            addComponentListener(this);
        }

        /**
         * Sets a new currently displayed sample image.
         *
         * @param newColorsSample The new currently displayed sample image.
         * @throws IOException If an error occurs while reading the image.
         */
        public void setColorsSample(final File newColorsSample) throws IOException {
            this.colorsSample = newColorsSample;
            if (colorsSample != null) {
                colorsSampleImage = ImageIO.read(colorsSample);
            } else {
                colorsSampleImage = null;
            }

            drawScaled();
        }

        /**
         * Draw the sample image scaled.
         */
        private void drawScaled() {
            if (colorsSampleImage != null && scaledImage != null) {
                final int x;
                final int y;
                final int w;
                final int h;
                final double r = (double) colorsSampleImage.getWidth() / colorsSampleImage.getHeight();

                if (r > (double) scaledImage.getWidth() / scaledImage.getHeight()) {
                    w = scaledImage.getWidth();
                    h = (int) Math.round(scaledImage.getWidth() / r);
                    x = 0;
                    y = (scaledImage.getHeight() - h) / 2;
                } else {
                    w = (int) Math.round(scaledImage.getHeight() * r);
                    h = scaledImage.getHeight();
                    x = (scaledImage.getWidth() - w) / 2;
                    y = 0;
                }

                final Graphics2D g = (Graphics2D) scaledImage.getGraphics();
                g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
                g.drawImage(
                        colorsSampleImage, x, y, x + w, y + h, 0, 0, colorsSampleImage.getWidth(), colorsSampleImage.getHeight(), this);
            } else {
                scaledImage = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_ARGB);
            }

            final int red = redSlider.getValue();
            final int green = greenSlider.getValue();
            final int blue = blueSlider.getValue();

            for (int x = 0; x < scaledImage.getWidth(); x++) {
                for (int y = 0; y < scaledImage.getHeight(); y++) {
                    final Color c = new Color(scaledImage.getRGB(x, y), true);
                    final Color balanced = new Color(
                            Math.min(255, (int) Math.round((c.getRed() / 100d) * red)),
                            Math.min(255, (int) Math.round((c.getGreen() / 100d) * green)),
                            Math.min(255, (int) Math.round((c.getBlue() / 100d) * blue)),
                            c.getAlpha());
                    scaledImage.setRGB(x, y, balanced.getRGB());
                }
            }

            repaint();
        }

        @Override
        protected void paintComponent(final Graphics g) {
            super.paintComponent(g);

            g.drawImage(scaledImage, 0, 0, this);
        }

        @Override
        public void componentResized(final ComponentEvent evt) {
            scaledImage = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_ARGB);
            drawScaled();
        }

        @Override
        public void componentMoved(final ComponentEvent evt) { }

        @Override
        public void componentShown(final ComponentEvent evt) { }

        @Override
        public void componentHidden(final ComponentEvent evt) { }

    }

    /**
     * The class listening for change events on the red green and bleu sliders.
     *
     * @author Zleurtor
     */
    private final class ColorsChangeListener implements ChangeListener {

        /**
         * The panel containing the sample images.
         */
        private final ColorsSamplePanel panel;

        /**
         * Construct a new colors change listener impacting the given panel.
         *
         * @param thePanel The panel containing the sample images.
         */
        private ColorsChangeListener(final ColorsSamplePanel thePanel) {
            this.panel = thePanel;
        }

        @Override
        public void stateChanged(final ChangeEvent evt) {
            panel.drawScaled();
        }

    }

}
