/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.io.BufferedReader;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;

import org.apache.log4j.lf5.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.control.LogLevelAction;
import zleurtor.scanmanager.view.ResourcesManager;

/**
 * A dialog allowing to display log (with syntax highlighting).
 *
 * @author Zleurtor
 */
public class LogDialog extends JDialog {

    private static final long serialVersionUID = 7476694054074966486L;

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(LogDialog.class);

    /**
     * The text component displaying the log.
     */
    private final JTextPane textPane;

    /**
     * The label displaying a wait animation.
     */
    private final WaitPanel waitPanel = new WaitPanel();

    /**
     * The label displaying a wait message.
     */
    private final JLabel waitLabel = new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_LOADINGLOG));

    /**
     * The last foreground color used (in order to well color the stack traces).
     */
    private Color lastColor = ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_OTHER);

    /**
     * The action performed when the user change the log level.
     */
    private final LogLevelAction logLevelAction;

    /**
     * The name of the slave for which the log is displayed.
     */
    private final String name;

    /**
     * The number of lines.
     */
    private int lineNumber = 0;

    /**
     * Creates a log dialog for the given slaves panel and displaying the log of the specified IP.
     *
     * @param slavesPanel The slaves panel for which this dialog is created.
     * @param newIp The IP of the slave of which the log is displayed.
     * @param newName The name of the slave of which the log is displayed.
     * @param logReader A input stream reader used to read the log content.
     * @param logLevel The level of the displayed log.
     */
    public LogDialog(final SlavesPanel slavesPanel, final String newIp, final String newName,
            final BufferedReader logReader, final LogLevel logLevel) {
        super(SwingUtilities.getWindowAncestor(slavesPanel), ModalityType.APPLICATION_MODAL);
        logLevelAction = new LogLevelAction(slavesPanel.getMaster());
        this.name = newName;

        // Create and configure the text panel
        textPane = new JTextPane();
        textPane.setPreferredSize(new Dimension(
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_LOGPANEL_WIDTH),
                ResourcesManager.getInstance().getInt(ResourcesManager.KEY_LOGPANEL_HEIGHT)));
        textPane.setBackground(ResourcesManager.getInstance().getColor(ResourcesManager.KEY_PANEL_BACKGROUND));
        textPane.setOpaque(false);
        textPane.setSelectionColor(ResourcesManager.getInstance().getColor(ResourcesManager.KEY_TEXTFIELD_SELECTIONBACKGROUND));
        textPane.setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
        textPane.setAutoscrolls(true);
        textPane.setEditable(false);

        final JScrollPane scrollPane = new JScrollPane(textPane,
                JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scrollPane.setOpaque(false);
        scrollPane.getVerticalScrollBar().setOpaque(false);
        scrollPane.getHorizontalScrollBar().setOpaque(false);

        // Configure the dialog
        setTitle(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_LOGDIALOG, newIp));
        setContentPane(new ContentPanel());
        getContentPane().setLayout(new GridBagLayout());

        // Create the controls and listeners
        final JButton selectAllButton =
                new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_SELECTALL));
        final JButton copyToClipBoardButton =
                new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_COPYTOCLIPBOARD));
        final JComboBox<LogLevel> levelComboBox = new JComboBox<>();

        selectAllButton.addActionListener((final ActionEvent evt) -> selectAll());
        copyToClipBoardButton.addActionListener((final ActionEvent evt) -> copyToClipboard());
        levelComboBox.addItem(LogLevel.DEBUG);
        levelComboBox.addItem(LogLevel.INFO);
        levelComboBox.addItem(LogLevel.WARN);
        levelComboBox.addItem(LogLevel.ERROR);
        levelComboBox.setSelectedItem(logLevel);
        levelComboBox.addItemListener((final ItemEvent evt) -> changeLogLevel(evt));

        // Layout the components
        final int gap = 5;
        final int scrollPaneWidth = 3;
        final int levelComboBoxColumn = 0;
        final int selectAllButtonColumn = 1;
        final int copyToClipboardButtonColumn = 2;

        add(scrollPane,
                new GridBagConstraints(0, 0, scrollPaneWidth, 2, 1, 1,
                        GridBagConstraints.CENTER, GridBagConstraints.BOTH,
                        new Insets(gap, gap, gap / 2, gap), 0, 0));
        add(waitPanel,
                new GridBagConstraints(0, 0, scrollPaneWidth, 1, 1, 1,
                        GridBagConstraints.SOUTH, GridBagConstraints.NONE,
                        new Insets(gap, gap, gap / 2, gap), 0, 0));
        add(waitLabel,
                new GridBagConstraints(0, 1, scrollPaneWidth, 1, 1, 1,
                        GridBagConstraints.NORTH, GridBagConstraints.NONE,
                        new Insets(gap, gap, gap / 2, gap), 0, 0));
        add(levelComboBox,
                new GridBagConstraints(levelComboBoxColumn, 2, 1, 1, 1, 0,
                        GridBagConstraints.WEST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap, gap, gap / 2), 0, 0));
        add(selectAllButton,
                new GridBagConstraints(selectAllButtonColumn, 2, 1, 1, 0, 0,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap, gap, gap / 2), 0, 0));
        add(copyToClipBoardButton,
                new GridBagConstraints(copyToClipboardButtonColumn, 2, 1, 1, 0, 0,
                        GridBagConstraints.EAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap), 0, 0));

        pack();
        setLocationRelativeTo(slavesPanel);

        // Read the content of the log and add it to a document
        final Thread t = new Thread(() -> {
            try {
                scrollPane.setVisible(false);
                waitPanel.setVisible(true);
                waitLabel.setVisible(true);

                final StyledDocument document = new DefaultStyledDocument();
                for (String line = logReader.readLine(); line != null; line = logReader.readLine()) {
                    addLine(line, document);
                }
                logReader.close();

                textPane.setDocument(document);
                textPane.setCaretPosition(document.getLength());

                waitLabel.setVisible(false);
                waitPanel.setVisible(false);
                scrollPane.setVisible(true);
            } catch (final IOException | BadLocationException e) {
                GuiUtils.showWarning(e, slavesPanel);
                LOGGER.warn(e.getMessage(), e);
            }
        });
        t.setName("Log reader");
        t.start();

        setVisible(true);
    }

    /**
     * Add a new log line to the text pane.
     *
     * @param line The new log line to add to the text pane.
     * @param document The document the line shall be added to.
     * @throws BadLocationException If the insert position is not a valid position within the document.
     */
    public final void addLine(final String line, final StyledDocument document) throws BadLocationException {
        final SimpleAttributeSet current = new SimpleAttributeSet();

        StyleConstants.setBold(current, true);

        if (line.contains("INFO")) {
            StyleConstants.setForeground(current,
                    ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_INFO));
        } else if (line.contains("WARN")) {
            StyleConstants.setForeground(current,
                    ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_WARN));
        } else if (line.contains("ERROR")) {
            StyleConstants.setForeground(current,
                    ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_ERROR));
        } else if (line.contains("DEBUG")) {
            StyleConstants.setForeground(current,
                    ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_DEBUG));
        } else if (line.contains("Exception") || line.trim().startsWith("at")) {
            StyleConstants.setForeground(current, lastColor);
        } else {
            StyleConstants.setForeground(current,
                    ResourcesManager.getInstance().getColor(ResourcesManager.KEY_LOGPANEL_FOREGROUND_OTHER));
        }

        document.insertString(document.getLength(), ++lineNumber + " - " + line + "\n", current);

        lastColor = StyleConstants.getForeground(current);
    }

    /**
     * Select all the log.
     */
    private void selectAll() {
        textPane.setSelectionStart(0);
        textPane.setSelectionEnd(textPane.getDocument().getLength());
    }

    /**
     * Copy the selected part of the log to the clipboard.
     */
    private void copyToClipboard() {
        final StringSelection selection = new StringSelection(textPane.getSelectedText());
        final Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        clipboard.setContents(selection, selection);
    }

    /**
     * Method invoked when the user change the log level.
     *
     * @param evt The event representing the log level changing.
     */
    private void changeLogLevel(final ItemEvent evt) {
        if (evt.getStateChange() == ItemEvent.SELECTED) {
            logLevelAction.actionPerformed(new ActionEvent(evt.getItem(), ActionEvent.ACTION_PERFORMED, name));
        }
    }

}
