/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import org.apache.log4j.lf5.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.control.LogButtonAction;
import zleurtor.scanmanager.control.ShutdownAction;
import zleurtor.scanmanager.control.StreamAction;
import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.ExifReader;
import zleurtor.scanmanager.model.Master;
import zleurtor.scanmanager.model.events.MasterEvent;
import zleurtor.scanmanager.model.events.MasterListener;
import zleurtor.scanmanager.view.ResourcesManager;
import zleurtor.scanmanager.view.SlaveInformation;

/**
 * The panel displaying all the slaves information.
 *
 * @author Zleurtor
 */
public class SlavesPanel extends JPanel
    implements ComponentListener, MasterListener, MouseListener, MouseMotionListener, KeyListener, ActionListener {

    private static final long serialVersionUID = -5461904024406510225L;

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(SlavesPanel.class);

    /**
     * The application configuration.
     */
    private static final Configuration CONFIGURATION = Configuration.getInstance();

    /**
     * The thumbnail width.
     */
    private static final int THUMBNAIL_WIDTH = CONFIGURATION.getInt("snapshot.thumbnailWidth");

    /**
     * The thumbnail height.
     */
    private static final int THUMBNAIL_HEIGHT = CONFIGURATION.getInt("snapshot.thumbnailHeight");

    /**
     * The thumbnail error image.
     */
    private static final Image THUMBNAIL_ERROR;

    /**
     * The thumbnail empty image.
     */
    private static final Image THUMBNAIL_EMPTY;

    /**
     * The stream mask image.
     */
    private static final Image STREAM;

    /**
     * The button log image.
     */
    private static final Image BUTTON_LOG;

    /**
     * The button log image.
     */
    private static final Image OVERFLY_BUTTON_LOG;

    /**
     * The button log image.
     */
    private static final Image ACTIVE_BUTTON_LOG;

    /**
     * The selection color.
     */
    private static final Color SELECTION_COLOR = ResourcesManager.getInstance().getColor(ResourcesManager.KEY_SLAVESPANEL_SELECTION);

    static {
        Image thumbnailError = null;
        try {
            thumbnailError = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_SNAPSHOTERROR);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }
        THUMBNAIL_ERROR = thumbnailError;

        Image thumbnailEmpty = null;
        try {
            thumbnailEmpty = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_SNAPSHOTEMPTY);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }
        THUMBNAIL_EMPTY = thumbnailEmpty;

        Image stream = null;
        try {
            stream = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_STREAM);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }
        STREAM = stream;

        Image buttonLog = null;
        try {
            buttonLog = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_ICON_LOGBUTTON);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }
        BUTTON_LOG = buttonLog;
        OVERFLY_BUTTON_LOG = ResourcesManager.colorize(buttonLog,
                ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE));
        ACTIVE_BUTTON_LOG = ResourcesManager.colorize(buttonLog,
                ResourcesManager.getInstance().getColor(ResourcesManager.KEY_BUTTON_ACTIVE).brighter());
    }

    /**
     * The slaves panel master.
     */
    private final Master master;

    /**
     * The list of all slaves information.
     */
    private final List<SlaveInformation> slaves = new CopyOnWriteArrayList<>();

    /**
     * The list of selected slaves information.
     */
    private final List<SlaveInformation> selected = new ArrayList<>();

    /**
     * The index of the last selected slave.
     */
    private int lastSelectedIndex = 0;

    /**
     * The map of slave thumbnail shapes.
     */
    private final Map<Rectangle, SlaveInformation> snapshots = Collections.synchronizedMap(new HashMap<>());

    /**
     * The map of slave shutdown button shapes.
     */
    private final Map<Rectangle, SlaveInformation> shutdownButtons = Collections.synchronizedMap(new HashMap<>());

    /**
     * The map of slave log button shapes.
     */
    private final Map<Rectangle, SlaveInformation> logButtons = Collections.synchronizedMap(new HashMap<>());

    /**
     * The overflown slave information.
     */
    private Rectangle overFlown = null;

    /**
     * The pressed slave information button's shape.
     */
    private Rectangle pressed = null;

    /**
     * The action triggered when the user click on a slave snapshot.
     */
    private final StreamAction streamAction;

    /**
     * The action triggered when the user click on a slave shutdown button.
     */
    private final ShutdownAction shutdownAction;

    /**
     * The action triggered when the user click on a slave log button.
     */
    private final LogButtonAction logAction;

    /**
     * The options text field.
     */
    private final JTextField optionsField;

    /**
     * Whether or not we are waiting for a snapshot in order to play the "quack" sound.
     */
    private boolean waitForQuack = false;

    /**
     * The containing scrollpane.
     */
    private JScrollPane scrollPane;

    /**
     * The buffer used to paint the slaves thumbnails.
     */
    private final BufferedImage thumbnailBuffer = new BufferedImage(THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT, BufferedImage.TYPE_INT_ARGB);

    /**
     * Constructs a new slaves panel using the specified master object.
     *
     * @param newMaster The master object which shall execute slaves panel actions.
     * @param theOptionsField The options text field.
     */
    public SlavesPanel(final Master newMaster, final JTextField theOptionsField) {
        super();
        this.master = newMaster;
        this.optionsField = theOptionsField;

        streamAction = new StreamAction(master);
        shutdownAction = new ShutdownAction(master, this);
        logAction = new LogButtonAction(master);
        addMouseListener(this);
        addMouseMotionListener(this);
        addKeyListener(this);
        setFocusable(true);
        setOpaque(false);

        theOptionsField.addActionListener(this);

        setFont(new Font(getFont().getName(), Font.BOLD, getFont().getSize() + 1));
    }

    /**
     * Returns the slaves panel master.
     *
     * @return The slaves panel master.
     */
    public final Master getMaster() {
        return master;
    }

    /**
     * Returns the list of all slaves information.
     *
     * @return The list of all slaves information.
     */
    public final List<SlaveInformation> getSlaves() {
        return slaves;
    }

    /**
     * Returns the list of selected slaves information.
     *
     * @return The list of selected slaves information.
     */
    public final List<SlaveInformation> getSelected() {
        return selected;
    }

    /**
     * Sets whether or not we are waiting for a snapshot in order to play the "quack" sound.
     *
     * @param newWaitForQuack true if we are waiting for a snapshot, false otherwise.
     */
    public final void setWaitForQuack(final boolean newWaitForQuack) {
        this.waitForQuack = newWaitForQuack;
    }

    @Override
    protected final void paintComponent(final Graphics g) {
        super.paintComponent(g);

        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);

        if (slaves.size() == 0) {
            return;
        }

        snapshots.clear();
        shutdownButtons.clear();
        logButtons.clear();

        final Dimension cellSize = computeCellDimension(g);

        final Stroke stroke = ((Graphics2D) g).getStroke();

        int maxLine = Integer.MIN_VALUE;
        int maxColumn = Integer.MIN_VALUE;
        for (final SlaveInformation si : slaves) {
            maxLine = (int) Math.max(maxLine, computeSlaveLocation(si)[1]);
            maxColumn = (int) Math.max(maxColumn, computeSlaveLocation(si)[2]);
        }
        for (int l = 1; l <= maxLine; l++) {
            for (int c = 0; c <= maxColumn; c++) {
                final int x = c * cellSize.width;
                final int y = (l - 1) * cellSize.height;

                final int alpha = 63;
                final float strokeWidth = 0.5f;
                g.setColor(new Color(
                        getForeground().getRed(),
                        getForeground().getGreen(),
                        getForeground().getBlue(),
                        alpha));
                ((Graphics2D) g).setStroke(new BasicStroke(strokeWidth, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));

                g.drawRect(x, y, cellSize.width - 1, cellSize.height - 1);
            }
        }

        ((Graphics2D) g).setStroke(stroke);
        g.setColor(getForeground());

        for (final SlaveInformation si : slaves) {
            paintSlave(g, si, cellSize);
        }
    }

    /**
     * Paint the slave specified by si at the ith position.
     *
     * @param g The graphics to which the slave shall be painted.
     * @param si The information on the slave to paint.
     * @param cellSize The minimum size of a cell.
     */
    private void paintSlave(final Graphics g, final SlaveInformation si, final Dimension cellSize) {
        final int[] location = computeSlaveLocation(si);
        final int w = cellSize.width;
        final int h = cellSize.height;

        // Paint cells borders
        final int l = location[1] - 1;
        final int c = location[2];

        int x = c * w;
        int y = l * h;
        if (selected.contains(si)) {
            final Color previous = g.getColor();
            g.setColor(SELECTION_COLOR);
            g.fillRect(x, y, w, h);
            g.setColor(previous);
        }
        g.drawRect(x, y, w - 1, h - 1);

        // Draw IP and mac
        final int interline = 2;
        final int arc = 5;
        final String nameVersion = si.getName() + " - v" + si.getVersion();
        g.setColor(getBackground());
        g.fillRoundRect(x + 1, y + 1,
                g.getFontMetrics().stringWidth(si.getIp()) + 2, g.getFontMetrics().getHeight() + 2,
                arc, arc);
        g.fillRoundRect(x + 1, y + 1 + g.getFontMetrics().getHeight() + interline,
                g.getFontMetrics().stringWidth(nameVersion) + 2, g.getFontMetrics().getHeight() + 2,
                arc, arc);
        g.setColor(getForeground());
        g.drawString(si.getIp(), x + 2, y + 2 + g.getFontMetrics().getAscent());
        g.drawString(nameVersion, x + 2, y + 2 + g.getFontMetrics().getHeight() + interline + g.getFontMetrics().getAscent());

        x = x + w / 2 - THUMBNAIL_WIDTH / 2;
        y = y + h - 2 - THUMBNAIL_HEIGHT;

        // Draw the thumbnail
        if (si.getThumbnail() == null) {
            g.fillRect(x, y, THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT);
        } else {
            thumbnailBuffer.getGraphics().clearRect(x, y, thumbnailBuffer.getWidth(), thumbnailBuffer.getHeight());
            thumbnailBuffer.getGraphics().drawImage(si.getThumbnail(),
                    0, 0, thumbnailBuffer.getWidth(), thumbnailBuffer.getHeight(),
                    0, 0, si.getThumbnail().getWidth(this), si.getThumbnail().getHeight(this),
                    this);

            if (si.getThumbnail() != THUMBNAIL_EMPTY && si.getThumbnail() != THUMBNAIL_ERROR) {
                applyColorBalance(si);
            }

            g.drawImage(thumbnailBuffer, x, y, this);
            if (si.getThumbnail() != THUMBNAIL_ERROR) {
                snapshots.put(new Rectangle(x, y, THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT), si);
            }
        }

        // Draw the stream mask
        if (snapshots.get(overFlown) == si) {
            g.drawImage(STREAM,
                    x, y, x + THUMBNAIL_WIDTH, y + THUMBNAIL_HEIGHT,
                    0, 0, STREAM.getWidth(this), STREAM.getHeight(this),
                    this);
        }

        x = c * w;
        y = l * h;
        final int bw = BUTTON_LOG.getWidth(this);
        final int bh = BUTTON_LOG.getHeight(this);
        logButtons.put(new Rectangle(x + w - bw - 2, y + h - bh - 2, bw, bh), si);

        final Image img = (logButtons.get(pressed) == si) ? ACTIVE_BUTTON_LOG
                : (logButtons.get(overFlown) == si) ? OVERFLY_BUTTON_LOG : BUTTON_LOG;
        g.drawImage(img,
                x + w - bw - 2, y + h - bh - 2, x + w - 2, y + h - 2,
                0, 0, bw, bh,
                this);
    }

    /**
     * Apply the color balance of the specified slave information to the thumbnail buffer.
     *
     * @param si The slave information of which the color balance shall be applied to the thumbnail buffer.
     */
    private void applyColorBalance(final SlaveInformation si) {
        final int maxColorComponent = 255;
        for (int xx = 0; xx < thumbnailBuffer.getWidth(); xx++) {
            for (int yy = 0; yy < thumbnailBuffer.getHeight(); yy++) {
                final Color cc = new Color(thumbnailBuffer.getRGB(xx, yy));
                thumbnailBuffer.setRGB(xx, yy,
                        new Color(
                                (int) Math.min(maxColorComponent,
                                        Math.round((cc.getRed() / (double) SlaveInformation.DEFAULT_COLOR_BALANCE)
                                                * si.getColorBalance()[0])),
                                (int) Math.min(maxColorComponent,
                                        Math.round((cc.getGreen() / (double) SlaveInformation.DEFAULT_COLOR_BALANCE)
                                                * si.getColorBalance()[1])),
                                (int) Math.min(maxColorComponent,
                                        Math.round((cc.getBlue() / (double) SlaveInformation.DEFAULT_COLOR_BALANCE)
                                                * si.getColorBalance()[2]))).getRGB());
            }
        }
    }

    @Override
    public final void componentResized(final ComponentEvent e) {
        scrollPane = ((JScrollPane) e.getComponent());

        adjustSize();
    }

    @Override
    public final void componentMoved(final ComponentEvent e) { }

    @Override
    public final void componentShown(final ComponentEvent e) { }

    @Override
    public final void componentHidden(final ComponentEvent e) { }

    @Override
    public final void slavesCleared() {
        slaves.clear();

        repaint();
    }

    @Override
    public final void newSlave(final MasterEvent evt) {
        final SlaveInformation si = new SlaveInformation(evt.getName(), evt.getIp(), THUMBNAIL_EMPTY);
        si.setOptions(((String[]) evt.getData())[0]);
        si.setVersion(((String[]) evt.getData())[1]);
        slaves.add(si);
        Collections.sort(slaves);
        adjustSize();

        repaint();
    }

    @Override
    public final void slaveQuit(final MasterEvent evt) {
        if (evt.getIp() == null) {
            return;
        }

        SlaveInformation toRemove = null;
        for (final SlaveInformation si : slaves) {
            if (si.getIp().equals(evt.getIp())) {
                toRemove = si;
            }
        }
        if (toRemove != null) {
            slaves.remove(toRemove);
        }

        displayOptions();
        repaint();
    }

    @Override
    public final synchronized void snapshotSuccess(final MasterEvent evt) {
        final File path = new File(CONFIGURATION.getString("snapshot.masterFolder") + File.separator + evt.getData());
        final File thumbnailPath = new File(path.getParentFile(), "thumbnails\\" + File.separator + path.getName());
        thumbnailPath.getParentFile().mkdirs();

        // Try to load the thumbnail from the exif data
        Image thumbnail = null;
        try (OutputStream out = new FileOutputStream(thumbnailPath)) {
            final ExifReader exifReader = new ExifReader(path);
            out.write(exifReader.getThumbnail());
            thumbnail = Toolkit.getDefaultToolkit().getImage(thumbnailPath.getPath());
        } catch (final IOException e) {
            LOGGER.warn(e.getMessage(), e);
        }

        if (thumbnail == null) {
            // Didn't retrieve the thumbnail, we construct it from the complete image
            Image tmp = Toolkit.getDefaultToolkit().getImage(path.getPath());
            thumbnail = tmp.getScaledInstance(CONFIGURATION.getInt("snapshot.thumbnailWidth"),
                    CONFIGURATION.getInt("snapshot.thumbnailHeight"), Image.SCALE_SMOOTH);

            tmp.flush();
            tmp = null;
            System.gc();
        }

        for (final SlaveInformation si : slaves) {
            if (si.getIp().equals(evt.getIp())) {
                flushThumbnail(si);
                si.setThumbnail(thumbnail);
                si.setSnapshotFilename(path.getPath());
            }
        }
        if (waitForQuack && Boolean.parseBoolean(CONFIGURATION.getString("sound"))) {
            waitForQuack = false;
            playQuack();
        }

        repaint();
    }

    @Override
    public final synchronized void snapshotError(final MasterEvent evt) {
        for (final SlaveInformation si : slaves) {
            if (si.getIp().equals(evt.getIp())) {
                flushThumbnail(si);
                si.setThumbnail(THUMBNAIL_ERROR);
                si.setSnapshotFilename(null);
            }
        }
        if (waitForQuack && Boolean.parseBoolean(CONFIGURATION.getString("sound"))) {
            waitForQuack = false;
            playQuack();
        }

        repaint();

        GuiUtils.showError(evt.getData().toString(), this);
    }

    @Override
    public final void snapshotsCleared() {
        for (final SlaveInformation si : slaves) {
            flushThumbnail(si);
            si.setThumbnail(THUMBNAIL_EMPTY);
            si.setSnapshotFilename(null);
        }
        repaint();
    }

    @Override
    public final void streamError(final MasterEvent evt) {
        synchronized (slaves) {
            for (final SlaveInformation si : slaves) {
                if (si.getIp().equals(evt.getIp())) {
                    flushThumbnail(si);
                    si.setThumbnail(THUMBNAIL_ERROR);
                    si.setSnapshotFilename(null);
                }
            }
        }
        repaint();

        GuiUtils.showWarning(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_STREAMERROR), this);
        LOGGER.warn("Stream error.");
    }

    @Override
    public final void log(final MasterEvent evt, final LogLevel level) {
        final BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream((byte[]) evt.getData())));
        new LogDialog(this, evt.getIp(), evt.getName(), reader, level);
    }

    @Override
    public final void logLevelError(final MasterEvent evt) {
        JOptionPane.showMessageDialog(
                this,
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_LOGLEVELERROR, evt.getIp(), evt.getData()),
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_ERROR),
                JOptionPane.ERROR_MESSAGE);
    }

    @Override
    public final void logLevelSuccess(final MasterEvent evt) {
        JOptionPane.showMessageDialog(
                this,
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_LOGLEVELSUCCESS, evt.getIp(), evt.getData()),
                ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_INFO),
                JOptionPane.INFORMATION_MESSAGE);
    }

    @Override
    public final void options(final MasterEvent evt) {
        for (final SlaveInformation slave : slaves) {
            if (slave.getIp().equals(evt.getIp())) {
                slave.setOptions((String) evt.getData());
                displayOptions();
            }
        }
    }

    @Override
    public void fileSaved(final MasterEvent evt) { }

    @Override
    public final void mouseDragged(final MouseEvent evt) { }

    @Override
    public final void mouseMoved(final MouseEvent evt) {
        final Rectangle previous = overFlown;
        overFlown = null;

        for (final Rectangle r : snapshots.keySet()) {
            if (r.contains(evt.getPoint())) {
                overFlown = r;
                setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
                break;
            }
        }

        for (final Rectangle r : shutdownButtons.keySet()) {
            if (r.contains(evt.getPoint())) {
                overFlown = r;
                setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
                break;
            }
        }

        for (final Rectangle r : logButtons.keySet()) {
            if (r.contains(evt.getPoint())) {
                overFlown = r;
                setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
                break;
            }
        }

        if (overFlown == null) {
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }

        if (previous != overFlown) {
            repaint();
        }
    }

    @Override
    public final void mouseClicked(final MouseEvent evt) { }

    @Override
    public final void mousePressed(final MouseEvent evt) {
        final Rectangle oldPressed = pressed;
        pressed = null;

        for (final Rectangle r : shutdownButtons.keySet()) {
            if (r.contains(evt.getPoint())) {
                pressed = r;
            }
        }

        for (final Rectangle r : logButtons.keySet()) {
            if (r.contains(evt.getPoint())) {
                pressed = r;
            }
        }

        if (oldPressed == null && pressed != null
                || oldPressed != null && pressed != null && !oldPressed.equals(pressed)) {
            repaint();
        }

        requestFocus();
    }

    @Override
    public final void mouseReleased(final MouseEvent evt) {
        if (pressed != null) {
            pressed = null;
            repaint();
        }

        if (overFlown != null
                && (evt.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) == 0
                && (evt.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) == 0
                && SwingUtilities.isLeftMouseButton(evt)) {
            buttonClick();
        } else if (overFlown == null
                && !slaves.isEmpty()
                && SwingUtilities.isLeftMouseButton(evt)) {
            changeSelection(evt);
        }
    }

    @Override
    public final void mouseEntered(final MouseEvent evt) {
        mouseMoved(evt);
    }

    @Override
    public final void mouseExited(final MouseEvent evt) {
        overFlown = null;
    }

    /**
     * The method to used when the user click on a slave's button.
     */
    private void buttonClick() {
        // Click without any modifiers
        if (snapshots.containsKey(overFlown)) {
            streamAction.actionPerformed(new ActionEvent(snapshots.get(overFlown), ActionEvent.ACTION_PERFORMED, "Stream"));
        } else if (shutdownButtons.containsKey(overFlown)) {
            shutdownAction.actionPerformed(new ActionEvent(shutdownButtons.get(overFlown), ActionEvent.ACTION_PERFORMED, "Shutdown"));
        } else if (logButtons.containsKey(overFlown)) {
            logAction.actionPerformed(new ActionEvent(logButtons.get(overFlown), ActionEvent.ACTION_PERFORMED, "Log"));
        }
    }

    /**
     * Change the selection depending on the user's mouse event.
     *
     * @param evt An event which indicates that a mouse action occurred in a component.
     */
    private void changeSelection(final MouseEvent evt) {
        final Dimension cellSize = computeCellDimension(getGraphics());

        int index = -1;
        for (int i = 0; i < slaves.size(); i++) {
            final int[] location = computeSlaveLocation(slaves.get(i));
            final int l = location[1] - 1;
            final int c = location[2];
            final int x = c * cellSize.width;
            final int y = l * cellSize.height;

            final Rectangle rect = new Rectangle(x, y, cellSize.width, cellSize.height);
            if (rect.contains(evt.getPoint())) {
                index = i;
                break;
            }
        }

        int delta;
        try {
            delta = (index - lastSelectedIndex) / Math.abs(index - lastSelectedIndex);
        } catch (final ArithmeticException e) {
            delta = 1;
        }

        if ((evt.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0
                && (evt.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0) {
            ctrlShiftSelection(index, delta);
        } else if ((evt.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0) {
            ctrlSelection(index);
        } else if ((evt.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0) {
            shiftSelection(index, delta);
        } else {
            noModifierSelection(index);
        }

        repaint();
        displayOptions();
    }

    /**
     * Change the selection when the user execute a mouse action with ctrl and shift modifiers.
     *
     * @param index The index (in the list of slave of the clicked save).
     * @param delta -1 or +1 for respectively a new selected slave before of ater the previous one
     */
    private void ctrlShiftSelection(final int index, final int delta) {
        // CTRL + SHIFT
        for (int i = lastSelectedIndex;
                i >= Math.min(index, lastSelectedIndex) && i <= Math.max(index, lastSelectedIndex) && i < slaves.size() && i >= 0;
                i += delta) {
            selected.add(slaves.get(i));
        }
        lastSelectedIndex = index;
    }

    /**
     * Change the selection when the user execute a mouse action with shift modifier.
     *
     * @param index The index (in the list of slave of the clicked save).
     * @param delta -1 or +1 for respectively a new selected slave before of ater the previous one
     */
    private void shiftSelection(final int index, final int delta) {
        // SHIFT
        selected.clear();
        for (int i = lastSelectedIndex;
                i >= Math.min(index, lastSelectedIndex) && i <= Math.max(index, lastSelectedIndex) && i < slaves.size() && i >= 0;
                i += delta) {
            selected.add(slaves.get(i));
        }
    }

    /**
     * Change the selection when the user execute a mouse action with ctrl modifier.
     *
     * @param index The index (in the list of slave of the clicked save).
     */
    private void ctrlSelection(final int index) {
        // CTRL
        if (selected.contains(slaves.get(index))) {
            selected.remove(slaves.get(index));
        } else {
            selected.add(slaves.get(index));
        }
        lastSelectedIndex = index;
    }

    /**
     * Change the selection when the user execute a mouse action without modifier.
     *
     * @param index The index (in the list of slave of the clicked save).
     */
    private void noModifierSelection(final int index) {
        // No modifier
        selected.clear();
        try {
            selected.add(slaves.get(index));
            lastSelectedIndex = index;
        } catch (final IndexOutOfBoundsException e) {
            lastSelectedIndex = 0;
        }
    }

    @Override
    public final void keyTyped(final KeyEvent evt) { }

    @Override
    public final void keyPressed(final KeyEvent evt) { }

    @Override
    public final void keyReleased(final KeyEvent evt) {
        if ((evt.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0
                && evt.getKeyCode() == KeyEvent.VK_A) {
            selected.clear();
            selected.addAll(slaves);

            repaint();
            displayOptions();
        }
    }

    @Override
    public final void actionPerformed(final ActionEvent evt) {
        if (selected.size() > 0) {
            for (final SlaveInformation si : selected) {
                master.setOptions(optionsField.getText(), si.getName());
            }
        } else {
            for (final SlaveInformation si : slaves) {
                master.setOptions(optionsField.getText(), si.getName());
            }
        }
    }

    /**
     * Adjust the cells size in order to have complete rows.
     */
    private void adjustSize() {
        if (slaves.size() == 0 || getGraphics() == null) {
            return;
        }

        // Compute cell size
        final Dimension cellSize = computeCellDimension(getGraphics());
        if (cellSize.getWidth() == 0) {
            return;
        }

        int maxLine = Integer.MIN_VALUE;
        int maxColumn = Integer.MIN_VALUE;
        for (final SlaveInformation si : slaves) {
            maxLine = (int) Math.max(maxLine, computeSlaveLocation(si)[1]);
            maxColumn = (int) Math.max(maxColumn, computeSlaveLocation(si)[2]);
        }

        // Compute the needed size
        final Dimension size = new Dimension((maxColumn + 1) * cellSize.width, maxLine * cellSize.height);
        setPreferredSize(size);
        setMinimumSize(size);
        setSize(size);

        scrollPane.revalidate();
        scrollPane.repaint();
    }

    /**
     * Compute the minimum dimension of a cell.
     *
     * @param g The graphics in which the cells shall be paint.
     * @return The minimum dimension of a cell.
     */
    private Dimension computeCellDimension(final Graphics g) {
        final FontMetrics metrics = g.getFontMetrics();

        int width = 0;
        int height = 0;
        for (final SlaveInformation si : slaves) {
            width = Math.max(width, 2 + metrics.stringWidth(si.getIp() + 2));
            width = Math.max(width, 2 + metrics.stringWidth(si.getName() + 2));
            final int bw = BUTTON_LOG.getWidth(this);
            width = Math.max(width, 2 + bw + THUMBNAIL_WIDTH + bw + 2);

            final int interline = 3;
            height = Math.max(height, 2 + metrics.getHeight() + interline + metrics.getHeight() + interline + THUMBNAIL_HEIGHT + 2);
        }

        return new Dimension(width, height);
    }

    /**
     * Compute the location (on the grid) of a slave, depending on its information.
     *
     * @param si The information on the slave we want to tcompute the location.
     * @return The location (on the grid) of the specified slave
     */
    private static int[] computeSlaveLocation(final SlaveInformation si) {
        final int lastIpByteIndex = 3;
        final int byteMaxLength = 3;

        String lastByte = si.getIp().split("\\.")[lastIpByteIndex];
        while (lastByte.length() < byteMaxLength) {
            lastByte = "0" + lastByte;
        }

        final int columnIndex = 0;
        final int lineIndex = 2;
        final int column = Integer.parseInt(lastByte.substring(columnIndex, columnIndex + 2));
        int line = Integer.parseInt(lastByte.substring(lineIndex, lineIndex + 1));

        // Hack for shifted slave
        final int shiftedSlavesColumn = 7;
        final int firstShiftedSlaveLine = 3;
        if (column == shiftedSlavesColumn && line >= firstShiftedSlaveLine) {
            line--;
        }

        // For test purpose: if the slave is on the column 0, put it on column 1
        // A physical column on a graphical line
        final int ipGap = 10;
        return new int[] { 0, column > ipGap ? column - ipGap : column, line };
    }

    /**
     * Display the slave option in the options text field.
     */
    public final void displayOptions() {
        for (final SlaveInformation slave : selected) {
            if (slaves.contains(slave)) {
                optionsField.setText(slave.getOptions());
                return;
            }
        }

        // No slave found, empty the options field
        optionsField.setText("");
    }

    /**
     * Flush the slave thumbnail image if needed.
     *
     * @param si The slave of which the thumbnail image shall be flushed.
     */
    private void flushThumbnail(final SlaveInformation si) {
        if (si.getThumbnail() != null && si.getThumbnail() != THUMBNAIL_EMPTY && si.getThumbnail() != THUMBNAIL_ERROR) {
            si.getThumbnail().flush();
            si.setThumbnail(null);
            System.gc();
        }
    }

    /**
     * Play the "quack" sound.
     */
    private void playQuack() {
        try {
            final AudioInputStream stream = ResourcesManager.getInstance().getAudio(ResourcesManager.KEY_SOUND_QUACK);
            final AudioFormat format = stream.getFormat();
            final DataLine.Info info = new DataLine.Info(Clip.class, format);
            final Clip clip = (Clip) AudioSystem.getLine(info);
            clip.open(stream);
            clip.start();
        } catch (final UnsupportedAudioFileException | IOException | LineUnavailableException e) {
            GuiUtils.showWarning(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_STREAMERROR), this);
            LOGGER.warn(e.getMessage(), e);
        }
    }

    /**
     * Compute the color balance common to all slaves. If the slaves don't have a common color balance, 100% is returned.
     *
     * @return The color balance common to all slaves, or 100% if the slaves don't have a common color balance.
     */
    public int[] getCommonColorBalance() {
        final int[] colorBalance = new int[] { -1, -1, -1};

        for (final SlaveInformation si : selected) {
            for (int i = 0; i < colorBalance.length; i++) {
                if (colorBalance[i] == -1) {
                    colorBalance[i] = si.getColorBalance()[i];
                } else if (colorBalance[i] != si.getColorBalance()[i]) {
                    colorBalance[i] = SlaveInformation.DEFAULT_COLOR_BALANCE;
                }
            }
        }

        colorBalance[0] = colorBalance[0] != -1 ? colorBalance[0] : SlaveInformation.DEFAULT_COLOR_BALANCE;
        colorBalance[1] = colorBalance[1] != -1 ? colorBalance[1] : SlaveInformation.DEFAULT_COLOR_BALANCE;
        colorBalance[2] = colorBalance[2] != -1 ? colorBalance[2] : SlaveInformation.DEFAULT_COLOR_BALANCE;
        return colorBalance;
    }

}
