/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.GridBagConstraints;
import java.awt.Insets;

import javax.swing.JDialog;
import javax.swing.JLabel;

import org.apache.log4j.lf5.LogLevel;

import zleurtor.scanmanager.model.events.MasterEvent;
import zleurtor.scanmanager.model.events.MasterListener;
import zleurtor.scanmanager.view.ResourcesManager;

/**
 * The dialog displaying information about the update events.
 *
 * @author Zleurtor
 */
public class SnapshotsDialog extends AbstractProgressDialog implements MasterListener {

    private static final long serialVersionUID = -8823495552344353036L;

    /**
     * The label displaying update download information.
     */
    private final JLabel label =
            new JLabel(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TEXT_AWAITINGSNAPSHOTS));

    /**
     * The number of awaited images.
     */
    private int totalImages = 0;

    /**
     * The number of received images.
     */
    private volatile int receivedImages = 0;

    /**
     * Constructs a new update dialog.
     */
    public SnapshotsDialog() {
        super();

        setTitle(ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_SNAPSHOTSDIALOG));

        final int margin = ResourcesManager.getInstance().getInt(ResourcesManager.KEY_WAITPANEL_MARGIN);

        add(label,
                new GridBagConstraints(
                        0, 0, 1, 1, 0, 0,
                        GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                        new Insets(margin, margin, margin / 2, margin), 0, 0));

        pack();
        setLocationRelativeTo(null);
        setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
    }

    /**
     * Sets the number of awaited images.
     *
     * @param newTotalImages The new number of awaited images.
     */
    public final void setTotalImages(final int newTotalImages) {
        this.totalImages = newTotalImages;
    }

    /**
     * Reset the progress bar.
     */
    public final void reset() {
        receivedImages = 0;
        getProgressBar().setMaximum(totalImages);
        getProgressBar().setMinimum(0);
        getProgressBar().setValue(receivedImages);
    }

    @Override
    public void slavesCleared() { }

    @Override
    public void newSlave(final MasterEvent evt) { }

    @Override
    public void slaveQuit(final MasterEvent evt) { }

    @Override
    public final synchronized void snapshotSuccess(final MasterEvent evt) {
        addImage();
    }

    @Override
    public final synchronized void snapshotError(final MasterEvent evt) {
        addImage();
    }

    /**
     * Add a new image to the progress bar.
     */
    private void addImage() {
        receivedImages++;
        if (receivedImages >= totalImages) {
            dispose();
        }

        getProgressBar().setMaximum(totalImages);
        getProgressBar().setMinimum(0);
        getProgressBar().setValue(receivedImages);
    }

    @Override
    public void snapshotsCleared() { }

    @Override
    public void streamError(final MasterEvent evt) { }

    @Override
    public void log(final MasterEvent evt, final LogLevel level) { }

    @Override
    public void logLevelError(final MasterEvent evt) { }

    @Override
    public void logLevelSuccess(final MasterEvent evt) { }

    @Override
    public void options(final MasterEvent evt) { }

    @Override
    public void fileSaved(final MasterEvent evt) { }

}
