/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

/**
 * The constraints used to render components contained in a table layout.
 *
 * @author Zleurtor
 */
public class TableConstraints {

    /**
     * The row at which the component shall be start.
     */
    private final int row;

    /**
     * The column at which the component shall be start.
     */
    private final int column;

    /**
     * The number of cells occupied by the component horizontally.
     */
    private final int width;

    /**
     * The number of cells occupied by the component vertically.
     */
    private final int height;

    /**
     * Constructs a new table layout constraints for the specified row, column, width and height.
     *
     * @param theRow The row at which the component shall be start.
     * @param theColumn The column at which the component shall be start.
     * @param theWidth The number of cells occupied by the component horizontally.
     * @param theHeight The number of cells occupied by the component vertically.
     */
    public TableConstraints(final int theRow, final int theColumn, final int theWidth, final int theHeight) {
        super();

        if (theRow < 0) {
            throw new IllegalArgumentException("Row must be superior than or equals to 0");
        }
        if (theColumn < 0) {
            throw new IllegalArgumentException("Column must be superior than or equals to 0");
        }
        if (theWidth < 1) {
            throw new IllegalArgumentException("Width must be superior than or equals to 1");
        }
        if (theHeight < 1) {
            throw new IllegalArgumentException("Height must be superior than or equals to 1");
        }

        this.row = theRow;
        this.column = theColumn;
        this.width = theWidth;
        this.height = theHeight;
    }

    /**
     * Returns the row at which the component shall be start.
     *
     * @return The row at which the component shall be start.
     */
    public final int getRow() {
        return row;
    }

    /**
     * Returns the column at which the component shall be start.
     *
     * @return The column at which the component shall be start.
     */
    public final int getColumn() {
        return column;
    }

    /**
     * Returns the number of cells occupied by the component horizontally.
     *
     * @return The number of cells occupied by the component horizontally.
     */
    public final int getWidth() {
        return width;
    }

    /**
     * Returns the number of cells occupied by the component vertically.
     *
     * @return The number of cells occupied by the component vertically.
     */
    public final int getHeight() {
        return height;
    }

}
