/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager2;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.JComponent;

/**
 * A graphics layout consisting in a table (like HTML ones) with components that can occupy two (or more) cells.
 *
 * @author Zleurtor
 *
 */
public class TableLayout implements LayoutManager2 {

    /**
     * The number of rows of the layout.
     */
    private final int rows;

    /**
     * The number of columns of the layout.
     */
    private final int columns;

    /**
     * The horizontal gap between two columns.
     */
    private final int hgap;

    /**
     * The vertical gap between two rows.
     */
    private final int vgap;

    /**
     * The components and their constraints contained by this layout.
     */
    private final Map<Component, TableConstraints> componentConstraints = new HashMap<>();

    /**
     * Constructs a new table layout with the specified rows, columns and gap.
     *
     * @param theRows The number of rows of the layout.
     * @param theColumns The number of columns of the layout.
     * @param theHGap The horizontal gap between two columns.
     * @param theVGap The vertical gap between two rows.
     */
    public TableLayout(final int theRows, final int theColumns, final int theHGap, final int theVGap) {
        super();
        this.rows = theRows;
        this.columns = theColumns;
        this.hgap = theHGap;
        this.vgap = theVGap;
    }

    @Override
    @Deprecated
    public final void addLayoutComponent(final String name, final Component comp) {
        throw new RuntimeException("Not implemented");
    }

    @Override
    public final void addLayoutComponent(final Component comp, final Object constraints) {
        if (constraints == null || !(constraints instanceof TableConstraints)) {
            throw new IllegalArgumentException("cannot add to layout: constraint must be a TableConstraints");
        }

        componentConstraints.put(comp, (TableConstraints) constraints);
    }

    @Override
    public final void removeLayoutComponent(final Component comp) {
        componentConstraints.remove(comp);
    }

    @Override
    public final Dimension preferredLayoutSize(final Container parent) {
        int cellWidth = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellWidth = Math.max(cellWidth, entry.getKey().getPreferredSize().width / entry.getValue().getWidth());
        }

        int cellHeight = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellHeight = Math.max(cellHeight, entry.getKey().getPreferredSize().height / entry.getValue().getHeight());
        }

        int borderWidth = 0;
        int borderHeight = 0;
        if (parent instanceof JComponent && ((JComponent) parent).getBorder() != null) {
            final Insets borderInsets = ((JComponent) parent).getBorder().getBorderInsets(parent);
            borderWidth = borderInsets.left + borderInsets.right;
            borderHeight = borderInsets.top + borderInsets.bottom;
        } else {
            borderWidth = 0;
            borderHeight = 0;
        }

        final int totalHGap = (columns + 1) * this.hgap;
        final int totalVGap = (rows + 1) * this.hgap;
        return new Dimension(cellWidth * columns + borderWidth + totalHGap, cellHeight * rows + borderHeight + totalVGap);
    }

    @Override
    public final Dimension minimumLayoutSize(final Container parent) {
        int cellWidth = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellWidth = Math.max(cellWidth, entry.getKey().getMinimumSize().width / entry.getValue().getWidth());
        }

        int cellHeight = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellHeight = Math.max(cellHeight, entry.getKey().getMinimumSize().height / entry.getValue().getHeight());
        }

        int borderWidth = 0;
        int borderHeight = 0;
        if (parent instanceof JComponent && ((JComponent) parent).getBorder() != null) {
            final Insets borderInsets = ((JComponent) parent).getBorder().getBorderInsets(parent);
            borderWidth = borderInsets.left + borderInsets.right;
            borderHeight = borderInsets.top + borderInsets.bottom;
        } else {
            borderWidth = 0;
            borderHeight = 0;
        }

        final int totalHGap = (columns + 1) * this.hgap;
        final int totalVGap = (rows + 1) * this.hgap;
        return new Dimension(cellWidth * columns + borderWidth + totalHGap, cellHeight * rows + borderHeight + totalVGap);
    }

    @Override
    public final Dimension maximumLayoutSize(final Container parent) {
        int cellWidth = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellWidth = Math.max(cellWidth, entry.getKey().getMaximumSize().width / entry.getValue().getWidth());
        }

        int cellHeight = 0;
        for (final Entry<Component, TableConstraints> entry : componentConstraints.entrySet()) {
            cellHeight = Math.max(cellHeight, entry.getKey().getMaximumSize().height / entry.getValue().getHeight());
        }

        int borderWidth = 0;
        int borderHeight = 0;
        if (parent instanceof JComponent && ((JComponent) parent).getBorder() != null) {
            final Insets borderInsets = ((JComponent) parent).getBorder().getBorderInsets(parent);
            borderWidth = borderInsets.left + borderInsets.right;
            borderHeight = borderInsets.top + borderInsets.bottom;
        } else {
            borderWidth = 0;
            borderHeight = 0;
        }

        final int totalHGap = (columns + 1) * this.hgap;
        final int totalVGap = (rows + 1) * this.hgap;
        return new Dimension(cellWidth * columns + borderWidth + totalHGap, cellHeight * rows + borderHeight + totalVGap);
    }


    @Override
    public final float getLayoutAlignmentX(final Container target) {
        final float layoutAlignmentX = 0.5f;
        return layoutAlignmentX;
    }

    @Override
    public final float getLayoutAlignmentY(final Container target) {
        final float layoutAlignmentY = 0.5f;
        return layoutAlignmentY;
    }

    @Override
    public final void layoutContainer(final Container parent) {
        Insets borderInsets = new Insets(0, 0, 0, 0);
        if (parent instanceof JComponent && ((JComponent) parent).getBorder() != null) {
            borderInsets = ((JComponent) parent).getBorder().getBorderInsets(parent);
        }

        final double cellWidth = (parent.getSize().getWidth() - (borderInsets.left + borderInsets.right) - ((columns + 1) * hgap))
                                    / columns;
        final double cellHeight = (parent.getSize().getHeight() - (borderInsets.top + borderInsets.bottom) - ((rows + 1) * vgap))
                                    / rows;

        for (final Component comp : componentConstraints.keySet()) {
            final TableConstraints constraints = componentConstraints.get(comp);
            comp.setBounds(
                    (int) Math.round((cellWidth * constraints.getColumn()) + borderInsets.left + ((constraints.getColumn() + 1) * hgap)),
                    (int) Math.round((cellHeight * constraints.getRow()) + borderInsets.top + ((constraints.getRow() + 1) * vgap)),
                    (int) Math.round((cellWidth * constraints.getWidth()) + ((constraints.getWidth() - 1) * hgap)),
                    (int) Math.round((cellHeight * constraints.getHeight()) + ((constraints.getHeight() - 1) * vgap)));
        }
    }

    @Override
    public void invalidateLayout(final Container target) { }

}
