/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.io.IOException;

import javax.swing.border.Border;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.view.ResourcesManager;

/**
 * The border of the toolbar.
 *
 * @author Zleurtor
 */
public class ToolBarBorder implements Border {

    /**
     * A logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(ToolBarBorder.class);

    /**
     * The border's bottom image.
     */
    private static final Image BOTTOM;

    /**
     * The border's left image.
     */
    private static final Image LEFT;

    /**
     * The border's right image.
     */
    private static final Image RIGHT;

    static {
        Image bottom = null;
        Image left = null;
        Image right = null;

        try {
            bottom = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_TOOLBAR_BORDER_BOTTOM);
            left = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_TOOLBAR_BORDER_LEFT);
            right = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_TOOLBAR_BORDER_RIGHT);
        } catch (final IOException e) {
            GuiUtils.showWarning("Cannot load the border images: " + e.getMessage(), e, null);
            LOGGER.warn("Cannot load the border images: " + e.getMessage(), e);
        }

        BOTTOM = bottom;
        LEFT = left;
        RIGHT = right;
    }

    @Override
    public final void paintBorder(final Component c, final Graphics g, final int x, final int y, final int width, final int height) {
        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        final Insets i = getBorderInsets(c);

        g.drawImage(BOTTOM,
                i.left, c.getHeight() - i.bottom + 1, c.getWidth() - i.right, c.getHeight() + 1,
                0, 0, BOTTOM.getWidth(c), BOTTOM.getHeight(c),
                c);
        g.drawImage(LEFT,
                0, 0, i.left, c.getHeight(),
                0, 0, LEFT.getWidth(c), LEFT.getHeight(c),
                c);
        g.drawImage(RIGHT,
                c.getWidth() - i.right, 0, c.getWidth(), c.getHeight(),
                0, 0, RIGHT.getWidth(c), RIGHT.getHeight(c),
                c);
    }

    @Override
    public final Insets getBorderInsets(final Component c) {
        final int bottomHeight = BOTTOM.getHeight(c);
        final int height = c.getHeight() + bottomHeight;

        final double leftRatio = (double) LEFT.getWidth(c) / LEFT.getHeight(c);
        final double rightRatio = (double) LEFT.getWidth(c) / LEFT.getHeight(c);

        return new Insets(0, (int) Math.round(height * leftRatio), bottomHeight, (int) Math.round(height * rightRatio));
    }

    @Override
    public final boolean isBorderOpaque() {
        return false;
    }

}
