/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;
import java.net.MalformedURLException;
import java.text.MessageFormat;
import java.util.Map;
import java.util.TreeMap;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import zleurtor.scanmanager.model.Configuration;
import zleurtor.scanmanager.model.Utils;
import zleurtor.scanmanager.view.ResourcesManager;

/**
 * The dialog displaying the list of snapshot sets, allowing the user to choose which sets to delete.
 *
 * @author Zleurtor
 */
public class TrashDialog extends JDialog {

    private static final long serialVersionUID = 6893761654085801505L;

    /**
     * The max number of column displayed.
     */
    private static final int MAX_COLUMNS = 5;

    /**
     * The mapping between the check boxes and the corresponding snapshot sets.
     */
    private final Map<JCheckBox, File> snapshotSetsCheckBoxes = new TreeMap<>(
        (final JCheckBox cb1, final JCheckBox cb2) -> cb1.getText().compareTo(cb2.getText()));

    /**
     * The sets of snapshots selected by the user.
     */
    private File[] selectedSnapshotSets;

    /**
     * The button to select all check boxes.
     */
    private final JButton deSelectAllButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_DESELECTALL));

    /**
     * The button to cancel the trash action.
     */
    private final JButton cancelButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_CANCEL));

    /**
     * The button to launch the trash action on the selected snapshot sets.
     */
    private final JButton deleteButton =
            new JButton(ResourcesManager.getInstance().getString(ResourcesManager.KEY_BUTTON_DELETE));

    /**
     * The centerPanel (containing the check box list.
     */
    private final JPanel centerPanel = new JPanel() {
        private static final long serialVersionUID = -6845131611365199787L;

        @Override
        protected void paintComponent(final Graphics g) {
            super.paintComponent(g);

            final int alpha = 63;
            final float strokeWidth = 0.5f;
            g.setColor(new Color(
                    getForeground().getRed(),
                    getForeground().getGreen(),
                    getForeground().getBlue(),
                    alpha));
            ((Graphics2D) g).setStroke(new BasicStroke(strokeWidth, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));

            final double w = (double) getWidth() / nbColumns;
            final double h = (double) getHeight() / nbLines;

            for (double x = 0; x < getWidth(); x += w) {
                g.drawLine((int) Math.round(x), 0, (int) Math.round(x), getHeight());
            }
            g.drawLine(getWidth() - 1, 0, getWidth() - 1, getHeight());
            for (double y = 0; y < getHeight(); y += h) {
                g.drawLine(0, (int) Math.round(y), getWidth(), (int) Math.round(y));
            }
            g.drawLine(0, getHeight() - 1, getWidth(), getHeight() - 1);
        }
    };

    /**
     * The number of lines displayed.
     */
    private int nbLines;

    /**
     * The number of columns displayed.
     */
    private int nbColumns;

    /**
     * Constructs a new dialog having the specified window as owner.
     *
     * @param window The Window from which the dialog is displayed.
     */
    public TrashDialog(final Window window) {
        super(window, ResourcesManager.getInstance().getString(ResourcesManager.KEY_TITLE_TRASH), ModalityType.APPLICATION_MODAL);

        setContentPane(new ContentPanel());
        initializeComponents();
    }

    /**
     * Returns the sets of snapshots selected by the user.
     *
     * @return The sets of snapshots selected by the user.
     */
    public File[] getSelectedSnapshotSets() {
        return selectedSnapshotSets;
    }

    /**
     * Set a new list of snapshot sets.
     *
     * @param snapshotSets The new snapshot sets to set.
     */
    public void setSnapshotSets(final File[] snapshotSets) {
        mapSnapshotSetsToCheckBoxes(snapshotSets);
        initializeCheckBoxes();
    }

    /**
     * Initializes the components (buttons and listeners).
     */
    private void initializeComponents() {
        centerPanel.setOpaque(false);
        centerPanel.setLayout(new GridBagLayout());

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(createDeSelectAllButtonPanel(), BorderLayout.NORTH);
        getContentPane().add(new JScrollPane(centerPanel), BorderLayout.CENTER);
        getContentPane().add(createButtonsPanel(), BorderLayout.SOUTH);

        deSelectAllButton.addActionListener((final ActionEvent evt) -> {
            boolean atLeastOneNotSelected = false;
            for (final JCheckBox checkBox : snapshotSetsCheckBoxes.keySet()) {
                atLeastOneNotSelected |= !checkBox.isSelected();
            }

            for (final JCheckBox checkBox : snapshotSetsCheckBoxes.keySet()) {
                checkBox.setSelected(atLeastOneNotSelected);
            }
        });
        cancelButton.addActionListener((final ActionEvent evt) -> {
            selectedSnapshotSets = null;
            dispose();
        });
        deleteButton.addActionListener((final ActionEvent evt) -> {
            storeSelectedSnapshotSets();
            dispose();
        });
    }

    /**
     * Map the check boxes and the corresponding snapshot sets.
     *
     * @param snapshotSets The snapshots to map to their corresponding check boxes.
     */
    private void mapSnapshotSetsToCheckBoxes(final File[] snapshotSets) {
        snapshotSetsCheckBoxes.clear();

        for (final File snapshotSet : snapshotSets) {
            final int primary = Utils.ipToInt(Configuration.getInstance().getString("snapshot.primaryThumbnail"));
            int nearest = 0;
            String nearestThumbnail = "";

            for (final String thumbnailName : new File(snapshotSet, "thumbnails")
                    .list((File dir, String name) -> !new File(dir, name).isDirectory())) {
                final int thumbnailIp = Utils.ipToInt(thumbnailName.replace(".jpg", ""));
                if (Math.abs(primary - thumbnailIp) < Math.abs(primary - nearest)) {
                    nearest = thumbnailIp;
                    nearestThumbnail = thumbnailName;
                }
            }

            JCheckBox checkBox = null;
            try {
                checkBox = new JCheckBox(MessageFormat.format("<html>{0}<br><img src=\"{1}\"></html>",
                        snapshotSet.getName(), new File(new File(snapshotSet, "thumbnails"), nearestThumbnail).toURI().toURL()));
            } catch (final MalformedURLException e) {
                // Should never occur !
            }

            snapshotSetsCheckBoxes.put(checkBox, snapshotSet);
        }
    }

    /**
     * Initializes the snapshot sets check boxes.
     */
    private void initializeCheckBoxes() {
        centerPanel.removeAll();

        final int gap = 6;

        nbColumns = snapshotSetsCheckBoxes.size() < MAX_COLUMNS ? snapshotSetsCheckBoxes.size() : MAX_COLUMNS;
        nbLines = (snapshotSetsCheckBoxes.size() / MAX_COLUMNS) + (snapshotSetsCheckBoxes.size() % MAX_COLUMNS == 0 ? 0 : 1);

        int index = 0;
        for (final JCheckBox checkBox : snapshotSetsCheckBoxes.keySet()) {

            final int gridx = index % MAX_COLUMNS;
            final int gridy = index / MAX_COLUMNS;

            centerPanel.add(checkBox, new GridBagConstraints(gridx, gridy, 1, 1, 1, 1,
                    GridBagConstraints.WEST, GridBagConstraints.NONE,
                    new Insets(gridy == 0 ? gap : gap / 2, gridx == 0 ? gap : gap / 2,
                               gridy == nbLines - 1 ? gap : gap / 2, gridx == nbColumns - 1 ? gap : gap / 2),
                    0, 0));
            index++;
        }
    }

    /**
     * Creates the cancel/delete buttons panel.
     *
     * @return The newly created buttons panel.
     */
    private JPanel createButtonsPanel() {
        final int gap = 5;

        final JPanel theButtonsPanel = new JPanel();
        theButtonsPanel.setLayout(new GridBagLayout());

        final int cancelColumn = 0;
        final int deleteColumn = 1;

        theButtonsPanel.add(cancelButton,
                new GridBagConstraints(cancelColumn, 0, 1, 1, 1, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap / 2), 0, 0));
        theButtonsPanel.add(deleteButton,
                new GridBagConstraints(deleteColumn, 0, 1, 1, 0, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap / 2, gap / 2, gap, gap), 0, 0));

        return theButtonsPanel;
    }

    /**
     * Creates the select all button panel.
     *
     * @return The newly created buttons panel.
     */
    private JPanel createDeSelectAllButtonPanel() {
        final int gap = 5;

        final JPanel theSelectAllButtonPanel = new JPanel();
        theSelectAllButtonPanel.setOpaque(false);
        theSelectAllButtonPanel.setLayout(new GridBagLayout());

        final int selectAllColumn = 0;

        theSelectAllButtonPanel.add(deSelectAllButton,
                new GridBagConstraints(selectAllColumn, 0, 1, 1, 1, 0, GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE,
                        new Insets(gap, gap, gap, gap), 0, 0));

        return theSelectAllButtonPanel;
    }

    /**
     * Store the sets of snapshots selected by the user.
     */
    private void storeSelectedSnapshotSets() {
        int length = 0;
        for (final JCheckBox checkBox : snapshotSetsCheckBoxes.keySet()) {
            if (checkBox.isSelected()) {
                length++;
            }
        }

        selectedSnapshotSets = new File[length];
        int index = 0;
        for (final JCheckBox checkBox : snapshotSetsCheckBoxes.keySet()) {
            if (checkBox.isSelected()) {
                selectedSnapshotSets[index++] = snapshotSetsCheckBoxes.get(checkBox);
            }
        }
    }

}
