/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.gui;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.geom.AffineTransform;
import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.JPanel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zleurtor.scanmanager.view.ResourcesManager;

/**
 * A panel displaying a wait animation.
 */
public class WaitPanel extends JPanel implements ComponentListener {

    private static final long serialVersionUID = 2546248206324654210L;

    /**
     * A Logger...
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(WaitPanel.class);

    /**
     * The wait image used to display the wait animation.
     */
    private static final Image WAIT_IMAGE;

    /**
     * The angle to add at each step.
     */
    private static final double DELTA_ANGLE = Math.PI / 10;

    /**
     * A timer having in charge the redrawing of this component.
     */
    private Timer timer = null;

    static {
        Image waitImage = null;
        try {
            waitImage = ResourcesManager.getInstance().getImage(ResourcesManager.KEY_IMAGE_WAIT);
        } catch (final IOException e) {
            GuiUtils.showWarning(e, null);
            LOGGER.warn(e.getMessage(), e);
        }
        WAIT_IMAGE = ResourcesManager.colorize(waitImage,
                ResourcesManager.getInstance().getColor(ResourcesManager.KEY_WAITPANEL_FOREGROUND));
    }

    /**
     * The current step.
     */
    private int step = 0;

    /**
     * Constructs a new wait panel.
     */
    public WaitPanel() {
        super(true);

        this.addComponentListener(this);
    }

    @Override
    protected final void paintComponent(final Graphics g) {
        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);

        final double locationX = WAIT_IMAGE.getWidth(this) / 2;
        final double locationY = WAIT_IMAGE.getHeight(this) / 2;
        final AffineTransform rx = AffineTransform.getTranslateInstance(
                (getWidth() - WAIT_IMAGE.getWidth(this)) / 2,
                (getHeight() - WAIT_IMAGE.getHeight(this)) / 2);
        rx.concatenate(AffineTransform.getRotateInstance(step * DELTA_ANGLE, locationX, locationY));
        ((Graphics2D) g).drawImage(WAIT_IMAGE, rx, this);
    }

    @Override
    public final Dimension getPreferredSize() {
        if (WAIT_IMAGE == null) {
            return new Dimension(0, 0);
        } else {
            return new Dimension(WAIT_IMAGE.getWidth(this), WAIT_IMAGE.getHeight(this));
        }
    }

    @Override
    public final void componentResized(final ComponentEvent evt) {
        final int oneSecond = 1000;

        timer = new Timer();
        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                step++;
                if (step * DELTA_ANGLE >= 2 * Math.PI) {
                    step = 0;
                }
                if (WaitPanel.this.getParent() != null) {
                    WaitPanel.this.getParent().repaint();
                }
            }
        },
                oneSecond / ResourcesManager.getInstance().getInt(ResourcesManager.KEY_WAITPANEL_FPS),
                oneSecond / ResourcesManager.getInstance().getInt(ResourcesManager.KEY_WAITPANEL_FPS));
    }

    @Override
    public final void componentMoved(final ComponentEvent evt) { }

    @Override
    public final void componentShown(final ComponentEvent evt) { }

    @Override
    public final void componentHidden(final ComponentEvent evt) {
        if (timer != null) {
            timer.cancel();
            timer.purge();
            timer = null;
        }
    }

}
