/*
 * 3D Scanner Manager
 * Copyright (c) 2021-2024 Zleurtor
 *
 * This file is part of 3D Scanner Manager.
 *
 * 3D Scanner Manager is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Scanner Manager is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Scanner Manager.  If not, see <https://www.gnu.org/licenses/>.
 */

package zleurtor.scanmanager.view.plaf;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Point2D;

import javax.swing.JComponent;
import javax.swing.JSlider;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.basic.BasicSliderUI;

import zleurtor.scanmanager.view.ResourcesManager;

/**
 * The 3D Scan Manager personalized label UI.
 *
 * @author Zleurtor
 */
public class ScanManagerSliderUI extends BasicSliderUI implements MouseListener, MouseMotionListener, MouseWheelListener {

    /**
     * The color of the slider's cursor.
     */
    private static final Color THUMB_COLOR = ResourcesManager.getInstance().getColor(ResourcesManager.KEY_SLIDER_THUMBCOLOR);

    /**
     * The thumb's width.
     */
    private static final int THUMB_WIDTH = 12;

    /**
     * The thumb's width.
     */
    private static final int THUMB_HEIGHT = 12;

    /**
     * The distance that the track is from the side of the control.
     */
    private static final int TRACK_BUFFER = 24;

    /**
     * The track background color.
     */
    private static final Color TRACK_COLOR = ResourcesManager.getInstance().getColor(ResourcesManager.KEY_RULE_TRACKCOLOR);

    /**
     * The track width.
     */
    private static final int TRACK_WIDTH = 4;

    /**
     * The ticks' color.
     */
    private static final Color TICK_COLOR = ResourcesManager.getInstance().getColor(ResourcesManager.KEY_SLIDER_FOREGROUND);

    /**
     * The width of the increment/decrement buttons.
     */
    private static final int BUTTONS_WIDTH = 16;

    /**
     * Whether the decrement button is focused.
     */
    private boolean decrementFocused = false;

    /**
     * Whether the increment button is focused.
     */
    private boolean incrementFocused = false;

    /**
     * Constructs a ScanManagerSliderUI.
     *
     * @param theSlider A slider.
     */
    public ScanManagerSliderUI(final JSlider theSlider) {
        super(theSlider);
        theSlider.addChangeListener(evt -> theSlider.repaint());
    }

    /**
     * Create the UI.
     *
     * @param c The component to which the UI shall be applied.
     * @return The unique UI instance.
     */
    public static ComponentUI createUI(final JComponent c) {
        final ScanManagerSliderUI sliderUI = new ScanManagerSliderUI((JSlider) c);

        final JSlider s = (JSlider) c;
        s.setOpaque(false);
        s.addMouseListener(sliderUI);
        s.addMouseMotionListener(sliderUI);
        s.addMouseWheelListener(sliderUI);

        return sliderUI;
    }

    @Override
    protected final Dimension getThumbSize() {
        return new Dimension(THUMB_WIDTH, THUMB_HEIGHT);
    }

    @Override
    protected final void calculateTrackBuffer() {
        trackBuffer = TRACK_BUFFER;
    }

    @Override
    public final void paint(final Graphics g, final JComponent c) {
        super.paint(g, c);

        paintButtons(g);
    }

    @Override
    public final void paintFocus(final Graphics g) { }

    @Override
    public final void paintThumb(final Graphics g) {
        ((Graphics2D) g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

        g.setColor(THUMB_COLOR);
        g.fillOval(thumbRect.x, thumbRect.y, thumbRect.width, thumbRect.height);
    }

    @Override
    public final void paintTrack(final Graphics g) {
        g.setColor(TRACK_COLOR);
        g.fillRect(thumbRect.x, trackRect.y + trackRect.height / 2 - TRACK_WIDTH / 2,
                   trackRect.width - (thumbRect.x - trackRect.x), TRACK_WIDTH);

        final int alphaRatio = 4;
        ((Graphics2D) g).setPaint(new GradientPaint(
                new Point2D.Double(trackRect.x, trackRect.y + trackRect.getHeight() / 2),
                new Color(THUMB_COLOR.getRed(), THUMB_COLOR.getGreen(), THUMB_COLOR.getBlue(), THUMB_COLOR.getAlpha() / alphaRatio),
                new Point2D.Double(thumbRect.x, trackRect.y + trackRect.getHeight() / 2),
                THUMB_COLOR));

        g.drawRect(trackRect.x, trackRect.y + trackRect.height / 2 - TRACK_WIDTH / 2, thumbRect.x - trackRect.x, TRACK_WIDTH);
    }

    @Override
    public final void paintTicks(final Graphics g) {
        g.setColor(TICK_COLOR);

        for (double i = slider.getMinimum(); i <= slider.getMaximum(); i += slider.getMajorTickSpacing()) {
            final double r = (i - slider.getMinimum()) / (slider.getMaximum() - slider.getMinimum());
            final int x = trackRect.x + (int) Math.round(r * trackRect.width);
            g.drawLine(x, trackRect.y,
                       x, trackRect.y - 2);
            g.drawLine(x, trackRect.y + trackRect.height,
                       x, trackRect.y + trackRect.height + 2);
        }
    }

    @Override
    public final void scrollByBlock(final int direction) {
        if (decrementFocused || incrementFocused) {
            if (direction == BasicSliderUI.POSITIVE_SCROLL) {
                slider.setValue(slider.getValue() + 1);
            } else if (direction == BasicSliderUI.NEGATIVE_SCROLL) {
                slider.setValue(slider.getValue() - 1);
            }
        } else {
            super.scrollByBlock(direction);
        }
    }

    /**
     * Paints the decrease/increase buttons.
     *
     * @param g The graphics in wich the buttons shall be painted.
     */
    public final void paintButtons(final Graphics g) {
        // Draw the buttons
        if (decrementFocused) {
            g.setColor(THUMB_COLOR);
        } else {
            g.setColor(TICK_COLOR);
        }

        g.drawOval(0, 0, BUTTONS_WIDTH, BUTTONS_WIDTH);
        int w = g.getFontMetrics().stringWidth("-");
        int h = g.getFontMetrics().getHeight();
        final int decrementDeltaY = 5;
        g.drawString("-",
                0 + BUTTONS_WIDTH / 2 - w / 2 + 1,
                0 + BUTTONS_WIDTH / 2 + h / 2 - decrementDeltaY);

        if (incrementFocused) {
            g.setColor(THUMB_COLOR);
        } else {
            g.setColor(TICK_COLOR);
        }

        g.drawOval(slider.getWidth() - BUTTONS_WIDTH - 1, 0, BUTTONS_WIDTH, BUTTONS_WIDTH);
        w = g.getFontMetrics().stringWidth("+");
        h = g.getFontMetrics().getHeight();
        final int incrementDeltaY = 3;
        g.drawString("+",
                (slider.getWidth() - BUTTONS_WIDTH) + BUTTONS_WIDTH / 2 - w / 2,
                0 + BUTTONS_WIDTH / 2 + h / 2 - incrementDeltaY);
    }

    @Override
    public final void mouseClicked(final MouseEvent evt) { }

    @Override
    public final void mousePressed(final MouseEvent evt) { }

    @Override
    public final void mouseReleased(final MouseEvent evt) { }

    @Override
    public final void mouseEntered(final MouseEvent evt) { }

    @Override
    public final void mouseExited(final MouseEvent evt) {
        decrementFocused = false;
        incrementFocused = false;

        slider.repaint();
    }

    @Override
    public final void mouseDragged(final MouseEvent evt) { }

    @Override
    public final void mouseMoved(final MouseEvent evt) {
        decrementFocused = false;
        incrementFocused = false;
        slider.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));

        if (evt.getX() >= 0 && evt.getX() < BUTTONS_WIDTH
                && evt.getY() >= 0 && evt.getY() < BUTTONS_WIDTH) {
            decrementFocused = true;
            slider.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        } else if (evt.getX() >= slider.getWidth() - BUTTONS_WIDTH && evt.getX() < slider.getWidth()
                && evt.getY() >= 0 && evt.getY() < BUTTONS_WIDTH) {
            incrementFocused = true;
            slider.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        } else if (thumbRect.contains(evt.getPoint())) {
            slider.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        }

        slider.repaint();
    }

    @Override
    public final void mouseWheelMoved(final MouseWheelEvent evt) {
        if ((evt.getModifiersEx() & MouseWheelEvent.CTRL_DOWN_MASK) == MouseWheelEvent.CTRL_DOWN_MASK) {
            final int millis = 1_000;
            slider.setValue((int) Math.round(Math.max(slider.getMinimum(),
                                             Math.min(slider.getMaximum(),
                                                      slider.getValue() + millis * evt.getPreciseWheelRotation()))));
        } else {
            slider.setValue((int) Math.round(Math.max(slider.getMinimum(),
                                             Math.min(slider.getMaximum(),
                                                      slider.getValue() + evt.getPreciseWheelRotation()))));
        }
    }

}
